
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

import java.io.*;
import java.util.*;
import java.util.zip.*;


/*
 *  BuildOneWireProgram KLA
 *
 *  History:
 *  09-15-2000 First version, used TINI overlay to help 
 *             transition from TINI1.01 to use OneWireAPI 0.00.
 *  11-14-2000 Updated for TINI 1.02, which has the new OneWireAPI.  
 *             Changed to pass unknown arguments to TINIConvertor,
 *             so that all the power of TINIConvertor is accessible.
 *  11-30-2000 Updated to allow containers to be loaded from a JAR file.
 *
 */

public class BuildOneWireProgram
{
    static Vector filesToBeDeleted = new Vector();
    static boolean useJAR = false;
    static String jarname    = "";
    static byte[] buffer = null;
    static boolean DEBUG = false;

   /**
    * Method main
    *
    *
    * @param args
    *
    * @throws IOException
    *
    */
   public static void main (String[] args)
      throws IOException
   {
      if ((args == null) || (args.length == 0))
      {
         printUsageString();
         return;
      }

      Vector containers = new Vector();
      Vector originals = new Vector();
      String path       = "com/dalsemi/onewire/container/OneWireContainer";
      String prepath    = "";
      String dep_path   = "";
      

      boolean depListOnly = false;
      String depTarget = null;

      //first let's see if TINIConvertor is in our classpath
      try
      {
         Class c = Class.forName("TINIConvertor");
      }
      catch (Exception e)
      {
         System.out.println("TINIConvertor is not in your classpath!");
         System.out.println("***************************************\r\n");
         System.out.println("Make sure the file tini.jar is in your classpath.");
         System.out.println("You can do this from the command line by calling");
         System.out.println("    java -classpath d:\\tini\\bin\\tini.jar;%classpath% BuildOneWireProgram [OPTIONS]");
         System.out.println("\r\nIt may also be useful to include '.' in your classpath:");
         System.out.println("    java -classpath d:\\tini\\bin\\tini.jar;.;%classpath% ...");
         return;
      }
      
      Vector iButtonNames = new Vector(5);
      int totalArgs = args.length;

      for (int i = 0; i < args.length; i++)
      {
         String thisarg = args [i];

         if (thisarg.equals("-debug"))
         {
            DEBUG = true;
            args[i] = null;
            totalArgs--;
         }
         else if (thisarg.startsWith("-dep"))
         {
            if (thisarg.equals("-dep"))
            {
                depListOnly = true;
            }
            else
            {
                depListOnly = true;
                depTarget = thisarg.substring(4);
            }
         }
         else if (thisarg.equals("-p"))
         {

            //then the next string is the desired path to the containers
            i++;

            prepath = args [i];
            
            try
            {
                File f = new File(prepath);
                if (!f.exists())
                {
                    System.out.println("Error: Cannot find the File/Directory with the One Wire Containers.");
                    System.exit(0);
                }
                if (!f.isDirectory())
                {
                    useJAR = true;
                    jarname = prepath;
                    prepath = "";
                    buffer = new byte[2048];
                }
            }
            catch(Exception e)
            {
                System.out.println("Error trying to validate path to One Wire Containers: " +e.toString());
                System.exit(0);
            }
                
            
            
            args[i-1] = null;
            args[i] = null;
            totalArgs -= 2;
         }
         else if (thisarg.equals("-x"))
         {
            //then the next string is the desired path to the TINI dependency file
            i++;

            dep_path = args [i];

            args[i-1] = null;
            args[i] = null;
            totalArgs -= 2;
         }
         else
         {
            if (isTwoDigitHex(thisarg))
            {
                iButtonNames.addElement(thisarg);
                args[i] = null;
                totalArgs --;
            }
         }
      }
      
      initializeDependencyFile(dep_path);

      if (depListOnly) 
      {
         if (depTarget==null)  //then print them all
         {
             Enumeration enum = dependency.keys();
             for (Object o = enum.nextElement();enum.hasMoreElements();o = enum.nextElement())
             {
                 String key = (String)o;
                 System.out.println("Dependency list for OneWireContainer"+key+":");
                 String[] dep = getDependencies(key);
                 if (dep.length==0)
                 {
                     System.out.println("There are no dependencies for OneWireContainer"+key+".");
                 }
                 else
                 {
                     for (int k=0;k<dep.length;k++)
                     {
                         System.out.println(k + " : "+dep[k]);
                     }
                 }
                 System.out.println("---------------------------------------");
             }
             System.exit(0);
         }
         else  //then only print container %depTarget%'s depdendency list
         {
             System.out.println("Dependency list for OneWireContainer"+depTarget+":");
             String[] dep = getDependencies(depTarget);
             if (dep.length==0)
             {
                 if (dependency.get(depTarget)==null)
                 {
                     System.out.println("There is no entry for a OneWireContainer"+depTarget+".");
                     System.exit(0);
                 }
                 else
                 {
                     System.out.println("There are no dependencies for OneWireContainer"+depTarget+".");
                     System.exit(0);
                 }
             }
             else
             {
                 for (int k=0;k<dep.length;k++)
                 {
                     System.out.println(k + " : "+dep[k]);
                 }
                 System.exit(0);
             }
         }
      }


      try
      {
            for (int i=0;i<iButtonNames.size();i++)
            {
                    //now add the dependencies here...
                    String[] dep = getDependencies((String)iButtonNames.elementAt(i));

                    for (int x = 0; x < dep.length; x++)
                    {
                        if (DEBUG) System.out.println("Dep " + x + " : " + dep [x]);

                        String element = prepath + "com/dalsemi/onewire/container/"
                                            + dep [x] + ".class";
                        String original_element = element;                                            

                         if (originals.indexOf(element) == -1)
                         {
                            originals.addElement(element);
                            element = fileCheck(element);                                
                            containers.addElement(element);
                         }
                    }

                    String ibutton = prepath + path + (String)iButtonNames.elementAt(i) + ".class";
                    if (originals.indexOf(ibutton) == -1)
                    {
                        originals.addElement(ibutton);
                        ibutton = fileCheck(ibutton);
                        containers.addElement(ibutton);
                    }
            }

            String[] arguments = new String[2 * containers.size() + totalArgs];
            int j=0;

            for (int i = 0; i < containers.size(); i++)
            {
                arguments [j++] = "-f";
                arguments [j++] = (String)containers.elementAt(i);
                if (DEBUG)
                    System.out.println("-f arg: "+(String)containers.elementAt(i));                
            }
            for (int i=0;i<args.length;i++)
            {
                if (args[i]!=null)
                    arguments [j++] = args[i];
            }

            PrintStream out = System.out;
            try
            {

                PrintStream os = new PrintStream(new ByteArrayOutputStream());
                if (!DEBUG) System.setOut(os);
                
                TINIConvertor.main(arguments);

                System.setOut(out);
                out.println("\r\n\r\nBuild successful");
            }
            catch (Exception e)
            {

                System.setOut(out);
                out.println("\r\n\r\nCould not create the application: "
                                    + e.toString());
            }
      }
      catch(Exception uhoh)
      {
            //need to clean up, which we'll do just below...so let this drain
      }
      
      //clean up - delete any temporary files
      while (filesToBeDeleted.size()>0)
      {
          String curr_filename = (String)filesToBeDeleted.elementAt(0);
          filesToBeDeleted.removeElementAt(0);
          try
          {
              File death_file = new File(curr_filename);
              death_file.delete();
          }
          catch(Exception e)
          {
              System.out.println("Could not delete the file "+curr_filename);
              //drain it...let's try to delete the next file
          }
      }
   }

   private static Hashtable dependency   = null;
   private static String    dep_filename = null;

   private static void initializeDependencyFile (String filename)
   {
      if (dep_filename == null)
         dep_filename = new String(filename);
      else
      {
         if (dep_filename.equals(filename))
            return;   //we've already initialized on this file
         else         //we have not yet initialized on this file
            dep_filename = new String(filename);
      }

      boolean fileExists = false;
      try
      {
         File test_file = new File(dep_filename);
         if (test_file.exists())
             fileExists = true;
      }
      catch(Exception e)
      {
         //assume file doesn't exist...
      }
      
      Hashtable temp = new Hashtable();

      if (fileExists)
      {
        try
        {
            FileInputStream fin  = new FileInputStream(dep_filename);
            BufferedReader  in   =
                new BufferedReader(new InputStreamReader(fin));
            String          line = in.readLine();

            while (line != null)
            {
                if (line.length() > 1)
                {

                //line should be 2 characters of hex, then a space, then all the dependencies
                String hex = line.substring(0, 2);

                if (line.length() > 2)
                    line = line.substring(3);
                else
                    line = "";

                //                    System.out.println("Hash "+hex+", storing "+line);
                temp.put(hex, line);
                }

                line = in.readLine();
            }
        }
        catch (Exception e)
        {
            System.out.println("Failed to find dependencies: " + e.toString());

            return;
        }

        dependency = temp;
        return;
      }
      //else...
      System.out.println("Could not find dependency file, using defaults (as released in OneWireAPI 0.00)");
      temp.put("01","");
      temp.put("02","");
      temp.put("04", "../utils/Bit ClockContainer MemoryBank MemoryBankNV MemoryBankROM MemoryBankScratch OneWireSensor PagedMemoryBank ScratchPad");
      temp.put("05", "OneWireSensor SwitchContainer");
      temp.put("06", "MemoryBank MemoryBankNV MemoryBankROM MemoryBankScratch PagedMemoryBank ScratchPad");
      temp.put("08", "MemoryBank MemoryBankNV MemoryBankROM MemoryBankScratch PagedMemoryBank ScratchPad");
      temp.put("09", "MemoryBank MemoryBankEPROM MemoryBankROM MemoryBankScratch OTPMemoryBank PagedMemoryBank");
      temp.put("0A", "MemoryBank MemoryBankNV MemoryBankROM MemoryBankScratch PagedMemoryBank ScratchPad");
      temp.put("0B", "MemoryBank MemoryBankEPROM MemoryBankROM OTPMemoryBank PagedMemoryBank");
      temp.put("0C", "MemoryBank MemoryBankNV MemoryBankROM MemoryBankScratch PagedMemoryBank ScratchPad");
      temp.put("0F", "MemoryBank MemoryBankEPROM MemoryBankROM OTPMemoryBank PagedMemoryBank");
      temp.put("10", "OneWireSensor TemperatureContainer");
      temp.put("12", "MemoryBank MemoryBankEPROM MemoryBankROM OTPMemoryBank PagedMemoryBank OneWireSensor SwitchContainer");
      temp.put("13", "MemoryBank MemoryBankEPROM MemoryBankROM OTPMemoryBank PagedMemoryBank");
      temp.put("14", "MemoryBank MemoryBankAppReg MemoryBankEE MemoryBankROM OTPMemoryBank PagedMemoryBank");
      temp.put("16", "CommandAPDU JibComm$BlockDataFragmenter JibComm ResponseAPDU");
      temp.put("18", "MemoryBank MemoryBankNV MemoryBankNVCRC MemoryBankROM MemoryBankScratch MemoryBankScratchCRC MemoryBankScratchEx MemoryBankScratchSHA PagedMemoryBank ScratchPad SHAiButton");
      temp.put("1A", "MemoryBank MemoryBankNV MemoryBankNVCRC MemoryBankROM MemoryBankScratch MemoryBankScratchEx PagedMemoryBank ScratchPad");
      temp.put("1D", "MemoryBank MemoryBankNV MemoryBankNVCRC MemoryBankROM MemoryBankScratch MemoryBankScratchEx PagedMemoryBank ScratchPad");
      temp.put("1F", "../utils/Bit OneWireSensor SwitchContainer");
      temp.put("20", "../utils/Bit ADContainer MemoryBank MemoryBankAD MemoryBankROM OneWireSensor PagedMemoryBank");
      temp.put("21", "MemoryBank MemoryBankNV MemoryBankNVCRC MemoryBankROM MemoryBankScratch MemoryBankScratchCRC MemoryBankScratchEx ClockContainer OneWireSensor TemperatureContainer PagedMemoryBank ScratchPad");
      temp.put("23", "MemoryBank MemoryBankNV MemoryBankROM MemoryBankScratch MemoryBankScratchEE PagedMemoryBank ScratchPad");
      temp.put("26", "ADContainer ClockContainer OneWireSensor TemperatureContainer");
      temp.put("28", "OneWireSensor TemperatureContainer");
      temp.put("2C", "OneWireSensor PotentiometerContainer");
      temp.put("30", "ADContainer OneWireSensor TemperatureContainer");
      dependency = temp;
   }

   private static String[] getDependencies (String hex)
   {
      if (dependency == null)
         return new String [0];

      //hex should be a 2 digit hex string
      String          line = ( String ) dependency.get(hex);
      if (line==null)
      {
        return new String[0];
        /*System.out.println("Cannot find data on this container: "+hex);
        System.out.println("Make sure it has a container in the correct place, and that there is a dependency file for it.");
        System.exit(0);*/
      }
      StringTokenizer st   = new StringTokenizer(line);
      int             len  = st.countTokens();
      String[]        temp = new String [len];
      int             i    = 0;

      while (st.hasMoreTokens())
      {
         temp [i++] = st.nextToken();
      }

      return temp;
   }


   static boolean isTwoDigitHex(String arg)
   {
      if (arg.length()!=2) return false;
      char c1 = arg.charAt(0);
      char c2 = arg.charAt(1);
      if (c1 < 0x30) return false;
      if (c1 > 0x39)
      {
         if (c1 < 0x41) return false;
         if (c1 > 0x46) return false;
      }
      if (c2 < 0x30) return false;
      if (c2 > 0x39)
      {
         if (c2 < 0x41) return false;
         if (c2 > 0x46) return false;
      }
      return true;
   }
   
   static String fileCheck(String filename) throws Exception
   {
       if (DEBUG)
            System.out.println("filecheck on "+filename);
       try
       {
            File f = new File(filename);
            if (!f.exists())
            {
                //first check the jar file
                if (useJAR)
                {
                    //we need to see here if filename has anything like blah/anotherblah/../file.txt in it
                    //java zip no lo comprende, so we need to go ahead and parse that
                    
                    int parent_index = filename.indexOf("..");
                    if (parent_index != -1)
                    {
                        if (DEBUG)
                            System.out.println("File name was "+filename);
                        
                        char[] c_filename = filename.toCharArray();
                        
                        int endregion = parent_index + 2;
                        parent_index -= 2;
                        while ((c_filename[parent_index] != '/') && (c_filename[parent_index] != '\\'))
                        {
                            parent_index--;
                        }
                        int startregion = parent_index;

                        filename = filename.substring(0,startregion) + filename.substring(endregion);
                        if (DEBUG)  
                            System.out.println("Filename is "+filename);                        
                    }

                    
                    String old_filename = filename;
                    
                    try
                    {
                        //open the zip and look for the entry
                        ZipFile zip = new ZipFile(jarname);
                        filename = filename.replace('/','_');
                        filename = filename.replace('\\','_');
                        
                        f = new File(filename);
                        while (f.exists())
                        {
                            filename = "_"+filename;
                            f = new File(filename);
                        }
                        //now filename is the name of the file we will write to
                        ZipEntry ze = zip.getEntry(old_filename);
                        if (ze.isDirectory())
                        {
                            System.out.println("This zip entry is a directory: "+old_filename);
                            throw new Exception();
                        }

                        filesToBeDeleted.addElement(filename);
                        FileOutputStream fout = new FileOutputStream(f);
                        
                        InputStream fin = zip.getInputStream(ze);
                        
                        int amount_read = fin.read(buffer);
                        while (amount_read != -1)
                        {
                            fout.write(buffer,0,amount_read);
                            amount_read = fin.read(buffer);
                        }
                        
                        fin.close();
                        fout.close();
                        zip.close();
                        return filename;  
                    }
                    catch(Exception e1)
                    {
                        System.out.println("There was a problem accessing the zip file "+jarname+", looking for "+old_filename);
                        throw new Exception();
                    }
                }                
                
                System.out.println("\r\nCannot find the file "+filename+".");
                System.out.println("This file is required to build the TINI application.");
                throw new Exception();  //tell the main to clean up
            }
       }
       catch(Exception e)
       {
            System.out.println("Exception occurred checking for file existance: "+e.toString());
            throw new Exception();  //tell the main to clean up
       }
       return null;  //should not get here...
   }

   public static void printUsageString()
   {
         System.out.println("Dallas Semiconductor TINI1.02 OneWire Program Building Tool (11-30-00 KLA)");
         System.out.println("---------------------------------------------------------------------------\r\n");
         System.out.println("This program is meant to be used to help build one wire applications for the");
         System.out.println("TINI board, firmware release 1.02.  With the release of the OneWireAPI for Java,");
         System.out.println("many OneWireContainers have several dependencies that must be built into any");
         System.out.println("TINI application.  This program helps abstract developers from those");
         System.out.println("depdendencies.\r\n\r\n");
         System.out.println("Usage: ");
         System.out.println("   java BuildOneWireProgram [options] [containers]");
         System.out.println();
         System.out.println("options: -p PATH     -- Path to your containers.  If your container ");
         System.out.println("                        OneWireContainerXX lives in the directory ");
         System.out.println("                        d:\\develop\\com\\dalsemi\\onewire\\container\\,");
         System.out.println("                        then use 'd:\\develop\\' for PATH");
         System.out.println("         -x DEP_FILE -- Filename of the dependency text database file.  This");
         System.out.println("                        file is called tini_dependencies.txt in the OneWireAPI");
         System.out.println("                        0.00, located in the onewire distribution's lib");
         System.out.println("                        directory.  Use the complete filename for this");
         System.out.println("                        argument.");
         System.out.println("                        If a dependency file is not specified or does not");
         System.out.println("                        exist, then a default dependency map will be used");
         System.out.println("                        that matches the dependency file released in the");
         System.out.println("                        OneWireAPI version 0.00");
         System.out.println("         -debug      -- See the entire output of TINIConvertor");
         System.out.println("         -dep        -- See the entire dependency list using the specified");
         System.out.println("                        depdendency file or the default.");
         System.out.println("         -depXX      -- See the depdendency list for OneWireContainerXX.");
         System.out.println("         -Any TINIConvertor Option");
         System.out.println("                     -- BuildOneWireProgram now supports all TINIConvertor");
         System.out.println("                        options by directly passing any other options to");
         System.out.println("                        TINIConvertor.");
         System.out.println("containers: Two digit hex numbers corresponding to the OneWire containers");
         System.out.println("            to build into your application.");
         System.out.println();
         System.out.println("For example, for a program that uses containers for the DS1920, DS1921, and");
         System.out.println("the DS1996, I might try running: ");
         System.out.println("   java BuildOneWireProgram -f ReadTemp.class 10 21 0C -o test.tini -d d:\\tini1.01\\bin\\tini.db");
         System.out.println();
         System.out.println("************************************************************************");
         System.out.println("* NOTE: 1. Container hex digits must be uppercase--'0C' and not '0c'   *");
         System.out.println("*       2. THE CONTAINERS MUST BE LOCATED IN THE DIRECTORY             *");
         System.out.println("*             PATH/com/dalsemi/onewire/container/                      *");
         System.out.println("*       3. THE CONTAINERS MUST ALREADY BE COMPILED!                    *");
         System.out.println("*       4. PATH and DEP_PATH default to the current directory          *");
         System.out.println("*       5. Any argument (-f filename, -o outputfile, etc.) with a name *");
         System.out.println("*          that is two hex digits will not function correctly.         *");
         System.out.println("*          BuildOneWireProgram will interpret those digits as the name *");
         System.out.println("*          of a OneWireContainer.                                      *");
         System.out.println("************************************************************************");
   }

}
