/**
 * Copyright (C) 1997-2000 Dallas Semiconductor Corporation.
 * All rights Reserved. Printed in U.S.A.
 * This software is protected by copyright laws of
 * the United States and of foreign countries.
 * This material may also be protected by patent laws of the United States
 * and of foreign countries.
 * This software is furnished under a license agreement and/or a
 * nondisclosure agreement and may only be used or copied in accordance
 * with the terms of those agreements.
 * The mere transfer of this software does not imply any licenses
 * of trade secrets, proprietary technology, copyrights, patents,
 * trademarks, maskwork rights, or any other form of intellectual
 * property whatsoever. Dallas Semiconductor retains all ownership rights.
 */
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * <p>This class represents an initialization file.  The data is initially stored
 * in a file, then all of the data is transfered into memory where all of the
 * key/value pairs can be accessed from this class.  Data is only stored back
 * into the specified file if the method to do so is used.
 * <p>The file has a user specified header version string which can be used to
 * determine that the ini file is supported by the current version of the software
 * <p>The file consistes of any number of key/value pairs separated into groups.
 * <p>All fields are case sensitive and should not begin or end with whitespace
 * characters.
 */

public class IniFile extends Object
{
    private String version = null;
    private String filename = null;
    private File file = null;
    private Vector groups = null;
    private int lineNumber = 0;


    /**
     * Class Constructor.
     * <p>Creates an empty <code>IniFile</code> with 0 (zero) groups.
     * @param filename the ini file name that will be written to or read from
     * @param version the version string used for this <code>IniFile</code>
     */
    public IniFile(String filename, String version)
    {
        this.version = version;
        this.filename = filename;
        groups = new Vector(4);
    }


    /**
     * Read the initialization file into memory.
     * <p>The <code>IniFile</code> object will be filled with data from the
     * specified file if it exists and has the same version string.
     * @exception IniFileException
     */
    public void readFromDisk() throws IniFileException
    {
        try
        {
            file = new File( filename );

            if ( file.exists() )
                readFileIntoGroups();
            else    // file didn't exist so groups remains empty
            {
            }
        }
        catch(IOException e)
        {
            throw new IniFileException( "Could not test existance of file",
                                        IniFileException.READWRITE_ERROR );
        }
    }

    private void readFileIntoGroups()
        throws IniFileException, IOException
    {
        if ( (file.canRead() == false) || (file.canWrite() == false) )
            throw new IniFileException("Must be able to read and write to file",
                                       IniFileException.READWRITE_ERROR);
        BufferedReader input = new BufferedReader(
                                   new InputStreamReader(
                                       new FileInputStream( file ) ) );

        // first line of file must be the version string
        String fileVersion = input.readLine();

        // if line was null or versions not equal, throw an exception
        if ( (fileVersion == null) || (! version.trim().equals(fileVersion.trim())) )
        {
            input.close();
            throw new IniFileException("Version specified and version in file are not the same",
                                       IniFileException.VERSION_ERROR,
                                       fileVersion );
        }

        // now start reading the data
        try
        {
            String line;
            while ( (line=input.readLine()) != null )
            {
                // read a line
                lineNumber++;

                // get rid of leading and trailing whitespace
                line = line.trim();

                // skip lines containing only white space
                if ( line.equals( "" ) )
                    continue;

                // should now start seeing group names enclosed in [ ]
                if ( line.startsWith("[") && line.endsWith("]") )
                {
                    int begin   = line.indexOf('[') + 1;        // after the [
                    int end     = line.lastIndexOf(']');        // before the ]
                    String groupName = line.substring(begin,end);

                    // create a new group
                    addGroup(groupName);

                    // now parse out and key/value pairs in this group
                    parseKeyValuePairs( input, groupName );
                }
                // didn't see what we should have so cannot parse
                else
                {
                    throw new IniFileException("Could not parse file",
                                               IniFileException.PARSE_ERROR,
                                               lineNumber );
                }
            }
        }
        catch (IniFileException e)      // if it was a IniFileException, let it go
        {
            throw e;
        }
        catch (Exception e)             // if it was some other exception, create new
        {
            input.close();
            throw new IniFileException( "Could not parse file",
                                        IniFileException.PARSE_ERROR,
                                        lineNumber);
        }

        // close the stream
        input.close();
    }

    private void parseKeyValuePairs( BufferedReader input,
                                     String groupName ) throws Exception
    {
        String line = null;
        while ( (line=input.readLine()) != null )
        {
            // read another line
            lineNumber++;

            // get rid of leading and trailing whitespace
            line = line.trim();

            // a blank line designates the end of this group so return
            if ( line.equals( "" ) )
                return;

            // look for the equals sign which separates key/value pairs
            int equalIndex = line.indexOf('=');

            // if this couldn't be found, there is a parse error
            if ( equalIndex == -1 )
                throw new IniFileException( "Unable to parse file",
                                            IniFileException.PARSE_ERROR,
                                            lineNumber);

            // now seeing key/value pair
            String key      = line.substring(0,equalIndex).trim();
            String value    = line.substring(equalIndex+1).trim();

            addPair(groupName, key, value);
        }
    }


   /**
     * Write all groups and key/value pairs to the file specified in the constructor.
     * <p>If the file does not exist, a new ini file will be created with the
     * specified version string.
     * <p>NOTE: the old ini file is destroyed and a new file is written
     */
    public void writeToDisk() throws IniFileException
    {
        writeToDisk(filename);
    }


    /**
     * Write all groups and key/value pairs to the file specified in the constructor.
     * <p>If the file does not exist, a new ini file will be created with the
     * specified version string.
     * <p>NOTE: the old ini file is destroyed and a new file is written
     * @param filename overrides the filename used in the constructor
     */
    public void writeToDisk(String filename) throws IniFileException
    {
        try
        {
            file = new File( filename );

            // delete the file if it already exists
            if ( file.exists() && (!file.delete()) )
            {
                throw new IniFileException("Unable to delete existing file",
                                           IniFileException.READWRITE_ERROR);
            }

            FileOutputStream output = new FileOutputStream(file);

            output.write( toString().getBytes() );
        }
        catch (Exception e)
        {
            throw new IniFileException("Unable to write to file",
                                       IniFileException.READWRITE_ERROR);
        }

    }


    /**
     * Creates a new group of the specified name.
     * <p>If group already exists by that name, a new group is not created.
     * @param group the name for this group
     */
    public void addGroup(String group)
    {
        int index=groups.indexOf(new IniGroup(group));

        if ( index == -1 )
        {   // group does not exist
            groups.addElement( new IniGroup(group) );
        }
    }

    /*
     * Gets a <code>IniGroup</code> based on the group name.
     * <p>If the group does not exist, <code>null</code> is returned.
     * @returns the IniGroup if found
     */
    private IniGroup getIniGroup(String group)
    {
        int index=groups.indexOf(new IniGroup(group));

        if ( index == -1 )
        {   // group does not exist
            return null;
        }
        else
        {   // return existing group
            return (IniGroup) groups.elementAt( index );
        }
    }

    /**
     * Gets an array of all group names
     * @return an array of all group names
     */
    public String[] getGroups()
    {
        String[] arr = new String[groups.size()];
        int i=0;

        for (Enumeration e = groups.elements() ; e.hasMoreElements() ;)
        {
            arr[i] = ((IniGroup)e.nextElement()).getName();
            i++;
        }

        return arr;
    }

    /**
     * Remove the group.
     * <p>A non-existant group does not cause an error.
     * @param group the name of the group to remove
     */
    public void removeGroup(String group)
    {
        int index=groups.indexOf( new IniGroup(group) );

        if ( index != -1 )
        {   // group does exist
            groups.removeElementAt( index );
        }
    }

    /**
     * Remove a key value pair.
     * <p>A non-existant group or key does not cause an error.
     * @param group the name of the group to remove
     * @param key the name of the key used for the key/value pair
     */
    public void removePair(String group, String key)
    {
        int index=groups.indexOf( new IniGroup(group) );

        if ( index != -1 )
        {   // group does exist
            IniGroup g = getIniGroup( group );

            g.removePair( key );
        }
    }

    /**
     * Adds the key/value pair to the specified group.
     * <p>If the group does not exist, it is created.
     * <p>If the key does not exist within the group, it is created.
     * <p>If the key does exist within the group, replace the old value.
     * @param group the name of the group to add the key/value pair to.
     * @param key the key used to locate the key/value pair
     * @param value the value corresponding to <code>key</code> in the key/value pair.
     */
    public void addPair(String group, String key, String value)
    {
        // add the group if it does not exist
        addGroup( group );

        // now get the IniGroup of the group just created
        IniGroup g = getIniGroup( group );

        g.addPair( key, value );
    }

    /**
     * Gets an array of all keys within the specified group.
     * <p><code>null</code> is returned if group does not exist or if 0 keys
     * @param group the name of the group to retrieve the keys from
     */
    public String[] getKeys(String group)
    {
        int index=groups.indexOf( new IniGroup(group) );

        if ( index != -1 )
        {   // group does exist
            return ((IniGroup)groups.elementAt(index)).getKeys();
        }
        else
        {
            return null;
        }
    }

    /**
     * Gets an array of all values within the specified group.
     * <p><code>null</code> is returned if group does not exist or if 0 values
     * @param group the name of the group to retrieve the values from
     */
    public String[] getValues(String group)
    {
        int index=groups.indexOf( new IniGroup(group) );

        if ( index != -1 )
        {   // group does exist
            return ((IniGroup)groups.elementAt(index)).getValues();
        }
        else
        {
            return null;
        }
    }

    // returns null if group or key does not exist
    /**
     * Get the value of key/value pair within the specified group.
     * <p><code>null</code> is returned if group or key does not exist
     * @param group the name of the group to retrieve the value from
     * @param key the key for the key/value pair to retrieve
     * @return the value specified
     */
    public String getValue(String group, String key)
    {
        int index=groups.indexOf( new IniGroup(group) );

        if ( index != -1 )
        {   // group does exist
            return ((IniGroup)groups.elementAt(index)).getValue(key);
        }
        else
        {   // group does not exist
            return null;
        }
    }

    /**
     * Set the value of key/value pair within the specified group.
     * <p>The value is set if and only if the group and key exist.
     * @param group the name of the group that the key/value pair is in.
     * @param key the key of the key/value pair.
     * @param value the value to set in the key/value pair.
     */
    public void setValue(String group, String key, String value)
    {
        int index=groups.indexOf( new IniGroup(group) );

        if ( index != -1 )
        {   // group does exist
            ((IniGroup)groups.elementAt(index)).setValue(key,value);
        }
    }

    /**
     * Creates a string that can be used to write the groups (with key/value pairs) to a file.
     * <p>The string is formatted so the group name is between <code>[</code>
     * and <code>]</code> followed by all key/value pairs followed by a newline.
     * @return a string useful for initialization file creation
     */
    public String toString()
    {
        String str = version
            + System.getProperty("line.separator")
            + System.getProperty("line.separator");
        String[] gStrs = getGroups();
        int i=0;

        for (Enumeration e = groups.elements() ; e.hasMoreElements() ;)
        {
            str += "[" + gStrs[i] + "]";
            str += System.getProperty("line.separator");
            str += (IniGroup)e.nextElement();
            str += System.getProperty("line.separator");
            i++;
        }

        return str;
    }


    /**
     * Adds a key/value pair that uses a numeric postfix at the specified location.
     * <p>The actual key is split into 2 parts, a non-digit
     * prefix and a integral postfix. This is useful for keeping
     * an ordered list in the <code>IniFile</code>.
     * <br>Any existing key/value pairs at and after the specified position
     * have their index increased by 1.
     * <p><b>NOTE:</b> It is intended for the first index to be 1.
     * <p>It would look like this:
     * <br>[Group]<br>
     * <br>filename1=foo.dat
     * <br>filename2=ofo.dat
     * <br>filename3=oof.dat
     * <p>
     * @param group the group to add the key/value pair into
     * @param keyPrefix the alphanumeric part of the key.
     * @param value the value to set for the key <code>"keyPrefix"+position</code>.
     * @param position the integral index of the key.
     * @param removeOthers remove all other occurrences of this <code>value</code> after
     * the position index if true.
     */
    public void addIntegralPostfixPair(String group, String keyPrefix,
                                String value, int position, boolean removeOthers)
    {
        shiftIntegralPostfixPairDown( group, keyPrefix, position );

        addPair( group, keyPrefix + position, value );

        // now remove other occurrences of this key/value pair
        //  after the one added
        if (removeOthers)
            findIntegralPostfixValueAndRemove( group, keyPrefix, value, position + 1 );
    }

    /**
     * Adds a key/value pair to the end of the key prefixes.
     * <p>The actual key is split into 2 parts, a non-digit
     * prefix and a integral postfix. This is useful for keeping
     * an ordered list in the <code>IniFile</code>.
     * @param group the group to add the key/value pair into
     * @param keyPrefix the alphanumeric part of the key.
     * @param value the value to set for the key.
     */
    public void addIntegralPostfixPair(String group, String keyPrefix, String value)
    {
        int i = 1;
        String currentValue = getValue(group, keyPrefix + i);

        while ( currentValue != null )
        {
            i++;
            currentValue = getValue(group, keyPrefix + i);
        }

        addPair( group, keyPrefix + i, value );
    }

    /**
     * Remove all key/value pairs with a numeric postfix >= specified.
     * <p>The actual key is split into 2 parts, a non-digit
     * prefix and a integral postfix. All pairs (if consecutive) are
     * removed following the and including the index specified
     * <p>Behavior is undefined if <code>keyPrefix</code> ends in a digit
     * @param group the group to remove the key/value pair from.
     * @param keyPrefix the alphanumeric part of the key.
     * @param startRemoved the index of the key/value to begin removing at.
     */
    public void removeIntegralPostfixPairs(String group, String keyPrefix, int startRemoved)
    {
        int i = startRemoved;
        String curValue = getValue(group, keyPrefix + i);

        while ( curValue != null )
        {
            removePair(group, keyPrefix + i);

            i++;
            curValue = getValue(group, keyPrefix + i);
        }
    }

    /**
     * Remove the specified key and decrement the index of the keys following the one removed.
     * <p>Behavior is undefined if <code>keyPrefix</code> ends in a digit
     * @param group the group to remove the key/value pair from.
     * @param keyPrefix the alphanumeric part of the key.
     * @param removeIndex the postfixIndex of the key to be removed.
     */
    public void removeIntegralPostfixPair(String group, String keyPrefix, int removeIndex)
    {
        shiftIntegralPostfixValuesUp(group, keyPrefix, removeIndex);
    }

    /**
     * Get all 1-N values with the specified group and key prefix.
     * <p>If this key prefix does not exist within the group, or the group
     * does not exist, <code>null</code> is returned.
     * @param group the group to get the values from.
     * @param keyPrefix the alphanumeric part of the key.
     */
    public String[] getIntegralPostfixValues(String group, String keyPrefix)
    {
        int i = 1;
        String value = getValue(group, keyPrefix + i);
        Vector v = new Vector();

        // value is null if value did not exist
        if ( value == null )
            return null;

        while ( value != null )
        {
            v.addElement(value);

            i++;
            value = getValue(group, keyPrefix + i);
        }

        // now do an enumeration
        String[] arr = new String[v.size()];
        i=0;

        for (Enumeration e = v.elements() ; e.hasMoreElements() ;)
        {
            arr[i] = (String)e.nextElement();
            i++;
        }

        return arr;
    }


    // the actual key is in the file as "keyPrefix"+"i"
    // the original key/value pair at "i" is unaffected
    private void shiftIntegralPostfixPairDown(String group, String keyPrefix, int i)
    {
        String curValue = getValue(group, keyPrefix + i);
        String nextValue = null;

        while ( curValue != null )
        {
            nextValue = getValue(group, keyPrefix + (i+1));
            addPair(group, keyPrefix + (i+1), curValue);
            curValue = nextValue;

            i++;
        }
    }


    // the actual key is in the file as "keyPrefix"+"i"
    // the i-th key is overwritten by the (i+1)th key
    // the last key is removed
    private void shiftIntegralPostfixValuesUp(String group, String keyPrefix, int i)
    {
        // don't shift if it causes non-positive numbers
        if ( i < 1 )
            return;

        String curValue = getValue(group, keyPrefix + (i+1));

        while ( curValue != null )
        {
            setValue(group, keyPrefix + i, curValue);

            i++;
            curValue = getValue(group, keyPrefix + (i+1));
        }

        removePair(group, keyPrefix + (i));
    }

    // removes all pairs with the specified value and properly prefixed key
    private void findIntegralPostfixValueAndRemove(String group, String keyPrefix, String value, int i)
    {
        // index of 1 is first index
        if ( i < 1 )
            return;

        String curValue = getValue(group, keyPrefix + i);

        while ( curValue != null )
        {
            if ( getValue(group, keyPrefix + i).equals( value ) )
            {   // found a match
                shiftIntegralPostfixValuesUp(group, keyPrefix, i);
                curValue = getValue(group, keyPrefix + i );
            }
            else
            {
                i++;
                curValue = getValue(group, keyPrefix + (i));
            }

        }
    }

}

