
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

/**
 * A communications port.  CommPort is an abstract class that describes a
 * communications port made available by the underlying system.  It includes
 * high-level methods for controlling I/O that are common to different kinds
 * of communications ports.
 */
package com.dalsemi.comm;



import javax.comm.UnsupportedCommOperationException;
import java.io.*;


/**
 * This class implements a generic communications port.
 */
public abstract class InternalCommPort
{

   /**
    * Opens the port for communication.
    * @exception IOException
    */
   public abstract void open ()
      throws IOException;

   /**
    * Closes the port for communication.
    * @exception IOException
    */
   public abstract void close ()
      throws IOException;

   /**
    * Gets an InputStream for the port.
    *
    * @return InputStream for the port
    * @exception IOException
    */
   public abstract InputStream getInputStream ()
      throws IOException;

   /**
    * Gets an OutputStream for the port.
    *
    * @return OutputStream attached to this port
    * @exception IOException
    */
   public abstract OutputStream getOutputStream ()
      throws IOException;

   /**
    * Enables timeouts while receiving data.
    * @exception UnsupportedCommOperationException
    */
   public abstract void enableReceiveTimeout ()
      throws UnsupportedCommOperationException;

   /**
    * Disables timeouts while receiving data.
    * @exception UnsupportedCommOperationException
    */
   public abstract void disableReceiveTimeout ()
      throws UnsupportedCommOperationException;

   /**
    * Sets the Receive timeout value.
    *
    * @param timeout timeout value in milliseconds
    * @exception UnsupportedCommOperationException
    */
   public abstract void setReceiveTimeout (int timeout)
      throws UnsupportedCommOperationException;

   /**
    * Returns true if ReceiveTimeout is enabled
    */
   public abstract boolean isReceiveTimeoutEnabled ();

   /**
    * Gets the size of the Input Buffer.
    */
   public abstract int getInputBufferSize ();

   /**
    * Gets the size of the Output Buffer.
    */
   public abstract int getOutputBufferSize ();

   /**
    * Sets the size of the InputBuffer.
    * @param size desired new size for the input buffer
    * @exception UnsupportedCommOperationException
    */
   public abstract void setInputBufferSize (int size)
      throws UnsupportedCommOperationException;

   /**
    * Sets the size of the OutputBuffer.
    * @param size desired new size for the output buffer
    * @exception UnsupportedCommOperationException
    */
   public abstract void setOutputBufferSize (int size)
      throws UnsupportedCommOperationException;
}
