
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;



import com.dalsemi.system.*;
import java.io.*;
import javax.comm.UnsupportedCommOperationException;


/**
 * Serial Port implements a RS-232 serial communications port.  SerialPort describes the low-level
 * interface to the serial communications port.
 * To change the serial port baud rate do the following at the beginning of
 * your program.<BR>
 * ....<BR>
 * public static void main(String[] args)<BR>
 * {<BR>
 * <T>SerialPort sp = new SerialPort(SerialPort.SERIAL0, SerialPort.STDOUT);<BR>
 * <T>sp.setBaudRate(SerialPort.SPEED_300);<BR>
 * <BR>
 * }<BR>
 * NOTE: Changing the speed on any of the streams (stdin, stdout, stderr) causes
 * all streams to change to the speed.<BR>
 * ....<BR>
 *
 */
class InternalSerialPort
   extends InternalCommPort
{

   /** Serial port 0
     */
   public static final int SERIAL0 = 0;

   /** Serial port 1
     */
   public static final int SERIAL1 = 1;

   /** Serial port 0
     */
   public static final int SERIAL2 = 2;

   /** Serial port 1
     */
   public static final int SERIAL3 = 3;

   /** Command stream
     * @deprecated
     */
   public static final int COMMAND = 0;

   /** Standard in
     */
   public static final int STDIN = 1;

   /** Standard out
     */
   public static final int STDOUT = 2;

   /** Standard error
     */
   public static final int STDERR = 3;

   /** Generic stream
     *
     */
   public static final int GENERIC = 4;

   /** Baud Rate 300 bps
     */
   public static final int SPEED_300 = 1;

   /** Baud Rate 600 bps
     */
   public static final int SPEED_600 = 2;

   /** Baud Rate 1200 bps
     */
   public static final int SPEED_1200 = 3;

   /** Baud Rate 2400 bps
     */
   public static final int SPEED_2400 = 4;

   /** Baud Rate 4800 bps
     */
   public static final int SPEED_4800 = 5;

   /** Baud Rate 9600 bps
     */
   public static final int SPEED_9600 = 6;

   /** Baud Rate 19200 bps
     */
   public static final int SPEED_19200 = 7;

   /** Baud Rate 38400 bps
     */
   public static final int SPEED_38400 = 8;

   /** Baud Rate 57600 bps
     */
   public static final int SPEED_57600 = 9;

   /** Baud Rate 115200 bps
     */
   public static final int SPEED_115200 = 10;

   private static final int SPEED_230400 = 11;
   /** Default serial port read timeout
     */
   public static final int SERIAL_DEFAULT_READ_TIMEOUT = 0x3FFFFFFF;
   private int             SerialHandle                = 0;
   private int             PortType                    = 0;
   private int             StreamType                  = 0;
   private int             baudRate                    = SPEED_115200;
   private boolean         IsOpen                      = false;
   private int             SerialReadTimeout           =
      SERIAL_DEFAULT_READ_TIMEOUT;
   private boolean         receiveTimeout              = false;
   private byte[]          SerialPortWriteArray;
   private byte[]          SerialPortReadArray;
   private int             SERIAL_PORT_ARRAY_LENGTH      = 1024;
   private int             SERIAL_PORT_READ_ARRAY_LENGTH = 1024;

   /**
    * Method declaration
    *
    *
    * @param baudRateSpecifier static final int specifying the desired baud rate
    *
    * @see
    */
   private static native void SetSerBaud (int portNum, int baudRateSpecifier);

   /**
    * Constructor declaration
    *
    *
    * @see
    */

   /* private */
   InternalSerialPort ()
   {
   }

   /**
    * Constructor declaration
    *
    *
    * @param portnum static final int specifying port number
    *
    * @see
    */

   /* private */
   InternalSerialPort (int portnum)
   {
   }

   /**
    * Constructs a new serial port object.
    *
    *
    * @param portnum - the serial port number that this should be associated with.
    * @param stream - the stream to open.
    * @exception IOException
    *
    * @see
    */
   public InternalSerialPort (int portnum, int stream)
      throws IOException
   {
      switch (portnum)
      {

         case SERIAL0 :
            PortType = NativeComm.PORT_SERIAL0;
            break;

         case SERIAL1 :
            PortType = NativeComm.PORT_SERIAL1;
            break;

         case SERIAL2 :
            PortType = NativeComm.PORT_SERIAL2;
            break;

         case SERIAL3 :
            PortType = NativeComm.PORT_SERIAL3;
            break;

         default :
            throw new IOException("Invalid Serial Port Number");
      }

      if (stream == STDIN)
         StreamType = NativeComm.STREAM_STDIN;
      else if (stream == STDOUT)
         StreamType = NativeComm.STREAM_STDOUT;
      else if (stream == STDERR)
         StreamType = NativeComm.STREAM_STDERR;
      else
         throw new IOException("Invalid Serial Stream Type");

      SerialPortReadArray  = new byte [SERIAL_PORT_READ_ARRAY_LENGTH];
      SerialPortWriteArray = new byte [SERIAL_PORT_ARRAY_LENGTH];
   }

   /**
    * Opens the serial port.  This should not be called directly.
    * In the future, you will be able to use this method, but for now it is broken.
    * The port is opened by the input/otuput stream classes.
    */
   public void open ()
   {
      if (!IsOpen)
      {
         SerialHandle = NativeComm.open(PortType, StreamType);

         NativeComm.setHandle(SerialHandle);

         IsOpen = true;
      }
   }

   /**
    * Close the serial port. This should not be called directly.
    * In the future, you will be able to use this method, but for now it is broken.
    * The port is closed by the input/otuput stream classes.
    * @exception IOException
    */
   public void close ()
      throws IOException
   {
      if (IsOpen)
      {

         //com.dalsemi.system.Debug.debugDump("Calling NativeComm.close");
         NativeComm.close(SerialHandle);

         //com.dalsemi.system.Debug.debugDump("SETTING ISOPEN = FALSE");
         IsOpen = false;
      }
   }

   /**
    * Writes the entire byte array <code>arr</code> to the serial port.
    * @param arr the array to be written to the serial port.
    * @exception IOException
    */
   public void write (byte[] arr)
      throws IOException
   {
      if (IsOpen)
      {

         // com.dalsemi.jellybeans.Debug.debugDump(0x5b);
         NativeComm.write(SerialHandle, arr);
      }
      else
         throw new IOException("Could not write to serial port");
   }

   /**
    * Writes <code>len</code> bytes of array <code>arr</code> from <code>offset</code> to the serial port.
    * @param arr     the array to be written.
    * @param offset  the offset in the array to begin sending.
    * @param len     the number of bytes to send.
    * @exception   IOException
    */
   public void write (byte[] arr, int offset, int len)
      throws IOException
   {

      /*  int bytesleft;
        int curoffset;
        int sendlength;
*/
      if (IsOpen)
      {

         // com.dalsemi.jellybeans.Debug.debugDump(0x5c);
         // Native doesn't handle offsets other than 0.
         // So we need to check for this and format it correctly
         // in a new buffer.
         //if (offset == 0)
         //    NativeComm.write(SerialHandle, arr, len);
         //else
         {

            // Copy the array to a serial buffer, and send each
            // one individually.

            /* bytesleft = len;
             curoffset = offset;

             while (bytesleft > 0)
             {
                 if (bytesleft > SERIAL_PORT_ARRAY_LENGTH)
                     sendlength = SERIAL_PORT_ARRAY_LENGTH;
                 else
                     sendlength = bytesleft;
                 synchronized (SerialPortWriteArray)
                 {
                    System.arraycopy(arr, curoffset, SerialPortWriteArray, 0,
                                     sendlength);
                    NativeComm.write(SerialHandle, SerialPortWriteArray,
                                     sendlength);
                 }
                 bytesleft -= sendlength;
                 curoffset += sendlength;

             }
             */
            NativeComm.write(SerialHandle, arr, offset, len);
         }
      }
      else
         throw new IOException("Could not write to serial port");
   }

   /**
    * Writes the integer <code>value</code> to the serial port.
    * @param value the integer to be written.
    * @exception IOException
    */
   public void write (int value)
      throws IOException
   {
      if (IsOpen)
      {
         synchronized (SerialPortWriteArray)
         {
            SerialPortWriteArray [0] = ( byte ) value;

            // com.dalsemi.jellybeans.Debug.debugDump(0x5d);
            NativeComm.write(SerialHandle, SerialPortWriteArray, 0, 1);

            // com.dalsemi.jellybeans.Debug.debugDump(0x5e);
         }
      }
      else
         throw new IOException("Could not write to serial port");
   }

   /**
    * Reads <code>len</code> bytes of array <code>arr</code> from <code>offset</code> to the serial port.
    * @param arr     the array for the data read to be placed in.
    * @param offset  the offset to begin saving data.
    * @param len     the number of bytes to read.
    *
    * @return
    * @exception IOException
    */
   public int read (byte[] arr, int offset, int len)
      throws IOException
   {
      int bytesleft;
      int curoffset;
      int sendlength;
      int retval;
      int numRequested = len;

      if (IsOpen)
      {

         /*if (offset != 0)
            com.dalsemi.system.Debug.debugDump("OFFSET NOT EQUAL TO ZERO");
         */
         retval = NativeComm.read(SerialHandle, arr, offset, len,
                                  SerialReadTimeout, receiveTimeout);

         return retval;

         /*
                    // Copy the serial buffer to the array
                    bytesleft = numRequested;
                    curoffset = offset;


                    while (bytesleft > 0)
                    {
                       if (bytesleft > SERIAL_PORT_READ_ARRAY_LENGTH)
                          sendlength = SERIAL_PORT_READ_ARRAY_LENGTH;
                       else
                          sendlength = bytesleft;

                       synchronized (SerialPortReadArray)
                       {
         //                 com.dalsemi.system.Debug.debugDump(0x62);
                          retval = NativeComm.read(SerialHandle,
                                                   SerialPortReadArray, sendlength,
                                                   SerialReadTimeout);

         //                 com.dalsemi.system.Debug.debugDump(0x63);
                          if (retval == 0)
                             return (curoffset - offset);

                          System.arraycopy(SerialPortReadArray, 0, arr, curoffset,
                                           retval);

                          // com.dalsemi.jellybeans.Debug.debugDump(0x64);

                          bytesleft -= retval;
                          curoffset += retval;
                       }
                    }

         */
      }
      else
         throw new IOException("Could not read from serial port");

      //return numRequested;
   }

   /**
    * Reads an integer from the serial port.
    * @return the byte read.
    * @exception IOException
    */
   public int read ()
      throws IOException
   {
      int retval;

      if (IsOpen)
      {
         synchronized (SerialPortReadArray)
         {
            retval = NativeComm.read(SerialHandle, SerialPortReadArray, 0, 1,
                                     SerialReadTimeout, receiveTimeout);

            // com.dalsemi.jellybeans.Debug.debugDump(0x65);
            if (retval == 0)
            {

               // com.dalsemi.jellybeans.Debug.debugDump(0x66);
               return -1;
            }
            else
            {

               // com.dalsemi.jellybeans.Debug.debugDump(0x67);
               retval = SerialPortReadArray [0] & 0xFF;

               // com.dalsemi.jellybeans.Debug.debugDump(0x99);
               return retval;

               // return (SerialPortReadArray[0] & 0xFF);
            }
         }
      }
      else
         throw new IOException("Could not read from serial port");
   }

   /**
    * Gets the serial port input stream
    * @return InputStream
    * @exception throws IOException
    */
   public InputStream getInputStream ()
      throws IOException
   {
      open();

      return new SerialInputStream(this);
   }

   /**
    * Gets the serial port output stream
    * @return OutputStream
    * @exception throws IOException
    */
   public OutputStream getOutputStream ()
      throws IOException
   {
      open();

      return new SerialOutputStream(this);
   }

   /**
    * enables the receive timeout - currently not implemented.
    * @exception UnsupportedCommOperationException
    */
   public void enableReceiveTimeout ()
      throws UnsupportedCommOperationException
   {
      receiveTimeout = true;
   }

   /**
    * disables the receive timeout. - currently not implemented.
    * Receive timeout is alway on.
    * @exception UnsupportedCommOperationException
    */
   public void disableReceiveTimeout ()
   {
      receiveTimeout    = false;
      SerialReadTimeout = SERIAL_DEFAULT_READ_TIMEOUT;   // this is basically forever or as close as I'm going to get.

      return;
   }

   /**
    * sets the receive timeout to the given value.
    * @param timeout the value to delay before timeing out.
    * @exception UnsupportedCommOperationException
    */
   public void setReceiveTimeout (int timeout)
      throws UnsupportedCommOperationException
   {
      if (timeout < 0)
         throw new UnsupportedCommOperationException("Timeout < 0");

      SerialReadTimeout = timeout;
   }

   /**
    * Returns true if the receive timeout is enabled.
    * currently always returns true.
    * @return boolean true if receive timeout is enabled
    */
   public boolean isReceiveTimeoutEnabled ()
   {
      return receiveTimeout;
   }

   /**
    * Gets the size of the input buffer. <I>This method
    * currently returns 1 which is not necessarily the correct size.</I>
    * @return the size of the input buffer.
    */
   public int getInputBufferSize ()
   {
      return 1;
   }

   /**
    * Gets the size of the output buffer. <I>This method
    * currently returns 1 which is not necessarily the correct size.</I>
    * @return the size of the output buffer.
    */
   public int getOutputBufferSize ()
   {
      return 1;
   }

   /**
    * Sets the size of the input buffer.
    * @param size the size of the input buffer.
    * @exception UnsupportedCommOperationException thrown if
    * the buffer cannot be resized.
    */
   public void setInputBufferSize (int size)
      throws UnsupportedCommOperationException
   {
      throw new UnsupportedCommOperationException();
   }

   /**
    * Sets the size of the output buffer.
    * @param size the size of the output buffer.
    * @exception UnsupportedCommOperationException thrown if
    * the output buffer cannot be resized.
    */
   public void setOutputBufferSize (int size)
      throws UnsupportedCommOperationException
   {
      throw new UnsupportedCommOperationException();
   }

   /**
    * Native method
    *
    * @param speed
    *
    * @throws UnsupportedCommOperationException
    *
    * @see
    */

   //  private native void setBaudRateNative(int speed)
   //      throws UnsupportedCommOperationException;

   /**
    * Sets the baud rate for the port.
    *
    * @param speed use the static final ints
    *
    * @throws UnsupportedCommOperationException
    *
    * @see
    */
   public void setBaudRate (int speed)
      throws UnsupportedCommOperationException
   {
      int tiniBaudRate = InternalSerialPort.SPEED_115200;

      switch (speed)
      {

         case 300 :
            tiniBaudRate = InternalSerialPort.SPEED_300;
            break;

         case 600 :
            tiniBaudRate = InternalSerialPort.SPEED_600;
            break;

         case 1200 :
            tiniBaudRate = InternalSerialPort.SPEED_1200;
            break;

         case 2400 :
            tiniBaudRate = InternalSerialPort.SPEED_2400;
            break;

         case 4800 :
            tiniBaudRate = InternalSerialPort.SPEED_4800;
            break;

         case 9600 :
            tiniBaudRate = InternalSerialPort.SPEED_9600;
            break;

         case 19200 :
            tiniBaudRate = InternalSerialPort.SPEED_19200;
            break;

         case 38400 :
            tiniBaudRate = InternalSerialPort.SPEED_38400;
            break;

         case 57600 :
            tiniBaudRate = InternalSerialPort.SPEED_57600;
            break;

         case 115200 :
            tiniBaudRate = InternalSerialPort.SPEED_115200;
            break;

         case 230400 :
            tiniBaudRate = InternalSerialPort.SPEED_230400;
            break;

         default :
            throw new UnsupportedCommOperationException(
               "unsupported baud rate");
      }

      switch (PortType)
      {

         case NativeComm.PORT_SERIAL0 :
            if ((tiniBaudRate >= SPEED_300) && (tiniBaudRate <= SPEED_115200))
            {
               SetSerBaud(NativeComm.PORT_SERIAL0, tiniBaudRate - 1);

               baudRate = tiniBaudRate;

               break;
            }
            else
            {
               throw new UnsupportedCommOperationException(
                  "Invalid baud rate");
            }
         case NativeComm.PORT_SERIAL1 :
            if ((tiniBaudRate >= SPEED_2400)
                    && (tiniBaudRate <= SPEED_115200))
            {

               //com.dalsemi.system.Debug.debugDump("Baud = ");
               //com.dalsemi.system.Debug.debugDump(tiniBaudRate);
               SetSerBaud(NativeComm.PORT_SERIAL1, tiniBaudRate - 1);

               baudRate = tiniBaudRate;

               break;
            }
            else
            {
               throw new UnsupportedCommOperationException(
                  "Invalid baud rate");
            }
         default :
         {
//            Debug.debugDump("INVALID PORT TYPE IN INTERNAL SETBAUDRATE");
//            Debug.debugDump(PortType);

            throw new UnsupportedCommOperationException();
         }
      }
   }

   /**
    * Method getBaudRate
    *
    *
    * @return
    *
    */
   public int getBaudRate ()
   {
      return baudRate;
   }

   /*
    * Sends a break on the serial port.
    *
    * @param timeMS the duration of the break.
    *
    * @see
    */
/*   public void sendBreak (int timeMS)
   {
      sendBreakNative(timeMS);
   }
*/

   /**
    * Returns the number of bytes available on the port.
    *
    * @return the number of bytes available.
    *
    * @see
    */
   public int available ()
   {
      return NativeComm.available(SerialHandle);
   }

   /**
    * Method setNumberBits
    *
    *
    * @param numbits
    *
    * @throws UnsupportedCommOperationException
    *
    */
   public void setNumberBits (int numbits)
      throws UnsupportedCommOperationException
   {
      if (numbits < 7 || numbits > 9)
      {
         throw new UnsupportedCommOperationException();
      }

      setNumberBitsNative(numbits);
   }

   private static native void setNumberBitsNative (int numbits);
}
