
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.comm;



import java.io.*;
import javax.comm.*;


/**
 * A Serial input stream is an input stream for reading data from an serial port.
 */
public class SerialInputStream
   extends java.io.InputStream
{
   InternalSerialPort SerialPortObject = null;
   private byte[]     pushbackByte     = new byte [1];
   private byte[]     readInt          = new byte [1];
   private boolean    havePushbackByte;
   private int        threshold        = 1;
   private boolean    thresholdEnabled = false;

   /**
    * Default constructor
    *
    * @see
    */
   private SerialInputStream ()
   {
   }

   /**
    * Constructor that specifies a particular port.
    *
    * @param sp
    *
    * @see
    */
   SerialInputStream (InternalSerialPort sp)
   {
      SerialPortObject = sp;
   }

   /**
    * Opens the Serial input stream for reading.
    *
    * @exception  IOException  if an I/O error occurs.
    */
   private void open ()
      throws IOException
   {
      SerialPortObject.open();
   }

   /**
    * Reads the next byte of data from this input stream. The value
    * byte is returned as an <code>int</code> in the range
    * <code>0</code> to <code>255</code>. If no byte is available
    * because the end of the stream has been reached, the value
    * <code>-1</code> is returned. This method blocks until input data
    * is available, the end of the stream is detected, or an exception
    * is thrown.
    * <p>
    *
    * @return     the next byte of data, or <code>-1</code> if the end of the
    * stream is reached.
    * @exception  IOException  if an I/O error occurs.
    */
   public synchronized int read ()
      throws IOException
   {

      // com.dalsemi.jellybeans.Debug.debugDump(0x77);
      //int retval = (SerialPortObject.read());
      // com.dalsemi.jellybeans.Debug.debugDump(0x78);
      if (thresholdEnabled && threshold == 0)
         return -1;

      return (read(readInt, 0, 1) == -1) ? -1
                                         : (readInt [0] & 0x0ff);
   }

   private int tread (byte[] barr, int offset, int len)
      throws IOException
   {
      int retval         = 0, bytesRead = 0, available = 0;
      int localThreshold = threshold;

      if (len < localThreshold)
         localThreshold = len;

      synchronized (this)
      {
         if (havePushbackByte)
         {
            barr [offset++]  = pushbackByte [0];
            havePushbackByte = false;

            if (localThreshold == 1)
               return 1;

            bytesRead = 1;

            localThreshold--;
         }
      }

      available = SerialPortObject.available();

      /* read as much as we possibly can up to len */
      if (available > localThreshold && available <= len)
         localThreshold = available;

      // do the read.
      bytesRead += (SerialPortObject.read(barr, offset, localThreshold));

      return bytesRead;
   }

   /**
    * Reads the number of bytes requested from this input stream.
    *
    * If no byte is available
    * because the end of the stream has been reached, the value
    * <code>-1</code> is returned. This method blocks until input data
    * is available, the end of the stream is detected, or an exception
    * is thrown.
    * <p>
    *
    * @param barr  the array to be filled.
    * @param offset  the offset to begin writting in the array.
    * @param len  the number of bytes requested.
    *
    * @return  the number of bytes read or <code>-1</code> if the end of the stream
    * has been reached.
    *
    * @throws IOException
    *
    * @see
    */
   public synchronized int read (byte[] barr, int offset, int len)
      throws IOException
   {
      int retval = 0, singleBytesRead = 0, available = available();

      if (threshold == 0 && thresholdEnabled)
         return 0;

      if (thresholdEnabled)
         return tread(barr, offset, len);

      synchronized (this)
      {
         if (havePushbackByte)
         {
            barr [offset++]  = pushbackByte [0];
            havePushbackByte = false;

            if ((len == 1) || (available == 1))
               return 1;

            singleBytesRead = 1;

            available--;
         }
      }

      // don't want to block forever anymore.  // Only read what is available (or up to threshold)
      if (available == 0)
      {
         singleBytesRead = (SerialPortObject.read(barr, offset++, 1));

         len--;

         available = SerialPortObject.available();

         if ((singleBytesRead == -1) || (available == 0) || (len == 0))
            return singleBytesRead;
      }

      if (available > len)
         available = len;

      if (singleBytesRead == 1)
         available -= 1;

      // com.dalsemi.jellybeans.Debug.debugDump(0x79);
      retval = (SerialPortObject.read(barr, offset, available));

      if (singleBytesRead == 1)   // we had a valid byte from the first read
      {
         if (retval == -1)
            retval = 1;
         else
            retval++;
      }

      // com.dalsemi.jellybeans.Debug.debugDump(0x7a);
      return retval;
   }

   /**
    * Method unread
    *
    *
    * @param b
    *
    * @throws IOException
    *
    */
   public synchronized void unread (byte b)
      throws IOException
   {
      synchronized (this)
      {
         if (havePushbackByte)
            throw new IOException();

         pushbackByte [0] = b;
         havePushbackByte = true;
      }
   }

   /**
    * Closes this Serial input stream and releases any system resources
    * associated with the stream.
    *
    * @exception  IOException  if an I/O error occurs.
    */
   public void close ()
      throws IOException
   {
      synchronized (this)
      {
         havePushbackByte = false;
      }

      SerialPortObject.close();
   }

   /**
    * Returns the number of bytes that can be read from this input
    * stream without blocking. The available method of
    * <code>InputStream</code> returns <code>0</code>. This method
    * <B>should</B> be overridden by subclasses.
    *
    * @return     the number of bytes that can be read from this input stream
    * without blocking.
    * @exception  IOException  if an I/O error occurs.
    */
   public int available ()
      throws IOException
   {
      int bytesAvailable = SerialPortObject.available();

      synchronized (this)
      {
         if (havePushbackByte)
            bytesAvailable++;

         return bytesAvailable;
      }
   }

   protected void enableReceiveThreshold (int thresh)
      throws UnsupportedCommOperationException
   {
      if (thresh >= 0)
      {
         threshold        = thresh;
         thresholdEnabled = true;
      }
      else
         throw new UnsupportedCommOperationException("Threshold < 0");
   }

   protected void disableReceiveThreshold ()
   {
      threshold        = 1;
      thresholdEnabled = false;
   }

   protected boolean isReceiveThresholdEnabled ()
   {
      return thresholdEnabled;
   }
}
