
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.io;



import com.dalsemi.system.ArrayUtils;
import java.io.CharConversionException;


/**
 *
 */
public class ByteToCharUTF8
   extends ByteToCharConverter
{

   //This buffer needs to be protected from multiple threads
   //trying to convert at once. ->  The InputStreamReader takes
   //care of this
   private byte[]  buffer        = new byte [2];
   private int     count         = 0;
   private boolean errorDetected = false;

   /**
    *
    */
   public int getMaxCharCount (byte[] forThis, int start, int end)
   {

      //return end - start;
      //this function needs to return how many unicode characters there are here
      //what we do is parse it, every byte that starts with :
      //    0xxxxxxx, 110xxxxx, 1110xxxx
      //is the mark of a new unicode character, and we need to make sure that
      //the end doesn't chop the middle of a character
      int count = 0;

      for (int i = start; i < end; i++)
      {
         if ((forThis [i] & 0x80) == 0x00)        //if this is a 0xxxxxxx byte
            count++;
         else if ((forThis [i] & 0xe0) == 0xc0)   //if this is a 110xxxx byte
         {
            i++;

            if (i < end)
               count++;
         }
         else if ((forThis [i] & 0xf0) == 0xe0)   //if this is a 1110xxxx byte
         {
            i++;
            i++;

            if (i < end)
               count++;
         }
      }

      return count;
   }

   /**
    *  KLA - removed throwing of exception, sun does not throw one.  also added error detection
    *  for multiple reads.
    */
   public int convert (byte[] src, int srcStart, int srcEnd, char[] dst,
                       int dstStart, int dstEnd)
   {
      int pos = dstStart;

      if (errorDetected)   //must flush to get rid of this error
         return 0;

      if (pos == dstEnd)   //if we have the number of bytes the caller wanted
         return (pos - dstStart);

      for (int i = srcStart; i < srcEnd; i++)
      {
         int ch;

         if (count == 0)
         {
            ch = src [i] & 0x0FF;
         }
         else
         {
            ch = buffer [0] & 0x0FF;

            i--;
         }

         if ((ch & 0x80) == 0)
         {

            //Single byte encoded
            dst [pos++] = ( char ) ch;
         }
         else
         {
            int tmp = ch >> 4;

            if (tmp == 0x0C || tmp == 0x0D)
            {

               //2 byte encoded
               if ((i + 1) < srcEnd)
               {
                  tmp = src [++i];
               }
               else
               {

                  //Not enough data to continue.
                  //Store in buffer...
                  count            = 0;
                  buffer [count++] = ( byte ) ch;

                  return (pos - dstStart);
               }

               count = 0;

               if ((tmp & 0xC0) != 0x80)
               {

                  //2-byte encoding failure found
                  errorDetected = true;

                  return (pos - dstStart);
               }
               else
               {
                  dst [pos++] = ( char ) (((ch & 0x1F) << 6) | (tmp & 0x3F));
               }
            }
            else if (tmp == 0x0E)
            {

               //3 byte encoded
               if (count <= 1)
               {
                  if ((i + 1) < srcEnd)
                     tmp = src [++i];
                  else
                  {

                     //Not enough data to continue.
                     //Store in buffer...
                     buffer [count++] = ( byte ) ch;

                     return (pos - dstStart);
                  }
               }
               else
               {
                  tmp   = buffer [1];
                  count = 0;
               }

               if ((tmp & 0xC0) != 0x80)
               {

                  //3-byte encoding failure found
                  errorDetected = true;

                  return (pos - dstStart);
               }
               else
               {
                  buffer [0] = ( byte ) ch;
                  ch         = ( char ) (((ch & 0x1F) << 6) | (tmp & 0x3F));

                  if ((i + 1) < srcEnd)
                     tmp = src [++i];
                  else
                  {

                     //Not enough data to continue.
                     //Store in buffer...
                     buffer [1] = ( byte ) tmp;
                     count      = 2;

                     return (pos - dstStart);
                  }

                  if ((tmp & 0xC0) != 0x80)
                  {

                     //3-byte encoding failure found
                     errorDetected = true;

                     return (pos - dstStart);
                  }
                  else
                  {
                     dst [pos++] = ( char ) ((ch << 6) | (tmp & 0x3F));
                  }
               }
            }
            else
            {

               //encoding failure found, this was expected to be a first byte
               //but it wasn't correctly formatted
               errorDetected = true;

               return (pos - dstStart);
            }
         }
      }

      return (pos - dstStart);
   }

   /**
    *
    */
   public int flush (char[] buff, int start, int length)
      throws CharConversionException
   {
      count         = 0;
      errorDetected = false;

      if (length == 0)
         return 0;

      throw new CharConversionException("UTF8 Cannot convert more data");
   }

   /**
    *
    */
   public String getName ()
   {
      return "UTF8";
   }
}
