
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.onewire.utils;

// imports
import java.util.Vector;
import java.util.Enumeration;
import com.dalsemi.onewire.OneWireException;
import com.dalsemi.onewire.adapter.OneWireIOException;
import com.dalsemi.onewire.container.OneWireContainer;
import com.dalsemi.onewire.container.PagedMemoryBank;
import com.dalsemi.onewire.container.MemoryBank;
import com.dalsemi.onewire.container.OTPMemoryBank;


/**
 * Class to provide read/write cache services to a 1-Wire memory
 * device.  Writes are only performed when this classes <code>sync()</code>
 * method is called.  Provides page bitmap services for OTP devices.
 *
 * <p>Objectives:
 * <ul>
 * <li> Cache read/written pages
 * <li> write only on sync()
 * <li> write order is oldest to newest.
 * <li> Collect redirection information when appropriate
 * </ul>
 *
 * <p>TODO:
 * <ul>
 * <li> OTP devices not yet support for writing
 * <li> auto redirection to new pages if page occupied
 * <li> array of redirection jobs todo
 * <li> Constructor with array of memorybanks (multi-device filesystem)
 * </ul>
 *
 * @author  DS
 * @version 0.00, 20 November 2000
 * @see     com.dalsemi.onewire.utils.OWFile
 * @see     com.dalsemi.onewire.utils.OWFileDescriptor
 * @see     com.dalsemi.onewire.utils.OWFileInputStream
 * @see     com.dalsemi.onewire.utils.OWFileOutputStream
 */
class MemoryCache
{

   //--------
   //-------- Static Final Variables
   //--------

   /** Field EMPTY - flag to indicate cache page is empty */
   private static final int EMPTY = -1;

   /** Field NONE - flag to indicate last page read is not known */
   private static final int NONE = -100;

   /** Field USED - flag to indicate page bitmap file used */
   private static final int USED = 0;

   /** Field NOT_USED - flag to indicated page bitmap file un-used */
   private static final int NOT_USED = 1;

   /** Enable/disable debug messages */
   private static final boolean doDebugMessages = false;

   //--------
   //-------- Variables
   //--------

   /** Field owd - 1-Wire container that containes this memory to cache */
   private OneWireContainer owd;

   /** Field cache -  2 dimentional array to contain the cache */
   private byte[][] cache;

   /** Field len - array of lengths of packets found */
   private int[] len;

   /** Field changed - array of flags to indicate the page has been changed */
   private boolean[] changed;

   /** Field banks - vector of memory banks that contain the file system */
   private Vector banks;

   /** Field totalPages - total pages in this file system */
   private int totalPages;

   /** Field lastPageRead - last page read by this cache */
   private int lastPageRead;

   /** Field maxPacketDataLength - maximum data length on a page */
   private int maxPacketDataLength;

   /** Field bankPages - array of the number of pages in vector of memory banks */
   private int[] bankPages;

   /** Field writeLog - array to track the order of pages written to the cache */
   private int[] writeLog;

   /** Field tempExtra - temporary buffer used to to read the extra information from a page read */
   private byte[] tempExtra;

   /** Field tempPage - temporary buffer the size of a page */
   private byte[] tempPage;

   /** Field redirect - array of redirection bytes */
   private int[] redirect;

   /** Field owners - vector of classes that are using this cache */
   private Vector owners;

   /** Field openedToWrite - vector of files that have been opened to write on this filesystem */
   private Vector openedToWrite;

   /** Field getRedirection - flag to indicate page redirection information must be gathered */
   private boolean getRedirection;

   /** Field pbmBank - memory bank used for the page bitmap */
   private OTPMemoryBank pbmBank;

   /** Field pbmByteOffset - byte offset into page bitmap buffer */
   private int pbmByteOffset;

   /** Field pbmBitOffset - bit offset into page bitmap buffer  */
   private int pbmBitOffset;

   /** Field pbmCache - buffer to cache the page bitmap */
   private byte[] pbmCache;

   /** Field pbmCacheModified - modifified version of the page bitmap */
   private byte[] pbmCacheModified;

   /** Field pbmRead - flag indicating that the page bitmap has been read */
   private boolean pbmRead;

   /** Field lastFreePage - last free page found in the page bitmap */
   private int lastFreePage;

   //--------
   //-------- Constructor
   //--------

   /**
    * Construct a new memory cache for provided 1-wire container device.
    *
    * @param device 1-Wire container 
    */
   public MemoryCache(OneWireContainer device)
   {
      owd = device;

      PagedMemoryBank pmb = null;

      banks         = new Vector(1);
      owners        = new Vector(1);
      openedToWrite = new Vector(1);

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___Constructor MemoryCache: "
                            + device.getAddressAsString());

      // enumerate through the memory banks and collect the
      // general purpose banks in a vector 
      for (Enumeration bank_enum = owd.getMemoryBanks();
              bank_enum.hasMoreElements(); )
      {

         // get the next memory bank
         MemoryBank mb = (MemoryBank) bank_enum.nextElement();

         // look for pbm memory bank (used in file structure)
         if (mb.isWriteOnce() &&!mb.isGeneralPurposeMemory()
                 && mb.isNonVolatile() && (mb instanceof OTPMemoryBank))
         {

            // If only 4 pages then have DS2502 or DS2406 which have bitmap included
            // in the only status page.  All other EPROM devices have a special memory
            // bank that has 'Bitmap' in the title. 
            if ((totalPages == 4)
                    || (mb.getBankDescription().indexOf("Bitmap") != -1))
            {
               pbmBank = (OTPMemoryBank) mb;

               if (totalPages == 4)
                  pbmBitOffset = 4;

               pbmByteOffset  = 0;
               getRedirection = true;

               //\\//\\//\\//\\//\\//\\//\\//
               if (doDebugMessages)
                  System.out.println("_Paged BitMap MemoryBank: "
                                     + mb.getBankDescription()
                                     + " with bit offset " + pbmBitOffset);
            }
         }

         // check regular memory bank 
         if (!mb.isGeneralPurposeMemory() ||!mb.isNonVolatile()
                 ||!(mb instanceof PagedMemoryBank))
            continue;

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
            System.out.println("_Using MemoryBank: "
                               + mb.getBankDescription());

         banks.addElement(mb);
      }

      // count total bankPages
      bankPages  = new int [banks.size()];
      totalPages = 0;

      for (int b = 0; b < banks.size(); b++)
      {
         pmb           = (PagedMemoryBank) banks.elementAt(b);
         bankPages [b] = pmb.getNumberPages();
         totalPages    += bankPages [b];
      }

      // create the cache   
      cache               = new byte [totalPages][pmb.getPageLength()];
      len                 = new int [totalPages];
      changed             = new boolean [totalPages];
      writeLog            = new int [totalPages];
      maxPacketDataLength = pmb.getMaxPacketDataLength();
      tempPage            = new byte [pmb.getPageLength()];
      redirect            = new int [totalPages];

      // initialize some of the flag arrays 
      for (int p = 0; p < totalPages; p++)
      {
         changed [p]  = false;
         len [p]      = EMPTY;
         writeLog [p] = EMPTY;
      }

      // if getting redirection information, create necessarey arrays
      if (getRedirection)
      {
         tempExtra        = new byte [pmb.getExtraInfoLength()];
         pbmCache         = new byte [pbmBank.getSize()];
         pbmCacheModified = new byte [pbmBank.getSize()];
         pbmRead          = false;
      }
      else
         pbmRead = true;

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("_Total Pages: " + totalPages
                            + ", get Redirection = " + getRedirection);
   }

   /**
    * Gets the number of pages in this cache
    *
    * @return number of pages in the cache
    */
   public int getNumberPages()
   {
      return totalPages;
   }

   /**
    * Gets the maximum number of bytes for data in each page.
    *
    * @return max number of data bytes per page
    */
   public int getMaxPacketDataLength()
   {
      return maxPacketDataLength;
   }

   /**
    * Check if this memory device is write-once.  If this is true then
    * the page bitmap facilities in this class will be used.
    *
    * @return true if this device is write-once
    */
   public boolean isWriteOnce()
   {
      return getRedirection;
   }

   /**
    * Check if the page specified is redirected. 
    *
    * @param page to check
    *
    * @return page that this page is redirected to or
    *        0 if the page is not redirected 
    */
   public int isPageRedirected(int page)
   {
      if ((redirect == null) || (len [page] == EMPTY))
      {

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
            System.out.println("___isPageRedirected (" + page
                               + "), redirect null or page not read");

         return 0;
      }
      else
      {

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
            System.out.println("___isPageRedirected (" + page + ") = "
                               + redirect [page]);

         return redirect [page];
      }
   }

   /**
    * Read a page packet.  If the page is available in the cache
    * then return that data.
    *
    * @param page  page to read
    * @param readBuf buffer to place the data in
    * @param offset offset into the read buffer 
    *
    * @return the number byte in the packet
    *
    * @throws OneWireException when the adapter is not setup properly
    * @throws OneWireIOException when an 1-Wire IO error occures
    */
   public int readPagePacket(int page, byte[] readBuf, int offset)
      throws OneWireIOException, OneWireException
   {

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___readPagePacket (" + page + ") ");

      // check if need to read the page bitmap for the first time
      if (!pbmRead)
      {

         // read the pbm 
         pbmBank.read(0, false, pbmCache, 0, pbmCache.length);

         // make a copy of it
         System.arraycopy(pbmCache, 0, pbmCacheModified, 0, pbmCache.length);

         // mark as read
         pbmRead = true;

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
         {
            System.out.print("_Page bitmap read: ");
            debugDump(pbmCache, 0, pbmCache.length);
         }
      }

      // check to see if the requested page is cached
      if ((len [page] == EMPTY) || (redirect [page] != 0))
      {

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
            System.out.println("_Not in cache or redirected, length="
                               + len [page] + " redirect=" + redirect [page]);

         // page not cached, so read it
         // loop to get the bank specific page number
         int cnt = bankPages [0], bank_num = 0;

         for (;;)
         {
            if (page < cnt)
            {

               // read the page into cache
               PagedMemoryBank pmb =
                  (PagedMemoryBank) banks.elementAt(bank_num);

               //\\//\\//\\//\\//\\//\\//\\//
               if (doDebugMessages)
                  System.out.println("_Look in MemoryBank "
                                     + pmb.getBankDescription());

               if (getRedirection)
               {

                  // don't use multi-bank page reference (would not work with redirect)
                  // loop while page is redirected
                  int loopcnt = 0;

                  for (;;)
                  {

                     // check for redirection
                     if (redirect [page] == 0)
                     {

                        // check if already in cache
                        if (len [page] != EMPTY)
                           break;

                        // read the page with device generated CRC
                        pmb.readPageCRC(page, (lastPageRead == (page - 1)),
                                        cache [page], 0, tempExtra);

                        // set the last page read
                        lastPageRead = page;

                        // get the redirection byte
                        redirect [page] = ~tempExtra [0] & 0x00FF;

                        //\\//\\//\\//\\//\\//\\//\\//
                        if (doDebugMessages)
                           System.out.println("_Page: " + page + "->"
                                              + redirect [page]
                                              + " with packet length byte "
                                              + (int) cache [page][0]);

                        // not redirected so look at packet
                        if (redirect [page] == 0)
                        {

                           // check if length is realistic
                           if (cache [page][0] > maxPacketDataLength)
                              throw new OneWireIOException(
                                 "Invalid length in packet");

                           // verify the CRC is correct
                           if (CRC16.compute(cache [page], 0, cache [page][0]
                                             + 3, page) == 0x0000B001)
                           {

                              // get the length
                              len [page] = cache [page][0];

                              //System.out.print("[" + page + "]"); //????????????????

                              break;
                           }
                           else
                              throw new OneWireIOException(
                                 "Invalid CRC16 in packet read " + page);
                        }
                     }
                     else
                        page = redirect [page];

                     // check for looping redirection
                     if (loopcnt++ > totalPages)
                        throw new OneWireIOException(
                           "Circular redirection of pages");
                  }

                  //\\//\\//\\//\\//\\//\\//\\//
                  if (doDebugMessages)
                  {
                     System.out.print("_Data found (" + len [page] + "):");
                     debugDump(cache [page], 1, len [page]);
                  }

                  // get copy of data for caller
                  if (readBuf != null)
                     System.arraycopy(cache [page], 1, readBuf, offset,
                                      len [page]);

                  return len [page];
               }
               else
               {

                  // check if 'page' is in this memory bank
                  if ((cnt + bankPages [bank_num]) >= page)
                  {

                     // loop if get a crc error in packet data until get same data twice
                     for (;;)
                     {
                        pmb.readPage(page - (cnt - bankPages [bank_num]),
                                     (lastPageRead == (page - 1)), tempPage,
                                     0);

                        //\\//\\//\\//\\//\\//\\//\\//
                        if (doDebugMessages)
                           System.out.println("_Page: " + page
                                              + " translates to "
                                              + (page
                                                 - (cnt
                                                    - bankPages [bank_num])));

                        // set the last page read
                        lastPageRead = page;

                        // verify length is realistic
                        if ((tempPage [0] & 0x00FF) <= maxPacketDataLength)
                        {

                           // verify the CRC is correct
                           if (CRC16.compute(tempPage, 0, tempPage [0]
                                             + 3, page) == 0x0000B001)
                           {

                              // valid data so put into cache
                              System.arraycopy(tempPage, 0, cache [page], 0,
                                               tempPage.length);

                              // get the length
                              len [page] = tempPage [0];

                              //System.out.print("[" + page + "]"); //????????????????

                              break;
                           }
                        }

                        //\\//\\//\\//\\//\\//\\//\\//
                        if (doDebugMessages)
                        {
                           System.out.print("_Invalid CRC, raw: ");
                           debugDump(tempPage, 0, tempPage.length);
                        }

                        // must have been invalid packet
                        // compare with data currently in the cache
                        boolean same_data = true;

                        for (int i = 0; i < tempPage.length; i++)
                        {
                           if ((tempPage [i] & 0x00FF)
                                   != (cache [page][i] & 0x00FF))
                           {

                              //\\//\\//\\//\\//\\//\\//\\//
                              if (doDebugMessages)
                                 System.out.println("_Differenet at position="
                                                    + i);

                              same_data = false;

                              break;
                           }
                        }

                        // if the same then throw the exception, else loop again
                        if (same_data)
                           throw new OneWireIOException(
                              "Invalid CRC16 in packet read");
                        else
                           System.arraycopy(tempPage, 0, cache [page], 0,
                                            tempPage.length);
                     }
                  }

                  //\\//\\//\\//\\//\\//\\//\\//
                  if (doDebugMessages)
                  {
                     System.out.print("_Data found (" + len [page] + "):");
                     debugDump(cache [page], 1, len [page]);
                  }

                  // get copy of data for caller
                  if (readBuf != null)
                     System.arraycopy(cache [page], 1, readBuf, offset,
                                      len [page]);

                  return len [page];
               }
            }

            lastPageRead = NONE;

            bank_num++;

            cnt += bankPages [bank_num];
         }
      }
      else
      {

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
         {
            System.out.print("_In cache (" + len [page] + "):");
            debugDump(cache [page], 1, len [page]);
         }

         // get from cache
         if (readBuf != null)
            System.arraycopy(cache [page], 1, readBuf, offset, len [page]);

         return len [page];
      }
   }

   /**
    * Write a page packet into the cache.
    *
    * @param page  page to write
    * @param writeBuf buffer container the data to write
    * @param offset offset into write buffer
    * @param buflen length of data to write
    */
   public void writePagePacket(int page, byte[] writeBuf, int offset,
                               int buflen)
   {
      int log;

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
      {
         System.out.print("___writePagePacket on page " + page
                          + " with data (" + buflen + "): ");
         debugDump(writeBuf, offset, buflen);
      }

      // put in cache
      System.arraycopy(writeBuf, offset, cache [page], 1, buflen);

      len [page]      = buflen;
      cache [page][0] = (byte) buflen;

      // set changed flag
      changed [page] = true;

      // record write in log
      // search the write log until find 'page' or EMPTY
      for (log = 0; log < totalPages; log++)
      {
         if ((writeLog [log] == page) || (writeLog [log] == EMPTY))
            break;
      }

      // shift write log down 1 to 'log'
      for (; log > 0; log--)
         writeLog [log] = writeLog [log - 1];

      // add page at top
      writeLog [0] = page;
   }

   /**
    * Flush the pages written back to the 1-Wire device.
    *
    * @throws OneWireException when the adapter is not setup properly
    * @throws OneWireIOException when an 1-Wire IO error occures
    */
   public void sync()
      throws OneWireIOException, OneWireException
   {
      int cnt, bank_num, page, log;

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___sync");

      if (getRedirection)
         throw new OneWireException("Write to OTP not yet supported!");

      // loop through write log and write the oldest pages first
      for (log = totalPages - 1; log >= 0; log--)
      {

         // check if this is a valid log entry
         if (writeLog [log] != EMPTY)
         {

            // get page number to write
            page = writeLog [log];

            // make sure page has changed
            if (changed [page])
            {
               //\\//\\//\\//\\//\\//\\//\\//
               if (doDebugMessages)
               {
                  System.out.print("_write page " + page + " with data ("
                                   + len [page] + "): ");
                  debugDump(cache [page], 1, len [page]);
               }

               // loop to get the bank specific page number
               cnt      = bankPages [0];
               bank_num = 0;

               for (;;)
               {
                  if (page < cnt)
                  {

                     // write data from cache 
                     PagedMemoryBank pmb =
                        (PagedMemoryBank) banks.elementAt(bank_num);

                     pmb.writePagePacket(page - (cnt - bankPages [bank_num]),
                                         cache [page], 1, len [page]);

                     //System.out.print("{" + page + "}"); //????????????????

                     // clear changed flag
                     changed [page] = false;
                     lastPageRead   = NONE;

                     break;
                  }

                  bank_num++;

                  cnt += bankPages [bank_num];
               }
            }
         }
      }

      // clear the write log
      for (log = 0; log < totalPages; log++)
         writeLog [log] = EMPTY;

      // write the bitmap of used pages???????

      //???????? could blast all cached pages to force reads???????
   }

   //--------
   //-------- Owner tracking methods
   //--------

   /**
    * Add an owner to this memory cache.  This will be tracked
    * for later cleanup. 
    *
    * @param tobj owner of instance
    */
   public void addOwner(Object tobj)
   {

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___addOwner");

      if (owners.indexOf(tobj) == -1)
      {
         owners.addElement(tobj);
      }
   }

   /**
    * Remove the specified owner of this memory cache.  
    *
    * @param tobj owner of instance
    */
   public void removeOwner(Object tobj)
   {

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___removeOwner");

      owners.removeElement(tobj);
   }

   /**
    * Check to see if there on no owners of this memory cache.
    *
    * @return true if not owners of this memory cache
    */
   public boolean noOwners()
   {

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___noOwners = " + owners.isEmpty());

      return owners.isEmpty();
   }

   //--------
   //-------- Write file tracking methods
   //--------

   /**
    * Remove the provided filePath from the list of files
    * currently opened to write.
    *
    * @param filePath file to remove from write list
    */
   public void removeWriteOpen(String filePath)
   {
      int index = openedToWrite.indexOf(filePath);

      if (index != -1)
         openedToWrite.removeElementAt(index);
   }

   /**
    * Check to see if the provided filePath is currently opened
    * to write.  Optionally add it to the list if it not already
    * there.  
    *
    * @param filePath  file to check to see if opened to write
    * @param addToList true to add file to list if not present
    *
    * @return true if file was not in the opened to write list
    */
   public boolean isOpenedToWrite(String filePath, boolean addToList)
   {
      int index = openedToWrite.indexOf(filePath);

      if (index != -1)
         return true;
      else
      {
         if (addToList)
            openedToWrite.addElement(filePath);

         return false;
      }
   }

   //--------
   //-------- Page-Bitmap methods
   //--------

   /**
    * Check to see if this memory cache should handle the page bitmap.
    *
    * @return true if this memory cache should handle the page bitmap
    */
   public boolean handlePageBitmap()
   {
      return !(pbmBank == null);
   }

   /**
    * Mark the specified page as used in the page bitmap.
    *
    * @param page number to mark as used 
    */
   public void markPageUsed(int page)
   {
      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___markPageUsed " + page);

      // mark page used in cached bitmap of used pages
      Bit.arrayWriteBit(USED, pbmBitOffset + page, pbmByteOffset,
                        pbmCacheModified);
   }

   /**
    * free the specified page as being un-used in the page bitmap
    *
    * @param page number to mark as un-used 
    *
    * @return true if the page as be been marked as un-used, false
    *      if the page is on an OTP device and cannot be freed
    */
   public boolean freePage(int page)
   {

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.print("___freePage " + page);

      // only free pages that have been written to cache
      // but not flushed to device
      if (Bit.arrayReadBit(pbmBitOffset + page, pbmByteOffset, pbmCache)
              == NOT_USED)
      {
         Bit.arrayWriteBit(NOT_USED, pbmBitOffset + page, pbmByteOffset,
                           pbmCacheModified);

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
            System.out.println("_ was cached so really free now ");

         return true;
      }
      else
      {

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
            System.out.println("_ not cached so not free");

         return false;
      }
   }

   /**
    * Get the first free page from the page bitmap.
    *
    * @return first page number that is free to write
    */
   public int getFirstFreePage()
   {

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.print("___getFirstFreePage ");

      lastFreePage = 0;

      return getNextFreePage();
   }

   /**
    * Get the next free page from the page bitmap.
    *
    * @return next page number that is free to write
    */
   public int getNextFreePage()
   {
      for (int pg = lastFreePage; pg < totalPages; pg++)
      {
         if (Bit.arrayReadBit(
                 pbmBitOffset
                 + pg, pbmByteOffset, pbmCacheModified) == NOT_USED)
         {

            //\\//\\//\\//\\//\\//\\//\\//
            if (doDebugMessages)
               System.out.println("___getNextFreePage " + pg);

            return pg;
         }
      }

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___getNextFreePage, no free pages ");

      return -1;
   }

   /**
    * Get the total number of free pages in this file system.
    *
    * @return number of pages free
    *
    * @throws OneWireException when an IO exception occurs
    */
   public int getNumberFreePages()
      throws OneWireException      
   {
      // check if need to read the page bitmap for the first time
      if (!pbmRead)
      {
         // read the pbm 
         pbmBank.read(0, false, pbmCache, 0, pbmCache.length);

         // make a copy of it
         System.arraycopy(pbmCache, 0, pbmCacheModified, 0, pbmCache.length);

         // mark as read
         pbmRead = true;

         //\\//\\//\\//\\//\\//\\//\\//
         if (doDebugMessages)
         {
            System.out.print("_Page bitmap read in getNumberFreePages: ");
            debugDump(pbmCache, 0, pbmCache.length);
         }
      }

      int free_pages = 0;
      for (int pg = 0; pg < totalPages; pg++)
      {
         if (Bit.arrayReadBit(
                 pbmBitOffset
                 + pg, pbmByteOffset, pbmCacheModified) == NOT_USED)
            free_pages++;
      }

      //\\//\\//\\//\\//\\//\\//\\//
      if (doDebugMessages)
         System.out.println("___getNumberFreePages = " + free_pages);

      return free_pages;
   }

   /**
    * Gets the page number used in the remote page bitmap in an OTP device.
    *
    * @return page number used in the directory for the remote page bitmap
    */
   public int getBitMapPageNumber()
   {
      return (pbmBank.getStartPhysicalAddress() / pbmBank.getPageLength());
   }

   /**
    * Get the number of pages used for the remote page bitmap in an
    * OTP device.
    *
    * @return number of pages used in page bitmap
    */
   public int getBitMapNumberOfPages()
   {
      return ((totalPages / 8) / pbmBank.getPageLength());
   }

   //--------
   //-------- Misc Utility Methods
   //--------

   /**
    * Debug dump utility method
    *
    * @param buf buffer to dump
    * @param offset offset to start in the buffer
    * @param len length to dump
    */
   private void debugDump(byte[] buf, int offset, int len)
   {
      for (int i = offset; i < (offset + len); i++)
      {
         System.out.print(Integer.toHexString((int) buf [i] & 0x00FF) + " ");
      }

      System.out.println();
   }
}
