/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

// KLA - rehack! threw the old version out, replaced with this leaner, 
//       meaner, 1300 byte version
package com.dalsemi.shell.server;

/**
 * <p>Utility loosely based on the getopt command on UNIX,  
 * used to parse command line arguments.</p>
 * <h3>Usage</h3>
 * <p>Normal operation using the <code>GetOpt</code> class will
 * look something like this:</p>
 * <code><pre>
 *       int opt;
 *       boolean use_ID = false;
 *       String name = null;
 *       String pass = null;
 *       //args is an array with the command line arguments
 *       GetOpt myopts = new GetOpt(args,"n:p:i");
 *       while ((opt = myopts.getopt()) != myopts.optEOF)
 *       {
 *           if (opt == 'i')
 *               use_ID = true;
 *           else if (opt == 'n')
 *               name = myopts.optArgGet();
 *           else if (opt == 'p')
 *               pass = myopts.optArgGet();
 *           else
 *           {
 *               //an error must have occurred
 *               err.println(getUsageString());
 *               return;
 *           }
 *       }
 * </pre></code>
 * <p>The format string passed to the Constructor tells <code>GetOpt</code> which arguments 
 * are valid, and which ones require more information.  For this example, the
 * arguments 'n', 'p', and 'i' are valid.  Arguments 'n' and 'p' require additional
 * information, as denoted by the colon ':' following the option letters.</p>
 * <p>Note that once an unexpected value is returned, the program stops parsing the
 * command line input.  Once <code>GetOpt</code> finds 
 * an error, it enters an unknown state.  Subsequent calls to getopt() 
 * cannot be expected to conform to normal behavior.</p>
 * <p><code>GetOpt</code> will accept several options defined together in one string.  With a format string
 * of "fdi", the argument string "-fdi" will be parsed as three separate arguments 'f',
 * 'd', and 'i'.  Note that if the format string is "f:di", it will fail unless the argument
 * string is changed to "-dif" so that 'f' is at the end of the argument.  Also, in this case
 * another argument must follow, i.e.:</p>
 * <code><pre>
 *         java MyProgram -dif argument
 * </pre></code>
 */
public class GetOpt
{
    private String[] args   = null;
    private int index = 0;
    private int internal_index = 0;
    private char[] optionArray = null;
    private boolean hasExtra = false;
    
    /**
     * Indicates there are no more arguments to draw from this object.
     */
    public static final int optEOF = -1; 

    /**
     * Indicates there was an error in parsing the arguments.  
     * Errors occur if an unspecified option is encountered or an option
     * that should be followed by a <code>String</code> value is not.  Once an error
     * is encountered, the program should stop parsing arguments.  Note that
     * an error is only reported once it is reached in the <code>getopt()</code> algorithm. 
     * It is possible to receive one or more valid options and argument strings
     * before receiving an error.
     */
    public static final int optERR = -2;
    
    /**
     * <p>Constructor for a <code>GetOpt</code> object to parse command line arguments.</p>
     *
     * <p> The first parameter should be a <code>String[]</code> object containing all of the
     * command line arguments.  If a program is called with the following 
     * command line:<br>
     * <code><pre>
     *       java MyProgram -abc one_argument another argument -de
     * </pre></code>
     * Then <code>args</code> should contain:<br>
     * <ol>
     *    <li>"-abc"</li>
     *    <li>"one_argument"</li>
     *    <li>"another"</li>
     *    <li>"argument"</li>
     *    <li>"-de"</li>
     * </ol></p>
     *
     * <p>The second parameter is a format string that tells <code>GetOpt</code> which options are 
     * valid and which of those options must have accompanying argument
     * strings.  All options can be only one character long.  A colon ':'
     * following an option in the option string means that an argument
     * is expected to follow this option.  For example, if the available
     * options for <code>MyProgram <code>are 'a', 'b', 'c', 'd', and 'e', and option 'c'
     * expects an extra argument, the format string is "abc:de".  Order of 
     * the characters is irrelevant:  the format string could also be
     * "ebadc:".
     *
     * @param args contains all of the command line arguments
     * @param opts format string that tells <code>GetOpt</code> which options are 
     *    valid and which of those options must have accompanying argument
     *    strings
     *
     */
    public GetOpt (String[] args, String opts)
    {
        this.args = args;
        optionArray = opts.toCharArray();
    }
   
    /**
     * <p>Returns the next option found on the command line.  If <code>args</code>
     * as passed to the Constructor represents the command line:</p>
     * <code><pre>
     *      java MyProgram -abc argument1 -d -e
     * </pre></code>
     * <p>Then <code>getopt()</code> will return in order 'a', 'b', 'c', 'd', 'e', and <code>optEOF</code>, 
     * assuming the format string is "abc:de".  If the format string is
     * "abc:d:e", then getopt() will return this order: 'a', 'b', 'c', 'd', and <code>optEOF</code>.
     * In this case, "-e" is the argument string for the option 'd'.</p>
     * 
     * @return the next option found on the command line
     * @see #optArgGet()
     */
    public synchronized int getopt()
    {
        int option = optEOF;
	    hasExtra = false;
	    if (index >= args.length) 
		    return option;
	    char[] temp = args[index].toCharArray();   
        //if we are in the middle of something that looks like
        //   -fad
        //then internal_index will not be 0
        if (internal_index!=0)
        {
            option = temp[internal_index+1];
            if (internal_index == temp.length-2)
            {
                //then we are done with this round of options,
                internal_index = 0;
                //move on to the next option
                index++;
            }
            else 
	        {
                internal_index++;
            }
        }
	    else
	    {
	        while (index<args.length)
	        {
	            if ((temp[0]=='-') && (temp.length > 1))
	                break;
	            index++;
		        temp = args[index].toCharArray();
	        }
	        if (index==args.length)
		        return optEOF;
		    option = temp[1];
		    if (temp.length>2)
			    internal_index = 1;
		    else
			    index++;
	    }
	    int about = aboutTheOption(option);
	    if (about==-1)
		    return optERR;
	    if (about==1)   //more required
	    {
		    //increment index so we know to look at next argument
		    index++;
		    hasExtra = true;
		    if ((internal_index!=0) || (index-1 >= args.length))
			    return optERR;
	    }

        return option;
    }

    //-1 : not an option
    // 0 : no extra's allowed
    // 1 : extra required
    private int aboutTheOption(int ch)
    {
	    if (ch!=':')
	    {
		    for (int i=0;i<optionArray.length;i++)
		    if (optionArray[i]==ch)
		    {
			    if ((i+1<optionArray.length) && (optionArray[i+1]==':'))
				    return 1;
			    return 0;
		    }
	    }
	    return -1;
    }

    /**
     * <p>Returns the argument string associated with the last option 
     * reported by <code>getopt()</code>, or <code>null</code> if no argument is 
     * associated with this argument.  In the following example:</p>
     * <code><pre>
     *      java MyProgram -abc argument1 -d -e
     * </pre></code>
     * <code>optArgGet()</code>'s output will differ based on the format string.
     * <table width=450>
     *   <tr>
     *     <td> Format String:   </td>  <td>"abc:de"  </td>        <td>"abc:d:e"</td>
     *   </tr>
     *   <tr>
     *     <td> Option/Argument 1:</td>  <td> 'a' / null </td>   <td> 'a' / null </td>
     *   </tr>
     *   <tr>
     *     <td> Option/Argument 2:</td>  <td> 'b' / null </td>   <td> 'b' / null </td>
     *   </tr>
     *   <tr>
     *     <td> Option/Argument 3:</td>  <td> 'c' / "argument1" </td>   <td> 'c' / "argument1" </td>
     *   </tr>
     *   <tr>
     *     <td> Option/Argument 4:</td>  <td> 'd' / null </td>   <td> 'd' / "-e" </td>
     *   </tr>
     *   <tr>
     *     <td> Option/Argument 5:</td>  <td> 'e' / null </td>   <td> NA </td>
     *   </tr>
     * </table>
     * 
     * 
     * @return the argument string associated with the last option reported
     * by <code>getopt()</code>
     */
    public String optArgGet()
    {
	    if ((!hasExtra) || (index-1 >= args.length))
		    return null;
	    return args[index-1];
    }
   
   

}
