
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell.server;



import java.io.*;
import java.util.*;


/**
 * A generic server.  Servers will listen on some system resource for
 * connection requests.  When someone attempts to login, the server
 * spins off a session to handle that login.  Servers are designed
 * to be multi-threaded, allowing multiple simultaneous connections.
 */
public abstract class Server
   extends Thread
{

   /**
    * List of sessions created by this server.
    */
   protected Vector sessions;

   /**
    * Stops this server when set to true.
    */
   protected boolean shutdown;

   /**
    * Initializes the server and creates the list of sessions (initially empty).
    */
   protected Server ()
   {
      sessions = new Vector();
      shutdown = false;
   }

   /**
    * Starts the server.  This method continuously loops checking for new
    * connections until the <code>shutdown</code> field is set to <code>true</code>.
    */
   public void run ()
   {
      while (!shutdown)
      {
         try
         {
            checkForNewSessions();
         }
         catch (Throwable t)
         {

            //Hmm... just no goodness can come out of a throwable coming all this
            //way up the stack.  All of the recoverable stuff will be handled further
            //down in the internals.  Bring down the server.
/*            com.dalsemi.system.Debug.debugDump("Server PANIC: "
                                               + t.toString());
*/
            try
            {
               if (!shutdown)
                  shutDown();
            }
            catch (Exception e)
            {

               //Be afraid.  Be very, very afraid.
               shutdown = true;
            }
         }
      }
   }

   /**
    * Listens for connection requests.  This method blocks until a request is received.
    * Once the new session is created, it is added to the list of known
    * sessions.
    */
   protected void checkForNewSessions ()
   {
      Session newSession = acceptNewSession();

      if (newSession != null)
      {
         sessions.addElement(newSession);
      }
   }

   /**
    * Waits for a connection request.  This method should block until a request is received.
    * When a request is made, it should create, initialize, and return a
    * new Session to handle that connection.
    *
    * @return  a new Session
    */
   protected abstract Session acceptNewSession ();

   /**
    * Requests that the server stop taking connections and terminate any current sessions. 
    * Cycles through the list of known sessions, closing down each one and then cleaning
    * up its own system resources.
    */
   public void shutDown ()
      throws IOException
   {
      if (!shutdown)
      {
         shutdown = true;

         if (sessions.size() > 0)
         {
            Enumeration sessionsEnum = sessions.elements();
            int         size         = sessions.size();

            for (int i = 0; i < size; i++)
            {
               (( Session ) sessions.elementAt(0)).forceEndSession();
            }

            if (!(this
                  instanceof com.dalsemi.shell.server.serial.SerialServer))
            {

               //If there are network connections open, we need to give the
               //connections time to close out.  If a particular socket
               //tries to close (i.e. send a FIN), and the other side doesn't
               //send a FIN to close his side, the socket will go into a
               //FIN_WAIT_2 wait state.  He will stay there until the other
               //side closes, or until the death timer fires.  (Currently after
               //3 seconds.)  So, we wait for 6 seconds to give plenty of close
               //out time.
               try
               {
                  Thread.sleep(6000);
               }
               catch (InterruptedException ie)
               {

                  //DRAIN
               }
            }
         }

         closeAllPorts();
      }
   }

   /**
    * Notifies this server that a particular session is ending.  Removes the closing session from the
    * list of known sessions.
    * 
    * @param session  the session that has terminated
    */
   public void sessionEnded (Session session)
   {
      sessions.removeElement(session);
      java.lang.System.gc();
   }

   /**
    * Gets an array of the names of all of the users that are currently connected
    * to this server.
    *
    * @return user names of all connected users
    */
   public String[] getConnectedUsers ()
   {
      String[]    users        = new String [sessions.size()];
      Enumeration sessionsEnum = sessions.elements();
      int         i            = 0;

      while (sessionsEnum.hasMoreElements())
      {
         users [i++] = (( Session ) sessionsEnum.nextElement()).getUserName();
      }

      return users;
   }

   /**
    * Cycles through the list of know sessions, sending the specified message.
    *
    * @param sendThis  message to send to all sessions
    */
   public void broadcast (String sendThis)
   {
      Enumeration sessionsEnum = sessions.elements();

      while (sessionsEnum.hasMoreElements())
      {
         (( Session ) sessionsEnum.nextElement()).broadcast(sendThis);
      }
   }

   /**
    * Cleans up any system resources held by this server.
    */
   protected abstract void closeAllPorts ()
      throws IOException;
}
