
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell.server.ftp;



import com.dalsemi.tininet.*;
import com.dalsemi.system.*;
import com.dalsemi.shell.server.*;
import java.io.*;
import java.net.*;
import java.util.Hashtable;
import java.util.*;


/**
 * This class encapsulates all functionality for an FTP session.
 * The FTP server listens to port 21 for connection requests.
 * When someone attempts to start an FTP session, they contact
 * port 21.  The server spins off a session (an instance of this
 * class) to handle that session.  The session allows you to log
 * into the system, then handles the communication between that
 * remote computer and the TINI system.  The FTP server and
 * sessions are designed to be multi-threaded, allowing multiple
 * simultaneous FTP connections.
 *
 * List of FTP return codes:
 * 120 Service ready in nnn minutes
 * 125 Data Connection already opening, transfer starting
 * 150 File Status OK, about to open data connection
 * 200 OK
 * 202 Command not implemented
 * 211 System status, or help reply
 * 212 Directory status
 * 213 File Status
 * 214 Help Message
 * 215 NAME system type
 * 220 Service ready for new user
 * 221 Service closing control connection
 * 225 Data connection open , no transfer in progress
 * 226 Closing data connection
 * 227 Entering passive mode (h1 h2 h3 h4 p1 p2)
 * 230 User logged in proceed
 * 250 Requested file action OK completed
 * 257 PATHNAME created
 * 331 User name OK, need password
 * 332 Need account for login
 * 350 Requested action pending further info
 * 421 Service not available, closing control connection
 * 425 Can't open data connection
 * 426 Connection closed, transfer aborted
 * 450 Req file action not taken
 * 451 Req action aborted - local error
 * 452 Req action not taken (out of memory)
 * 500 Syntax
 * 501 Syntax in parameters or arguments
 * 502 Command not implemented
 * 503 Bad sequence of commands
 * 504 Command not implemented for that parameter
 * 530 Not logged in
 * 532 Need acct for storing files
 * 550 Requested action not taken (file not found, no access, etc.)
 */
public final class FTPSession
   extends Session
{
   Socket socket;                         //Command socket.
   Socket datasocket;   //Transient datasocket.  For this session, the normal in, out

   //and err are the transient data streams, NOT the command streams!
   OutputStream        cmdOut;            //Command Port output
   FTPInputStream      cmdIn;             //Command Port input
   boolean             ASCII    = true;   // MODE of transmission
   private String      lastRNFR = "";     // tells if last command was RNFR
   private InetAddress dataSocketIP;   //Transient datasocket IP.  (Used with PORT commands.)
   private int         dataSocketPort;
   private boolean     loggedIn                = false;
   static byte[]       PORT_SUCCESSFUL         =
      "200 PORT Command successful.\r\n".getBytes();
   static byte[]       ROOT_ALLOWED            =
      "331 root login allowed. Password required.\r\n".getBytes();
   static byte[]       LOGGED_IN_A             = "230 User ".getBytes();
   static byte[]       LOGGED_IN_B             = " logged in.\r\n".getBytes();
   static byte[]       PASS_REQUIRED           =
      "331 Password Required for ".getBytes();
   static byte[]       OPENING_LIST_CONNECTION =
      "150 Opening ASCII mode data connection for file list.\r\n".getBytes();
   static byte[]       CLOSING_DATA_CONNECTION =
      "226 Closing data connection.\r\n".getBytes();
   static byte[]       LOGIN_READY             = null;
   static byte[]       GOODBYE                 =
      "221 Goodbye.\r\n".getBytes();
   static byte[]       NEWLINE                 = "\r\n".getBytes();

   /**
    * Initializes the session.
    *
    *
    * @param in        InputStream this session should use to get data from user.
    * @param out       OutputStream this session should use to output to user.
    * @param err       ErrorStream this session should use to output errors to user.
    * @param server    The server in charge of this session.
    */
   FTPSession (FTPInputStream in, SystemPrintStream out,
               SystemPrintStream err, Socket s, Server server)
   {
      super(in, out, err, server);

      password = null;

      if (LOGIN_READY == null)
         LOGIN_READY = ("220 Welcome to " + TINIOS.getShellName() + ".  ("
                        + TINIOS.getShellVersion()
                        + ")  Ready for user login.\r\n").getBytes();

      socket = s;
      cmdIn  = in;
      cmdOut = out;

      in.setSession(this);
      out.setSession(this);
   }

   /**
    * Logs the user into the system.  This function will prompt the user
    * for username and password.
    */
   public void login ()
      throws IOException
   {

      //Most of  the work for login happens when client
      //sends the USER and PASS commands, so just start
      //the connection dialog here...
      sendMessageFile("220-".getBytes(), FTPServer.getConnectionMsgFile());
      cmdOut.write(LOGIN_READY);
   }

   /**
    * Called to clean up when the session is ending.  (i.e. the
    * user typed "exit", or something bad happened that causes the
    * session to terminate.)  Closes all of the sockets associated
    * with this session.
    */
   protected void sessionEnding ()
   {
      try
      {
         try
         {
            cmdOut.write(GOODBYE);
         }
         catch (Throwable t)
         {
//            com.dalsemi.system.Debug.debugDump("Snif, didn't say goodbye.");   //DEBUG

            //DRAIN
         }

         in.close();
         out.close();
         cmdOut.close();
         cmdIn.close();
         socket.close();

         if (datasocket != null)
         {
            datasocket.close();

            datasocket = null;
         }

         cmdOut = null;
         cmdIn  = null;
         err    = null;
         out    = null;
         in     = null;
      }
      catch (Exception e)
      {

         // oh well, ending anyway
      }
   }

   /**
    * Called when the FTP command "USER username" was issued.
    */
   void setUser (String user)
      throws IOException
   {
      userName = user;
      loggedIn = false;

      if ((userName.equalsIgnoreCase("anonymous"))
              || userName.equalsIgnoreCase("ftp"))
      {
         if (FTPServer.isAnonymousAllowed())
         {
            cmdOut.write(
               "331 Anonymous login, please enter email address.\r\n".getBytes());
         }
         else
         {
            userName = null;

            cmdOut.write("500 Anonymous login not allowed.\r\n".getBytes());
         }
      }
      else if (userName.equalsIgnoreCase("root"))
      {
         if (FTPServer.isRootAllowed())
         {
            cmdOut.write(ROOT_ALLOWED);
         }
         else
         {
            userName = null;

            cmdOut.write("500 root login not allowed.\r\n".getBytes());
         }
      }
      else
      {
         cmdOut.write(PASS_REQUIRED);
         cmdOut.write(user.getBytes());
         cmdOut.write(NEWLINE);
      }
   }

   /**
    * Called when the FTP command "PASS password" was given.  This function will
    * attempt to log the user into the system.
    */
   void setPass (String pass)
      throws IOException
   {

      //Check anonymous login
      if (userName.equalsIgnoreCase("anonymous")
              || userName.equalsIgnoreCase("ftp"))
      {
         loggedIn = true;

         //Log the anonymous login if logging turned on...
         String ftpLog = FTPServer.logAnon();

         if (ftpLog != null)
         {
            try
            {
               BufferedOutputStream logFile = null;

               try
               {
                  logFile =
                     new BufferedOutputStream(new FileOutputStream(ftpLog,
                        true));
               }
               catch (FileNotFoundException fnfe)
               {
                  logFile =
                     new BufferedOutputStream(new FileOutputStream(ftpLog));
               }

               logFile.write(((new Date()).toString()).getBytes());
               logFile.write(':');
               logFile.write(pass.getBytes());
               logFile.write(NEWLINE);
               logFile.close();
            }
            catch (Exception e)
            {
/*               com.dalsemi.system.Debug.debugDump("Error logging: "
                                                  + e.toString());   //DEBUG
*/
            }
         }
      }
      else
      {
         if (TINIOS.login(userName, pass) == -1)
         {
            cmdOut.write(("530 Login incorrect.\r\n").getBytes());

            return;
         }
         else
         {
            loggedIn = true;
         }
      }

      sendMessageFile("230-".getBytes(), FTPServer.getWelcomeMsgFile());
      cmdOut.write(LOGGED_IN_A);
      cmdOut.write(userName.getBytes());
      cmdOut.write(LOGGED_IN_B);
      System.gc();
   }

   /**
    * This function will create a new data connection when the FTP command
    * "PORT xx,xx,xx,xx,xx,xx".  The port will be in the form:
    * "PORT 180,0,54,13,6,17"
    * The IP address in this case is 180.0.54.13
    * The port is (6 * 256 + 17)
    *
    * @param  port   The new IP and port in String form.
    */
   void setPort (String port)
      throws IOException
   {
      int commaCount = 0;
      int ipEnd      = 0;
      int portSep    = 0;

      for (int i = 0; i < port.length(); i++)
      {
         if (port.charAt(i) == ',')
         {
            commaCount++;

            if (commaCount == 4)
               ipEnd = i;
            else if (commaCount == 5)
               portSep = i;
         }
      }

      try
      {
         if (commaCount == 5)
         {
            dataSocketIP   = InetAddress.getByName(port.substring(0,
                    ipEnd).replace(',', '.'));
            dataSocketPort =
               (((Integer.parseInt(port.substring(ipEnd + 1, portSep), 10) & 0x0FF) << 8)
                | (Integer.parseInt(port.substring(portSep + 1), 10)
                   & 0x0FF));

            cmdOut.write(PORT_SUCCESSFUL);

            return;
         }
      }
      catch (NumberFormatException nfe)
      {
         dataSocketIP = null;
      }

      //String looks bad.
      throw new IOException("Port command syntax error.");
   }

   /**
    *
    */
   void openDataSocket ()
      throws Exception
   {
      if (datasocket != null)
      {
         datasocket.close();
      }

      datasocket = new Socket(dataSocketIP, dataSocketPort,
                              socket.getLocalAddress(), 20);

      //InetAddress.getByName(TININet.getIPAddress()), 20);
      in  = new FTPInputStream(datasocket.getInputStream(), true);   // this is a data connection
      out = new SystemPrintStream(datasocket.getOutputStream());
      err = out;
   }

   /**
    * Called when the FTP command "PASV" is issued to request us to create
    * a socket on which to listen for the user to contact us.  (As opposed to
    * using the PORT command.)  This is useful for users behind a firewall that
    * won't allow the FTP server to connect back to them.
    */
   void goPassiveMode ()
      throws IOException
   {
      if (datasocket != null)
      {
         datasocket.close();
      }

      //Randomly generate a local port.  We are going to put it in the
      //range of 30,000 to 60,000.
      int          port          = 30000
                                   + (Math.abs((new Random()).nextInt())
                                      % 30000);
      ServerSocket passiveSocket = new ServerSocket(port);

      passiveSocket.setSoTimeout(120000);   //Time out at 2 minutes?

      String passOut = "227 Entering Passive Mode ("
                       + TININet.getIPAddress().replace('.', ',') + ','
                       + Integer.toString(port / 256) + ','
                       + Integer.toString(port % 256) + ')' + "\r\n";

      cmdOut.write(passOut.getBytes());

      datasocket = passiveSocket.accept();
      in         = new FTPInputStream(datasocket.getInputStream(), true);   // this is a data connection
      out        = new SystemPrintStream(datasocket.getOutputStream());
      err        = out;
   }

   /**
    *
    */
   void putFile (String fileName)
      throws Exception
   {
      File currFile = new File(fileName);

      if (currFile.exists() &&!currFile.canWrite())
      {
         cmdOut.write(("553 " + fileName
                       + ": Permission denied.\r\n").getBytes());
      }
      else
      {
         cmdOut.write(("150 " + (ASCII ? "ASCII"
                                       : "BINARY") + " connection open, "
                                                   + "putting " + fileName
                                                   + "\r\n").getBytes());

         FileOutputStream fileOutStream = null;

         try
         {
            fileOutStream = new FileOutputStream(currFile);

            byte[] buffer    = new byte [512];
            int    bytesRead = 0;

            while (bytesRead != -1)
            {
               bytesRead = in.read(buffer, 0, 512);

               if (bytesRead != -1)
               {
                  fileOutStream.write(buffer, 0, bytesRead);
               }
            }
         }
         catch (Throwable t)
         {

            //DRAIN
         }

         fileOutStream.close();
         cmdOut.write(CLOSING_DATA_CONNECTION);

         if (datasocket != null)
         {
            datasocket.close();

            datasocket   = null;
            dataSocketIP = null;
         }
      }
   }

   /**
    *
    */
   void getFile (String fileName)
      throws Exception
   {
      File currFile = new File(fileName);

      if (!currFile.exists())
      {
         cmdOut.write(("550 " + fileName + ": No such file.\r\n").getBytes());
      }
      else if (!currFile.canRead())
      {
         cmdOut.write(("550 " + fileName
                       + ": Permission denied.\r\n").getBytes());
      }
      else if (currFile.isDirectory())
      {
         cmdOut.write(("550 " + fileName
                       + ": Not a plain file.\r\n").getBytes());
      }
      else
      {
         cmdOut.write(("150 " + (ASCII ? "ASCII"
                                       : "BINARY") + " connection open, "
                                                   + "getting " + fileName
                                                   + "\r\n").getBytes());

         FileInputStream fileInStream = null;

         try
         {
            fileInStream = new FileInputStream(currFile);

            long   fileLength       = currFile.length();
            long   fileStreamOffset = 0;
            byte[] fileData         = new byte [512];

            while (fileStreamOffset != fileLength)
            {
               int bytesRead = fileInStream.read(fileData);

               if (bytesRead <= 0)
                  break;

               out.write(fileData, 0, bytesRead);

               fileStreamOffset += bytesRead;
            }
         }
         catch (Throwable t)
         {

            //DRAIN
         }

         fileInStream.close();
         cmdOut.write(CLOSING_DATA_CONNECTION);

         if (datasocket != null)
         {
            datasocket.close();

            datasocket   = null;
            dataSocketIP = null;
         }
      }
   }

   /**
    *
    */
   void sendMessageFile (byte[] prepend, String messageFile)
   {
      try
      {
         if (messageFile != null)
         {
            File message = new File(messageFile);

            if (message.exists())
            {
               DataInputStream       inStream = new DataInputStream(
                  new BufferedInputStream(new FileInputStream(message)));
               ByteArrayOutputStream buff     = new ByteArrayOutputStream();
               String                line;

               while ((line = inStream.readLine()) != null)
               {
                  buff.write(prepend);
                  buff.write(line.getBytes());
                  buff.write(NEWLINE);
               }

               buff.write(prepend);
               buff.write(NEWLINE);
               inStream.close();
               buff.writeTo(cmdOut);
            }
         }
      }
      catch (Throwable t)
      {
//         com.dalsemi.system.Debug.debugDump("File error: " + t.toString());   //DEBUG
      }
   }

   /**
    * Called after each command is completed.  For example, the user
    * types "ls".  The command is received and parsed, then the appropriate
    * command is called in the shell.  Finally, we need send the correct
    * result code.
    */
   protected void currentCommandFinished ()
   {
      String currCommand = ( String ) environment.get(CURRENT_COMMAND);

      if ((currCommand == null) || (currCommand.length() == 0))
         return;

      try
      {
         if (currCommand.equals("ls"))
         {
            cmdOut.write(CLOSING_DATA_CONNECTION);

            if (datasocket != null)
            {
               datasocket.close();

               datasocket   = null;
               dataSocketIP = null;
            }
         }
         else if (currCommand.equals("move"))
         {
            cmdOut.write("250 File successfully renamed.\r\n".getBytes());
         }
         else if (currCommand.equals("cd"))
         {
            cmdOut.write("250 CWD command successful.\r\n".getBytes());

            //cmdOut.write("250 CWD command successful. \"/\" is current directory.\r\n".getBytes());
            //cmdOut.write(("250 Current Directory is "
            //            + environment.get(CURRENT_DIRECTORY)
            //            + "\r\n").getBytes());
         }
         else if (currCommand.equals("md"))
         {
            cmdOut.write("257 Directory created.\r\n".getBytes());
         }
         else if (currCommand.equals("del"))
         {
            cmdOut.write("250 File deleted.\r\n".getBytes());
         }
         else if (currCommand.equals("rd"))
         {
            cmdOut.write("250 Directory deleted.\r\n".getBytes());
         }
      }
      catch (IOException ioe)
      {
         forceEndSession();
      }
   }

   /**
    * Called when an exception is thrown in a command.  This function will
    * attempt to notify the user.  Any exceptions raised specifically by
    * shell commands should try to give as descriptive a message to the
    * exception as possible.
    *
    * @param ex  The exception thrown.
    */
   protected void exceptionThrown (Exception ex)
   {
      String currCommand = ( String ) environment.get(CURRENT_COMMAND);
      String message     = ex.getMessage();

      try
      {
         if (currCommand.equals("ls") || currCommand.equals("cd")
                 || currCommand.equals("move") || currCommand.equals("md")
                 || currCommand.equals("del") || currCommand.equals("rd"))
         {
            cmdOut.write(("550 " + message + "\r\n").getBytes());
         }
         else
         {
            cmdOut.write(("500 Command failed with the following reason: "
                          + message + "\r\n").getBytes());
         }

         if (currCommand.equals("ls"))
         {
            cmdOut.write(CLOSING_DATA_CONNECTION);

            if (datasocket != null)
            {
               datasocket.close();

               datasocket = null;
            }
         }
      }
      catch (IOException ioe)
      {
         forceEndSession();
      }

      environment.put(CURRENT_COMMAND, "");
   }

   /**
    * Gets the next command from this session's input stream.
    * This effectively performs functionality like a readLine(), but gives
    * the stream a chance to parse the incoming data.  This allows us to
    * translate FTP commands into system commands.
    *
    * @return  the next command from the input stream
    */
   public String getNextCommand ()
      throws IOException
   {
      String returnStr = "";
      String command   = cmdIn.readLine();

      if (command == null)
      {
//         com.dalsemi.system.Debug.debugDump("FS.gNC PANIC trouble.");
         forceEndSession();

         return null;
      }

      try
      {
         String args = null;
         int    last;

         if ((last = command.indexOf(' ')) != -1)
         {
            String temp = command;

            command = command.substring(0, last++);

            if (temp.length() > last)
               args = temp.substring(last);
         }

         /*this fixes the non-allowed r^H problem with a null username*/
         if (args == null)
            args = "";

         if (!(command.equalsIgnoreCase("RNTO")))
         {
            lastRNFR = "";
         }

         //Now, translate the FTP command into the correct system
         //command...
         if (command.equalsIgnoreCase("QUIT"))
         {
            returnStr = "quit";
         }
         else if (command.equalsIgnoreCase("USER"))
         {
            setUser(args);
         }
         else if (command.equalsIgnoreCase("PASS"))
         {
            setPass(args);
         }
         else if (!loggedIn)
         {
            cmdOut.write(
               "530 Please login with USER and PASS.\r\n".getBytes());
         }
         else if (command.equalsIgnoreCase("PORT"))
         {
            setPort(args);
         }
         else if (command.equalsIgnoreCase("XPWD")
                  || command.equalsIgnoreCase("PWD"))
         {
            cmdOut.write(("257 \"" + getFromEnvironment(CURRENT_DIRECTORY)
                          + "\" is current directory.\r\n").getBytes());
         }
         else if (command.equalsIgnoreCase("NLST"))
         {
            if ((dataSocketIP == null) && (datasocket == null))
            {
               cmdOut.write(
                  "425 Can't build data connection.\r\n".getBytes());
            }
            else
            {
               cmdOut.write(OPENING_LIST_CONNECTION);

               if ((args != null) && (args.length() != 0))
               {
                  returnStr = "ls " + args;
               }
               else
               {
                  returnStr = "ls";
               }

               if (dataSocketIP != null)
               {
                  openDataSocket();
               }
            }
         }
         else if (command.equalsIgnoreCase("LIST"))
         {
            if ((dataSocketIP == null) && (datasocket == null))
            {
               cmdOut.write(
                  "425 Can't build data connection.\r\n".getBytes());
            }
            else
            {
               cmdOut.write(OPENING_LIST_CONNECTION);

               if ((args != null) && (args.length() != 0))
               {
                  returnStr = "ls -l " + args;
               }
               else
               {
                  returnStr = "ls -l";
               }

               if (dataSocketIP != null)
               {
                  openDataSocket();
               }
            }
         }
         else if (command.equalsIgnoreCase("TYPE"))
         {

            // if args is null there's something wrong
            if (args.equalsIgnoreCase("I"))
            {
               ASCII = false;

               cmdOut.write("200 Type set to Binary\r\n".getBytes());
            }
            else if (args.equalsIgnoreCase("A"))
            {
               ASCII = true;

               cmdOut.write("200 Type set to ASCII\r\n".getBytes());
            }
            else
            {
               cmdOut.write(("504 Unknown parameter for TYPE: " + args
                             + "\r\n").getBytes());
            }
         }
         else if (command.equalsIgnoreCase("CWD"))
         {
            returnStr = "cd " + args;
         }
         else if (command.equalsIgnoreCase("RETR"))
         {
            if ((dataSocketIP == null) && (datasocket == null))
            {
               cmdOut.write(
                  "425 Can't build data connection.\r\n".getBytes());
            }
            else
            {
               if (dataSocketIP != null)
                  openDataSocket();

               getFile(args);
            }
         }
         else if (command.equalsIgnoreCase("STOR"))
         {
            if ((dataSocketIP == null) && (datasocket == null))
            {
               cmdOut.write(
                  "425 Can't build data connection.\r\n".getBytes());
            }
            else
            {
               if (dataSocketIP != null)
                  openDataSocket();

               putFile(args);
            }
         }
         else if (command.equalsIgnoreCase("RNTO"))
         {
            if (lastRNFR.length() == 0)
            {
               cmdOut.write("503 Bad command sequence\r\n".getBytes());

               returnStr = "";
            }
            else
            {
               returnStr = "move " + lastRNFR + ' ' + args;
               lastRNFR  = "";
            }
         }
         else if (command.equalsIgnoreCase("XMKD"))
         {
            returnStr = "md " + args;
         }
         else if (command.equalsIgnoreCase("RNFR"))
         {
            File temp = new File(args);

            if (temp.exists())
            {
               cmdOut.write(
                  "350 File found, ready to RNTO...\r\n".getBytes());

               lastRNFR = args;
            }
            else
            {
               cmdOut.write("550 File not found.\r\n".getBytes());

               lastRNFR = "";
            }
         }
         else if (command.equalsIgnoreCase("DELE"))
         {
            returnStr = "del " + args;
         }
         else if (command.equalsIgnoreCase("XRMD"))
         {
            returnStr = "rd " + args;
         }
         else if (command.equalsIgnoreCase("PASV"))
         {
            goPassiveMode();
         }
         else if (command.equalsIgnoreCase("SYST"))
         {
            cmdOut.write("215 UNIX Type: L8\r\n".getBytes());
         }
         else
         {
            cmdOut.write(
               "500 Unrecognized/Unsupported Command or Bad Syntax\r\n".getBytes());
         }
      }
      catch (Exception e)
      {
         try
         {
            returnStr = "";

            //Try to inform client that death and destruction has occurred.
            cmdOut.write(("550 Unknown internal failure: " + e.toString()
                          + "\r\n").getBytes());
         }
         catch (Exception ex)
         {

            //DRAIN
         }
         finally
         {

            //Panic, throw the exception.  This will drop this session.
            throw new IOException(e.toString());
         }
      }

      return returnStr;
   }
}
