
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell.server.serial;



import com.dalsemi.shell.server.*;
import java.io.*;
import java.net.*;
import java.util.Hashtable;
import com.dalsemi.system.*;


/**
 * This class encapsulates all functionality for a Serial session.
 * The Serial server listens to the serial port connection requests.
 * When someone attempts to start a Serial session, the server spins
 * off a session (an instance of this class) to handle that session.
 * The session allows you to log into the system, then handles the
 * communication between the user and the TINI system.
 */
public final class SerialSession
   extends Session
{

   /**
    * Initializes the session.
    *
    *
    * @param in        InputStream this session should use to get data from user.
    * @param out       OutputStream this session should use to output to user.
    * @param err       ErrorStream this session should use to output errors to user.
    * @param server    The server in charge of this session.
    */
   SerialSession (SystemInputStream in, SystemPrintStream out,
                  SystemPrintStream err, Server server)
   {
      super(in, out, err, server);
      prompt = ("\r\nTINI " + environment.get(CURRENT_DIRECTORY)
                + PROMPT).getBytes();
   }

   /**
    * Logs the user into the system.  This function will prompt the user
    * for username and password.
    */
   public void login ()
      throws IOException
   {
      out.println(welcomeMessage);

      boolean connected = false;

      while (!connected)
      {
         out.print("TINI login: ");

         userName = in.readLine();

         out.print("TINI password: ");
         in.setEcho(false);                                            // tell the input stream not to echo this

         password = in.readLine();

         in.setEcho(true);

         //Now, check to see if this is a correct user...
         if (TINIOS.login(userName, password) != -1)
         {
            connected = true;
         }
         else
         {
            out.println("Login incorrect.");
         }

         if (out.checkError())
         {
//            com.dalsemi.system.Debug.debugDump("SS.login trouble.");   //DEBUG
            forceEndSession();

            return;
         }
      }

      out.setSession(this);   // tell the output stream who the session is
      in.setSession(this);    // tell the input stream who the session is
      parseLogin();
      currentCommandFinished();
   }

   /**
    * Cleans up when the session is ending.  (i.e. the
    * user typed "exit", or something bad happened that caused
    * the session to terminate.)  Also resumes the Serial server so
    * new connections will be accepted.
    * 
    */
   protected void sessionEnding ()
   {
      out.setSession(null);
      in.setSession(null);
      server.resume();

      //Streams are closed in SerialServer...
   }

   /**
    * Called after each command is completed.  For example, the user
    * types "ls".  The command is received and parsed, then the appropriate
    * command is called in the shell.  Finally, we need to again display the
    * system prompt.
    */
   protected void currentCommandFinished ()
   {
      try
      {
         out.write(prompt);
      }
      catch (Exception e)
      {
//         com.dalsemi.system.Debug.debugDump("SERIAL: Prompt error.");
      }
   }

   /**
    * This method was added to speed up prompt printing.  It is called when
    * the prompt needs to change because the user changed directory.
    *
    * @param withThis  the current directory
    */
   public void updatePrompt (String withThis)
   {
      prompt = ("\r\nTINI " + withThis + PROMPT).getBytes();
   }

   /**
    * Called when an exception is thrown in a command.  This function will
    * attempt to notify the user.  Any exceptions raised specifically by
    * shell commands should try to give as descriptive a message to the
    * exception as possible.
    *
    * @param ex  The exception thrown.
    */
   protected void exceptionThrown (Exception ex)
   {
      String message = ex.getMessage();

      if ((message == null) || (message.equals("")))
      {
         out.println("Exception occurred: " + ex.toString());
      }
      else
         out.println(message);

      if (out.checkError())
      {
//         com.dalsemi.system.Debug.debugDump("SS.eT trouble.");   //DEBUG
         forceEndSession();
      }
   }

   /**
    * Gets the next command from the underlying session's input stream.
    * This effectively performs functionality like a readLine(), but gives
    * the stream a chance to parse the incoming data (to take care of backspace
    * events, etc.)
    *
    * @return  the next command from the input stream
    */
   public String getNextCommand ()
      throws IOException
   {
      return in.readLine();
   }
}
