
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.shell.server.telnet;



import com.dalsemi.shell.server.*;
import java.io.*;
import java.net.*;
import java.util.Hashtable;
import com.dalsemi.tininet.*;
import com.dalsemi.system.*;


/**
 * This class encapsulates all functionality for a Telnet session.
 * The Telnet server listens to port 23 for connection requests.
 * When someone attempts to start a Telnet session, they contact
 * port 23.  The server spins off a session (an instance of this
 * class) to handle that session.  The session allows you to log
 * into the system, then handles the communication between that
 * remote computer and the TINI system.  The Telnet server and
 * sessions are designed to be multi-threaded, allowing multiple
 * simultaneous Telnet connections.
 */
public final class TelnetSession
   extends Session
{
   Socket socket;

   /**
    * Initializes the session.
    *
    *
    * @param in        InputStream this session should use to get data from user.
    * @param out       OutputStream this session should use to output to user.
    * @param err       ErrorStream this session should use to output errors to user.
    * @param server    The server in charge of this session.
    */
   TelnetSession (SystemInputStream in, SystemPrintStream out,
                  SystemPrintStream err, Socket s, Server server)
   {
      super(in, out, err, server);

      socket = s;
      prompt = (TININet.getHostname() + ' '
                + environment.get(CURRENT_DIRECTORY) + PROMPT).getBytes();
   }

   /**
    * Logs the user into the system.  This function will prompt the user
    * for username and password.
    */
   public void login ()
      throws IOException
   {
      boolean connected = false;

      out.println(welcomeMessage);

      int loginAttempts = 0;

      while (!connected)
      {
         out.print(TININet.getHostname());
         out.print(" login: ");

         userName = in.readLine();

         if (userName == null)
         {
//            com.dalsemi.system.Debug.debugDump("TS.login PANIC trouble.");
            forceEndSession();

            return;
         }

         out.print(TININet.getHostname());
         out.print(" password: ");
         in.setEcho(false);   // tell the input stream not to echo this

         password = in.readLine();

         if (password == null)
         {
//            com.dalsemi.system.Debug.debugDump("TS.login PANIC trouble.");
            forceEndSession();

            return;
         }

         in.setEcho(true);

         //Now, check to see if this is a correct user/password...
         //First, though, check to see if user is attempting to login
         //as root.  If this is not allowed (based on the environment
         //variable "TELNET_ALLOW_ROOT"), then don't attempt login.
         if (userName.equals("root") &&!TelnetServer.isRootAllowed())
         {
            connected = false;
         }
         else if (TINIOS.login(userName, password) != -1)
         {
            connected = true;
         }

         if (!connected)
         {
            out.println("Login incorrect.");

            if (++loginAttempts == 5)
            {
               forceEndSession();

               return;
            }
         }

         if (out.checkError())
         {
//            com.dalsemi.system.Debug.debugDump("TS.login: PANIC");
            forceEndSession();

            return;
         }
      }

      out.setSession(this);   // tell the output stream who the session is
      in.setSession(this);    // tell the input stream who the session is

      //Now, display a welcome message if one was specified...
      String welcomeFile = TelnetServer.getWelcomeFile();

      if (welcomeFile != null)
      {
         try
         {
            File wf = new File(welcomeFile);

            if (wf.exists())
            {
               BufferedReader buff = new BufferedReader(new FileReader(wf));
               String         line;

               while ((line = buff.readLine()) != null)
               {
                  out.println(line);
               }

               buff.close();
            }
         }
         catch (Exception e)
         {
//            com.dalsemi.system.Debug.debugDump("TS welcome: " + e.toString());
         }
      }

      parseLogin();
      currentCommandFinished();
   }

   /**
    * Called to clean up when the session is ending.  (i.e. the
    * user typed "exit", or something bad happened that causes
    * the session to terminate.)  Closes all sockets opened by
    * this session.
    */
   protected void sessionEnding ()
   {
      try
      {
         try
         {
            err.println("Connection Terminated.");
         }
         catch (Throwable t)
         {
//            com.dalsemi.system.Debug.debugDump("Snif, didn't say goodbye.");

            //DRAIN
         }

         in.close();
         out.close();
         socket.close();

         out = null;
         err = null;
         in  = null;
      }
      catch (Throwable t2)
      {

         //DRAIN
      }
   }

   /**
    * Called after each command is completed.  For example, the user
    * types "ls".  The command is received and parsed, then the appropriate
    * command is called in the shell.  Finally, we need to again display the
    * system prompt.
    */
   protected void currentCommandFinished ()
   {
      try
      {
         out.write(prompt);
      }
      catch (Exception e)
      {
/*         com.dalsemi.system.Debug.debugDump("TELNET Prompt error: "
                                            + e.toString());
*/
      }
   }

   /**
    * This method was added to speed up prompt printing.  It is called when
    * the prompt needs to change because the user changed directory.
    *
    * @param withThis  the current directory
    */
   public void updatePrompt (String withThis)
   {

      //withThis will be the currentDirectory
      prompt = (TININet.getHostname() + ' ' + withThis + PROMPT).getBytes();
   }

   /**
    * Called when an exception is thrown in a command.  This function will
    * attempt to notify the user.  Any exceptions raised specifically by
    * shell commands should try to give as descriptive a message to the
    * exception as possible.
    *
    * @param ex  the exception thrown
    */
   protected void exceptionThrown (Exception ex)
   {
      try
      {
         String message = ex.getMessage();

         if ((message == null) || (message.length() == 0))
         {
            out.println("Exception occurred: " + ex.toString());
         }
         else
            out.println(message);

         if (out.checkError())
         {
//            com.dalsemi.system.Debug.debugDump("T.eT trouble.");
            forceEndSession();
         }
      }
      catch (Exception ioe)
      {
//         com.dalsemi.system.Debug.debugDump("Telnet.eT: PANIC!");
         forceEndSession();
      }
   }

   /**
    * Gets the next command from the underlying session's input stream.
    * This effectively performs functionality like a readLine(), but gives
    * the stream a chance to parse the incoming data (to take care of backspace
    * events, etc.)
    *
    * @return  the next command from the input stream
    */
   public String getNextCommand ()
      throws IOException
   {
      return in.readLine();
   }
}
