/****************************************************************************
*
*  Copyright (C) 2000 Dallas Semiconductor Corporation. 
*  All rights Reserved. Printed in U.S.A.
*  This software is protected by copyright laws of
*  the United States and of foreign countries.
*  This material may also be protected by patent laws of the United States 
*  and of foreign countries.
*  This software is furnished under a license agreement and/or a
*  nondisclosure agreement and may only be used or copied in accordance
*  with the terms of those agreements.
*  The mere transfer of this software does not imply any licenses
*  of trade secrets, proprietary technology, copyrights, patents,
*  trademarks, maskwork rights, or any other form of intellectual
*  property whatsoever. Dallas Semiconductor retains all ownership rights.
*
*     Module Name: PPPCommand.java
*/

package com.dalsemi.slush.command;

import java.io.*;
import java.util.*;
import javax.comm.*;
import com.dalsemi.slush.*;
import com.dalsemi.system.*;
import com.dalsemi.shell.server.*;
import com.dalsemi.tininet.ppp.*;


//
// Modem comamnd
//
class ModemCommand
{
    // Command to issue
    String command;
    // Expected response
    String response;
    // Maximum time to wait in seconds
    int timeout;
    
    // Do not allow uninitialized modem commands
    private ModemCommand(){};
    
    public ModemCommand(String command, String response, int timeout)
    {
        this.command = command;
        this.response = response;
        this.timeout = timeout;
    }
}

//
// PPP server/dialer
//
public class PPPCommand implements SlushCommand, PPPEventListener
{
    
    // Serial port interface
    public SerialPort serialPort = null;
    private InputStream inputStream;
    private OutputStream outputStream;
    private int serialPortNum = -1;
    
    // The PPP object
    PPP ppp = null;
    
    // If true will recycle server after disconnect
    private boolean recycle = true;
    
    // If true PPP server is running
    private boolean server = false;
    
    // True if connection is established
    private boolean connected = false;
    
    private byte[] localAddress = new byte[] {0, 0, 0, 0};
    private byte[] remoteAddress = new byte[] {0, 0, 0, 0};
    private String interfaceName = "ppp0";
    
    private ModemCommand[] dialSequence = {new ModemCommand("AT\r",          "OK",      5),
                                           new ModemCommand("ATZ\r",         "OK",      5),
                                           new ModemCommand("ATM1L0\r",      "OK",      5),
                                           new ModemCommand("ATDT1234567\r", "CONNECT", 30)
                                          };
    
	//
	// getUsageString method of the SlushCommand interface
	//
    public String getUsageString()
    {
        return "\r\nppp [options]\r\n\r\nSets the operating options for PPP connections." +
               "\r\n [-a xx.xx.xx.xx]    Set local IP address" +
               "\r\n [-c]                Close connection" +
               "\r\n [-p password]       Set login password" +
               "\r\n [-r xx.xx.xx.xx]    Set remote IP address" +
               "\r\n [-u username]       Set login username" +
               "\r\n [-s]                Start server" +
               "\r\n [-d]                Start client";

    }
    
    //
    // Slush command interface
    //
    public void execute(SystemInputStream in, SystemPrintStream out, SystemPrintStream err, 
                        String[] args, Hashtable env) throws Exception
    {
        // Enable external serial ports
        com.dalsemi.system.TINIOS.setExternalSerialPortEnable(2,true);
        com.dalsemi.system.TINIOS.setExternalSerialPortEnable(3,true);
        
        boolean startServer = false;
        boolean callServer = false;
        boolean closeConnection = false;
        String username = "";
        String password = "";
        String optionParam;
        
        // Only allow admin group to operate PPP
        if(!TINIOS.isCurrentUserAdmin())
        {
            throw new Exception("Permission denied.");
        }

        // Display current settings if no options specified
        if (args.length == 0)
        {
            // TODO; display command options or display connection status
        }
        
        // New ppp object if necessary
        if (ppp == null)
        {
            ppp = new PPP();
        }
          

        // Process command line option(s)
        int opt;
        GetOpt myopts = new GetOpt(args, "a:cdp:r:su:x:");
        
        while ((opt = myopts.getopt()) != myopts.optEOF)
        {
        	switch (opt) 
        	{
                // Local address
                case 'a':
                    String localAddressAsString = myopts.optArgGet();
                    localAddress = createIPFromString(localAddressAsString);
                    break;
                // Close connection
                case 'c':
                    closeConnection = true;
                    recycle = false;
                    break;
                // Dial
                case 'd':
                    callServer = true;
                    break;
                // Login password
                case 'p':
                    password = myopts.optArgGet();
                    break;
                // Remote address
                case 'r':
                    String remoteAddressAsString = myopts.optArgGet();
                    remoteAddress = createIPFromString(remoteAddressAsString);
                    break;
                // Start server
                case 's':
                    startServer = true;
                    break;
                // Login username
                case 'u':
                    username = myopts.optArgGet();
                    break;
                // Serial port number
                case 'x':
                    serialPortNum = Integer.parseInt(myopts.optArgGet());
                    break;
                // Ignore
                default:
                    break;
            }
        }
        
        //
        // Need to start server?
        //
        if (startServer)
        {
            if (connected == false)
            {
                server = true;
                
                recycle = true;
                
                // Open serial port
                if (openSerialPort() == false)
                {
                    closeSerialPort();
                    return;
                }
                    
                ppp.addEventListener(this);

                ppp.setACCM(0x00000000);
                ppp.setLocalAddress(localAddress);
                ppp.setRemoteAddress(remoteAddress);
                ppp.setAuthenticate(true);
                ppp.setDefaultInterface(true);
                ppp.setPassive(true);

                ppp.open();
            }
            else
            {
                out.println("Server already running");
            }
                            
        }
        //
        // Need to dial a server?
        //
        else if (callServer)
        {
            if (connected == false)
            {
                // Not a server
                server = false;
                // Do not recycle
                recycle = false;

                if (openSerialPort() == false)
                {
                    closeSerialPort();
                    return;
                }

                // Add this as an event listener
                ppp.addEventListener(this);
                
                // Set the local and remote IP address
                ppp.setLocalAddress(localAddress);
                ppp.setRemoteAddress(remoteAddress);
                
                // Set client peer type options
                ppp.setACCM(0x00000000);
                ppp.setAuthenticate(false);    
                ppp.setPassive(false);
                ppp.setUsername(username);
                ppp.setPassword(password);
                
                // Dialup server, ignoring errors
                for (int i = 0; i < dialSequence.length; ++i)
                {
                    atCommand(dialSequence[i]);
                }


                // Set connected flag
                connected = true;

                // Issue open command to PPP FSM
                ppp.open();
                
            }
            else
            {
                out.println("Connection already established");
            }
            
        }
        
        //
        // Bring server down or close dial-up connection
        //
        else if (closeConnection)
        {
            recycle = false;
            connected = false;
            ppp.close();
        }
            
    }
    
    //
    // PPP event listener interface
    //
    public void PPPEvent(PPPEvent ev)
    {
        switch (ev.getEventType())
        {
            case ev.STARTING:
                // This is the result of ppp.open()
                // If server operation is required wait for call
                if (server)
                {
                    if (waitForCall())
                    {
                        connected = true;
                    
                        // Signal PPP that serial port is available
                        ppp.up(serialPort);
                    }
                    else
                    {
                        ppp.close();
                    }
                }
                // Else place call
                else
                {
                    ppp.up(serialPort);
                }
                break;
                
            case ev.UP:
                ppp.addInterface(interfaceName);
                
                ((Slush)TINIOS.getShell()).broadcast("PPP connection established");
                
                break;
                
            case ev.STOPPED:
                // The remote peer has terminated the connection
                
                // Close PPP
                ppp.close();
                
                if (connected == false)
                    ((Slush)TINIOS.getShell()).broadcast("PPP connection failed");
               
                break;
                
            case ev.CLOSED:
                // This is the result of ppp.close()
                
                // Signal that the serial port is down
                ppp.down();
                
                connected = false;
                
                // Recycle PPP
                if (recycle == true)
                {
                    ppp.open();
                }
                else
                {
                    closePPP();
                    closeSerialPort();
                }

                ((Slush)TINIOS.getShell()).broadcast("PPP connection closed");

                ppp.removeInterface(interfaceName);

                break;
                
            case ev.AUTHENTICATION_REQUEST:
                // Retrieve username and password from authentication request
                // Calling getUsernamePassword with the second param as null will return the options length
                byte[] peerID = new byte[ppp.getUsernamePassword(ppp.PEER_ID, null)];
                ppp.getUsernamePassword(ppp.PEER_ID, peerID);
                byte[] password = new byte[ppp.getUsernamePassword(ppp.PEER_PASSWORD, null)];
                ppp.getUsernamePassword(ppp.PEER_PASSWORD, password);
                
                // Authenticate user, the slush user DB could be used
                if ( (new String(peerID).equals("root")) && (new String(password).equals("tini")) )
                    ppp.authenticate(true);
                else
                    ppp.authenticate(false);
                    
                break;
                
            // Ignore
            default:
                break;
        }
        
    }
    
    //
    // Cleanup PPP state
    //
    private void closePPP()
    {
        ppp.removeEventListener(this);
    }
    
    //
    // Close serial port
    //
    private void closeSerialPort()
    {
            /*
            // Reset modem
            serialPort.setDTR(false);
            serialPort.setRTS(false);
            */
            try
            {
                Thread.sleep(2000);
            }
            catch(Exception e){}
            /*
            serialPort.setDTR(true);
            serialPort.setRTS(true);
            */
        
        serialPort.close();
    }
    
    //
    // Open serial port
    //
    private boolean openSerialPort()
    {
        try 
        {
             
            CommPortIdentifier portId = CommPortIdentifier.getPortIdentifier("serial" + serialPortNum);
            
            serialPort = (SerialPort)portId.open(interfaceName, 0);
            
            serialPort.setSerialPortParams(19200, 
                                           SerialPort.DATABITS_8,
                                           SerialPort.STOPBITS_1,
                                           SerialPort.PARITY_NONE);
                                           
            serialPort.setFlowControlMode(serialPort.FLOWCONTROL_NONE);
            
            inputStream = serialPort.getInputStream();

            outputStream = serialPort.getOutputStream();

            long available = inputStream.available();
            if (available > 0)
                inputStream.skip(available);

            outputStream.flush();
            
            
            try 
            {
                /*
                // Reset modem
                serialPort.setDTR(false);
                serialPort.setRTS(false);
                */
                Thread.sleep(2000);
                
                serialPort.setDTR(true);
                serialPort.setRTS(true);
            }
            catch(Exception e)
            {
                com.dalsemi.system.Debug.debugDump(e.toString());
            }
            
        }
        catch (Exception e)
        {
            com.dalsemi.system.Debug.debugDump(e.toString());
            return false;
        }
        
        return true;
    }
    
    //
    // Create byte array from string
    //
    private static byte[] createIPFromString(String fromThis)
    {
        if (fromThis.equals(""))
            return new byte[0];
            
        byte[] toHere = null;
        try
        {
            int pos = 0;
            int oldPos = 0;
            int index = 0;
            
            while (pos != -1)
            {
                pos = fromThis.indexOf(".", oldPos);
                oldPos = pos + 1;
                if (pos != -1)
                    index++;
            }
            
            toHere = new byte[index + 1];
            
            index = 0;
            pos = 0;
            oldPos = 0;

            pos = fromThis.indexOf(".", oldPos);
            while (pos != -1)
            {
                toHere[index++] = (byte)Integer.parseInt(fromThis.substring(oldPos, pos));
                oldPos = pos + 1;
                pos = fromThis.indexOf(".", oldPos);
            }
            toHere[index++] = (byte)Integer.parseInt(fromThis.substring(oldPos));
        }
        catch(Exception e)
        {
            com.dalsemi.system.Debug.debugDump("Exception in TININet.createIPFromString: " + e.toString());
        }
        
        return toHere;
    }
    
    //
    // Establish incomming connection
    //
    private boolean waitForCall()
    {
        serialPort.setDTR(false);
        serialPort.setRTS(false);
        
        try
        {
            Thread.sleep(2000);
        }
        catch(Exception e){}
        
        serialPort.setDTR(true);
        serialPort.setRTS(true);
        
        try
        {
            Thread.sleep(2000);
        }
        catch(Exception e){}
        
        // Sync modem to serial port baud rate
        if (!atCommand(new ModemCommand("AT\r", "OK", 6)))
        {
            return false;
        }
            
        // Wait for ring
        waitFor("RING", null, 0);
        
        if (!atCommand(new ModemCommand("ATA\r", "CONNECT", 25)))
        {
            return false;
        }
        
        return true;
        
    }
    
    //
    // Wait for a string and send a response
    //
    private boolean waitFor(String expect, String response, int timeout)
    {
        boolean done = false;
        boolean returnValue = false;
        int numBytes;
        byte[] readBuffer = new byte[40];
        int bytesAvailable;
        String buffer = new String();
        
        timeout *= 2;
        
        while (done == false)
        {
            try {
                // Sleep for one-half second
                Thread.sleep(500);
                        
                if ((bytesAvailable = inputStream.available()) > 0)
                {
                    numBytes = inputStream.read(readBuffer, 0, bytesAvailable);
                    buffer = buffer.concat(new String(readBuffer, 0, numBytes));
                    if (buffer.indexOf(expect) != -1)
                    {
                        
                        try
                        {
                            if (response != null)
                                outputStream.write(response.getBytes());
                        }
                        catch(Exception e)
                        {
                            com.dalsemi.system.Debug.debugDump(e.toString() + " during write");
                        }
                        returnValue = true;
                        done = true;
                    }                
                }
                else
                {
                    // An initial timeout value of zero will wait forever
                    if (timeout != 0)
                    {
                        // Decrement seconds counter
                        if (--timeout == 0) 
                        {
                            returnValue = true; //false;
                            done = true;
                        }
                    }
                }
            }
            catch (Exception e)
            {
                com.dalsemi.system.Debug.debugDump(e.toString());
            }
        } 
        return returnValue;
    }
    
    //
    // Send modem an AT command and wait for response
    //
    public boolean atCommand(ModemCommand mc)
    {

        boolean returnValue = false;

        String response = new String();

        int timeout = mc.timeout * 2;
        
        try 
        {
            try
            {
                outputStream.write(mc.command.getBytes());
            }
            catch(Exception e)
            {
                com.dalsemi.system.Debug.debugDump("atCommand write " + e.toString());
                return false;
            }
            
            boolean done = false;
            while (done == false)
            {
                // Sleep for one-half second
                Thread.sleep(500);
                
                // Get available count
                int bytesAvailable = inputStream.available();
                if (bytesAvailable > 0)
                {
                    // Read available data
                    byte[] readBuffer = new byte[bytesAvailable];
                    int numBytes = inputStream.read(readBuffer, 0, bytesAvailable);
                    // Concat to existing response
                    response = response.concat(new String(readBuffer, 0, numBytes));
                    // See if expected response is contained in buffer
                    if (response.indexOf(mc.response) != -1)
                    {
                        // Set return value and escape loop
                        returnValue = true;
                        done = true;
                    }                
                }
                else
                {
                    // If not waiting forever
                    if (timeout > 0)
                    {
                        // Decrement seconds counter
                        if (--timeout == 0) 
                        {
                            returnValue = false;
                            done = true;
                        }
                    }
                }
            }            
        }
        catch (Exception e)
        {
            com.dalsemi.system.Debug.debugDump("atCommand all " + e.toString());
        }
        
        return returnValue;
    }
}