
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;



/**
 * <p>This class provides access to the TINI Real-Time clock. The methods support
 * setting and getting the clock values.  The Real-Time clock on TINI is an 8 <code>byte</code>
 * register that stores the year, month, date, day, hour, minute, second, and hundredth-second
 * in BCD format.<p>
 *
 * <h3>Usage</h3>
 *
 * <p>There are two basic ways to access TINI's Real-Time clock.  The faster method is to
 * directly use the values <code>year</code>, <code>month</code>, etc.</p>
 *
 * <pre><code>
 *      ...
 *      com.dalsemi.system.Clock C = new com.dalsemi.system.Clock();
 *      C.getRTC();
 *      System.out.println("The date is : " + C.getDate() + "/" +
 *                          C.getMonth() + "/" + C.getYear());
 *      System.out.println("The time is : " + C.getHour() + ":" +
 *                         C.getMinute() + ":" + C.getSecond() +
 *                         "." + C.getHundredth() + C.getPm() ? " PM" : " AM"));
 *      ...
 * </code></pre>
 *
 * <p>The other method is slower, but provides a mean to translate between the TINI Real-Time
 * clock and the Java timekeeping methods which require a <code>long</code> value of
 * milliseconds that have passed since January 1, 1970.  This method is slower, since the
 * <code>Clock</code> class must convert from its internal fields into an amount of
 * milliseconds.</p>
 *
 * <pre><code>
 *      ...
 *      com.dalsemi.system.Clock C = new com.dalsemi.system.Clock();
 *      long time = C.getTickCount();
 *      System.out.println(new java.util.Date(time));
 *      ...
 * </code></pre>
 *
 * <p>This class also provides means to set the Real-Time Clock settings.  The methods for
 * setting <code>Clock</code> values are analogous to the methods for getting, except that
 * after all fields have been set as desired, programs should call <code>setRTC()</code>.
 * The following code sets the TINI Real-Time clock to October 4, 2000 at 1:43 in the
 * afternoon.</p>
 *
 * <pre><code>
 *      C.setYear(2000);
 *      C.setMonth(10);
 *      C.setDate(4);
 *
 *      C.setHour(13);
 *      C.setPm(true);
 *      C.setMinute(43);
 *
 *      C.setRTC();
 *      System.out.println(new java.util.Date(C.getTickCount()));
 * </code></pre>
 * <p>Note that the time printed will probably not be 1:43 PM.  The time stored is
 * for GMT.  On a TINI with Central Standard Time as the default <code>TimeZone</code>, this program
 * prints out:</p>
 *
 * <code><pre>
 *    TINI /> java clockdemo.tini
 *    Wed Oct 4 08:43:59 CST 2000
 * </pre></code>
 * <p>which is correct because CST during Daylight Savings Time (CDT) is 5 hours behind GMT.
 * Also note that <code>getRTC()</code> is not called in the above code example.  It is
 * called internally by <code>getTickCount()</code>. Any other 'get' accesses should
 * call <code>getRTC()</code> first.</p>
 *
 */
public class Clock
{
    //modified 9-30-00 to fix the setTickCount problem on times at the 25th or later in the month KLA


   private byte[] rtcRegisters = new byte [8];

   /**
    * Stores the year mod 100.  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setYear(int)
    * @see #getYear()
    */
   protected int year;

   /**
    * Stores the month (1-12).  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setMonth(int)
    * @see #getMonth()
    */
   protected int month;

   /**
    * Stores the day of the week (1-7).  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setDay(int)
    * @see #getDay()
    */
   protected int day;

   /**
    * Stores the day of the month.  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setDate(int)
    * @see #getDate()
    */
   protected int date;

   /**
    * Stores the hour (1-12 or 0-23).  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setHour(int)
    * @see #getHour()
    */
   protected int hour;

   /**
    * Stores the minute.  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setMinute(int)
    * @see #getMinute()
    */
   protected int minute;

   /**
    * Stores the second.  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setSecond(int)
    * @see #getSecond()
    */
   protected int second;

   /**
    * Stores the hundredth-second.  This field is set by <code>getRTC()</code> and
    * used by the <code>setRTC()</code> method to set the Real-Time Clock.
    *
    * @see #setHundredth(int)
    * @see #getHundredth()
    */
   protected int hundredth;

   /**
    * Stores the AM/PM selection.  The hour is interpreted based on
    * the value of this flag if the is12Hour flag is set.
    *
    * @see #setPm(boolean)
    * @see #getPm()
    */
   protected boolean pm;

   /**
    * This flag signifies if the value stored in <code>hour</code>
    * is for a 1-12 hour clock that uses AM and PM settings or
    * a 0-23 hour clock that does not use AM and PM settings.
    *
    * @see #set12Hour(boolean)
    * @see #get12Hour()
    */
   protected boolean is12Hour;

   /*
    * Sum of the previous days in the previous months.
    * KLA - I added 365 to the end of this for my newly hacked up
    * setTickCount method.
    */
   private int[] sumOfDays =
   {
      0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365
   };

   /* replaced with a long with a mask of these values...
    * Used for day of the week calculations.
    */
/*   private static byte[] monthKeyValue =
   {
      1, 4, 4, 0, 2, 5, 0, 3, 6, 1, 4, 6
   };*/

   /*
    * Native function to get clock data.
    */
   private static native int getRTCNative (byte[] clockRegisters);

   /*
    * Native function to set clock data.
    */
   private static native int setRTCNative (byte[] clockRegisters);

   /**
    * Creates a new <code>Clock</code> instance for accessing
    * TINI's Real-Time clock.  Also performs a call to
    * <code>getRTC()</code> so the fields of this <code>Clock</code>
    * object initially have valid values.
    *
    * @see #getRTC()
    */
   public Clock ()
   {
      getRTC();
   }

   /**
    * Sets the year value of the Real-Time Clock.  The clock will only
    * store the value of year mod 100.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.
    *
    * @param year the year to be set
    *
    * @see #getYear()
    * @see #setRTC()
    */
   public synchronized void setYear (int year)
   {
      this.year = year;
   }

   /**
    * Get the year value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the year as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setYear(int)
    */
   public int getYear ()
   {
      return year;
   }

   /**
    * Sets the month value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.
    *
    * @param month the month to be set
    *
    * @see #getMonth()
    * @see #setRTC()
    */
   public synchronized void setMonth (int month)
   {
      this.month = month;
   }

   /**
    * Get the month value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the month as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setMonth(int)
    */
   public int getMonth ()
   {
      return month;
   }

   /**
    * Sets the day of the week value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.  To facilitate
    * setting the correct day of the week, first call <code>calculateDayOfWeek</code>
    * which calculates the day of the week depending on the year, month, and date.
    *
    * @param day the day of the week to be set
    *
    * @see #calculateDayOfWeek(int,int,int)
    * @see #getDay()
    * @see #setRTC()
    */
   public synchronized void setDay (int day)
   {
      this.day = day;
   }

   /**
    * Get the day of the week value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the day of the week as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setDay(int)
    */
   public int getDay ()
   {
      return day;
   }

   /**
    * Sets the date (day of the month) value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.
    *
    * @param date the day of the month to be set
    *
    * @see #getDate()
    * @see #setRTC()
    */
   public synchronized void setDate (int date)
   {
      this.date = date;
   }

   /**
    * Get the date (day of month) value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the day of the month as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setDate(int)
    */
   public int getDate ()
   {
      return date;
   }

   /**
    * Sets the hour value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.  The value
    * actually stored in TINI's Real-Time Clock is also determined
    * by the <code>is12Hour</code> flag and the <code>pm</code> flag.
    *
    * @param hour the hour to be set
    *
    * @see #set12Hour(boolean)
    * @see #setPm(boolean)
    * @see #getHour()
    * @see #setRTC()
    */
   public synchronized void setHour (int hour)
   {
      this.hour = hour;
   }

   /**
    * Get the hour value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.  The methods <code>get12Hour()</code>
    * and <code>getPm()</code> should also be called to find the exact meaning
    * of the <code>hour</code> value.
    *
    * @return the hour as reported by TINI's Real-Time Clock
    *
    * @see #getPm()
    * @see #get12Hour()
    * @see #getRTC()
    * @see #setHour(int)
    */
   public int getHour ()
   {
      return hour;
   }

   /**
    * Sets the minute value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.
    *
    * @param minute the minute to be set
    *
    * @see #getMinute()
    * @see #setRTC()
    */
   public synchronized void setMinute (int minute)
   {
      this.minute = minute;
   }

   /**
    * Get the minute value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the minute as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setMinute(int)
    */
   public int getMinute ()
   {
      return minute;
   }

   /**
    * Sets the second value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.
    *
    * @param second the second to be set
    *
    * @see #getSecond()
    * @see #setRTC()
    */
   public synchronized void setSecond (int second)
   {
      this.second = second;
   }

   /**
    * Get the second value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the second as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setSecond(int)
    */
   public int getSecond ()
   {
      return second;
   }

   /**
    * Sets the hundredth-second value of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to <code>setRTC()</code>
    * should  be performed to actually set the Real-Time Clock.
    *
    * @param hundredth the hundredth-second to be set
    *
    * @see #getHundredth()
    * @see #setRTC()
    */
   public synchronized void setHundredth (int hundredth)
   {
      this.hundredth = hundredth;
   }

   /**
    * Get the hundredth-second value for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.
    *
    * @return the hundredth-second as reported by TINI's Real-Time Clock
    *
    * @see #getRTC()
    * @see #setHundredth(int)
    */
   public int getHundredth ()
   {
      return hundredth;
   }

   /**
    * Sets the AM/PM value of the Real Time Clock.  This only sets the
    * instance variable for this object.  A call to setRTC should
    * be performed to actually set the real time clock.  Note that
    * the value pf <code>pm</code> only matters if <code>is12Hour</code>
    * is <code>true</code>.
    *
    * @param pm <code>true</code> for PM, <code>false</code> for AM
    *
    * @see #getPm()
    * @see #setRTC()
    * @see #set12Hour(boolean)
    */
   public synchronized void setPm (boolean pm)
   {
      this.pm = pm;
   }

   /**
    * Get the AM/PM value for the Real Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.  The value of the pm flag is
    * only relevant if the is12Hour flag is set.  If the is12Hour flag is not
    * set, the value for <code>hour</code> will range between 0 and 23, and the
    * pm flag is irrelevant.
    *
    * @return <code>true</code> for PM, <code>false</code> for AM
    *
    * @see #getRTC()
    * @see #setPm(boolean)
    * @see #get12Hour()
    * @see #getHour()
    */
   public boolean getPm ()
   {
      return this.pm;
   }

   /**
    * Sets the 12/24 hour mode of the Real-Time Clock.  This only sets the
    * instance variable for this object.  A call to setRTC should
    * be performed to actually set the real time clock.  If the argument
    * is <code>true</code>, then the value of the flag <code>pm</code>
    * matters when <code>setRTC()</code> is called.
    *
    * @param  is12Hour <code>true</code> for 12 hour clock, <code>false</code>
    * for 24 hour clock
    *
    * @see #setRTC()
    * @see #get12Hour()
    * @see #setPm()
    */
   public synchronized void set12Hour (boolean is12Hour)
   {
      this.is12Hour = is12Hour;
   }

   /**
    * Get the 12/24 Hour mode for the Real-Time Clock.  To get the latest value,
    * <code>getRTC()</code> should be called before this method to refresh
    * the value from TINI's Real-Time Clock.  If the <code>is12Hour</code> flag is not
    * set, the value for <code>hour</code> will range between 0 and 23, and the
    * <code>pm</code> flag is irrelevant.  If the <code>is12Hour</code> flag is set, the value for
    * <code>hour</code> will range between 1 and 12, and the <code>pm</code> flag
    * will be relevant.
    *
    * @return  <code>true</code> for 12 hour clock, <code>false</code> for
    * 24 hour clock
    *
    * @see #getRTC()
    * @see #getPm()
    * @see #set12Hour(boolean)
    * @see #getHour()
    */
   public boolean get12Hour ()
   {
      return is12Hour;
   }

   /**
    * Reads Real time clock values from hardware clock and places them into
    * <code>Clock</code> instance fields.  This method should be called
    * to refresh the <code>Clock</code> instance fields, and before
    * any 'get' methods are called.
    *
    * @see #setRTC()
    */
   public synchronized void getRTC ()
   {
        getRTCNative(rtcRegisters);

        year  = bcdToInt(rtcRegisters [7]);
        month = bcdToInt(rtcRegisters [6]);
        date  = bcdToInt(rtcRegisters [5]);
        day   = bcdToInt(rtcRegisters [4] & 0x07);

        // Do bounds checking on reads.
        if ((month < 1) || (month > 12))
        month = 1;

        if ((date < 1) || (date > 31))
        date = 1;

        if ((day < 1) || (day > 7))
        day = 1;

        hour = rtcRegisters [3] & 0xFF;   // Make sure we mask, so we don't go negative.

        if (hour > 127)   // are we in 12 hour mode???
        {                 // yes !!!
        pm   = ((hour & 0x20) == 0x20);
        hour &= 0x1f;
        hour = bcdToInt(hour);

        if ((hour < 1) || (hour > 12))
            hour = 1;

        is12Hour = true;
        }
        else              // no !!!
        {
        hour &= 0x3f;
        hour = bcdToInt(hour);

        if ((hour < 0) || (hour > 23))
            hour = 0;

        pm       = (hour >= 12);
        is12Hour = false;
        }

        minute    = bcdToInt(rtcRegisters [2]);
        second    = bcdToInt(rtcRegisters [1]);
        hundredth = bcdToInt(rtcRegisters [0]);

        // Bounds check reads.
        if ((minute < 0) || (minute > 59))
        minute = 0;

        if ((second < 0) || (second > 59))
        second = 0;

        if ((hundredth < 0) || (hundredth > 99))
        hundredth = 0;
   }

   /**
    * <p>Get the current time in milliseconds since midnight Jan 1, 1970, UTC.
    * This method calls <code>getRTC()</code>, so it is unnecessary for
    * callers to do so.  The value returned is suitable to be used in
    * calls to Java classes such as <code>java.util.Date</code> and
    * <code>java.util.Calendar</code>.</p>
    *
    * <p><b>Note: This method assumes the current year is the year
    * 2000 or later.</b>  Since the Real-Time Clock only stores the
    * 2 least significant digits of the year, this method assumes
    * that the clock represents a time after midnight, January 1, 2000.</p>
    *
    * @return number of milliseconds since midnight, Jan 1, 1970 UTC
    *
    * @see #setTickCount(long)
    */
   public synchronized long getTickCount ()
   {

        getRTC();

        long adjust = -1;
        int  year   = this.year + 2000;
        // Obtained the following formula in-house.  It calculates number of days since
        // Jan 1, 1970.
        // First, find if month is past February and year is a leap year...
        if ((month > 2)
                && ((year % 4 == 0 && year % 100 != 0) || (year % 400 == 0)))
        {

        //Is a leap year and we are past Feb.
        adjust = 0;
        }

        int hour = is12Hour ? (pm ? ((this.hour != 12) ? this.hour + 12
                                                    : this.hour)
                                : this.hour)
                            : this.hour;

        //Total days is the number of ideal days since 1970 (i.e. if there were no leap
        //years, plus the number of leap days since 1970 (take the diff of current year
        //minus 1969 and divide by 4 (right shift by 2)), plus the sum of the days in
        //the previous complete months in this year, plus the day, plus an adjust if this
        //is a leap year and we are after February.  Deep breath.
        long totalDays = (365L * (year - 1970L) + ((year - 1969L) >> 2)
                        + sumOfDays [month - 1] + date + adjust);

        return (totalDays * 86400000L) + (hour * 3600000L)
            + (minute * 60000L) + (second * 1000L) + (hundredth * 10L);
   }

   /**
    * <p>Set the current time from a value in milliseconds since Jan 1, 1970, UTC.
    * This method calls <code>setRTC()</code> before exiting.  This method
    * must parse the millisecond value into fields for the year, month, date,
    * and so on.  This is more time consuming than setting the fields
    * individually.</p>
    *
    * <p>This method is not written to handle negative values for
    * <code>millis</code>.  Also note that this method will only store the
    * 2 least significant digits of the year.  Subsequent calls to
    * <code>getTickCount()</code> will add these 2 digits to the year
    * 2000 to come up with the current year.  If you call
    * <code>setTickCount()</code> with a value for a date in 1999,
    * then <code>getTickCount()</code> will return you a value for
    * a date in 2099.</p>
    *
    * @param millis number of milliseconds since midnight, Jan 1, 1970 UTC
    *
    * @see #getTickCount()
    */
   public synchronized void setTickCount (long millis)
   {
      long n;
      long j = 0;
      long adjust        = 0;
      long daysSince1970 = millis / 86400000;

      n     = (daysSince1970 + 731) << 2;
      year  = ( int ) (n / 1461);
      n     = (n - 1461 * year) >> 2;

      /*
       * 'n' now contains the number of days into the year we are at.
       */

      year      += 1968;

      month = 1;
      while (n>sumOfDays[month-1])
        month++;

      if ((n > 58)     //if past the 28th of Feb and its a leap year
              && ((year % 4 == 0 && year % 100 != 0) || (year % 400 == 0)))
      {
        if (n==59) //handle the 29th of feb correctly
            month--;
        else   //Is a leap year and we are past Feb.
            adjust = 1;
      }

      //j needs to be how many days had passed on the first of the month
      j = sumOfDays [month - 1] + adjust;

      if (n < j)
      {
         month--;
         j = sumOfDays [month - 1] + adjust;
      }

      date      = ( int ) (n - j + 1);

      adjust    = millis - (daysSince1970 * 86400000L);
      hour      = ( int ) (adjust / 3600000L);
      adjust    -= (hour * 3600000L);
      minute    = ( int ) (adjust / 60000L);
      adjust    -= (minute * 60000L);
      second    = ( int ) (adjust / 1000L);
      adjust    -= (second * 1000L);
      hundredth = ( int ) (adjust / 10L);
      is12Hour  = false;
      day       = calculateDayOfWeek(month, date, year);

      year = year % 100;

      setRTC();
   }

   /**
    * Reads the stored values  <code>year</code>, <code>month</code>,
    * <code>date</code>, etc. and stores them in TINI's Real-Time Clock.
    * This method performs bounds checking on all values, however it
    * does not ensure that invalid dates (February 30) or days of
    * the week (Friday for October 5 2000) have been entered.
    *
    * @see #getRTC()
    */
   public synchronized void setRTC ()
   {
        // Do bounds checking on clock writes.
        if ((month < 1) || (month > 12))
        month = 1;

        if ((date < 1) || (date > 31))
        date = 1;

        if ((day < 1) || (day > 7))
        day = 1;

        if ((minute < 0) || (minute > 59))
        minute = 0;

        if ((second < 0) || (second > 59))
        second = 0;

        if ((hundredth < 0) || (hundredth > 99))
        hundredth = 0;

        if (is12Hour)
        if ((hour < 1) || (hour > 12))
            hour = 1;
        else if ((hour < 0) || (hour > 23))
            hour = 0;

        rtcRegisters [7] = intToBCD(year % 100);
        rtcRegisters [6] = intToBCD(month);
        rtcRegisters [5] = intToBCD(date);
        rtcRegisters [4] =
        ( byte ) ((intToBCD(day) | 0x10 /* resetBar bit */)
                    & 0x01F /* oscRunBar run */);
        rtcRegisters [3] = intToBCD(hour);

        if (is12Hour)
        rtcRegisters [3] =
            ( byte ) ((( byte ) rtcRegisters [3] & ( byte ) 0x1f) | (0x080   /* 12 hour bit */
                                        | (pm ? 0x20
                                            : 0) /* pm bit */));

        rtcRegisters [2] = intToBCD(minute);
        rtcRegisters [1] = intToBCD(second);
        rtcRegisters [0] = intToBCD(hundredth);

        setRTCNative(rtcRegisters);
   }

   /**
    * Calculates the correct day of the week given the month, day
    * of the month, and year.  This method will not work for dates
    * before 1900 or after the year 3000.
    *
    * @param  month     the month (1-12)
    * @param  date      the day of the month
    * @param  fullYear  the year (not mod 100)
    *
    * @return The day of the week (1-7).  1 is Saturday, 2 is Sunday, etc.
    */
   public static int calculateDayOfWeek (int month, int date, int fullYear)
   {

      /*  How to calculate the day of the week
          ====================================
          (1) Take the last two digits of the year.
          (2) Divide by 4, discarding any fraction.
          (3) Add the day of the month.
          (4) Add the month's key value: JFM AMJ JAS OND: 144 025 036 146
          (5) Subtract 1 for January or February of a leap year.
          A year will be a leap year if it is divisible by 4 but not by 100.
          If a year is divisible by 4 and by 100, it is not a leap year unless
          it is also divisible by 400.
          (6) For a Gregorian date, add 0 for 1900's, 6 for 2000's,
              4 for 1700's, 2 for 1800's; for other years, add or subtract
              multiples of 400.
          (7) Add the last two digits of the year.
          (8) Divide by 7 and take the remainder.
      */

      //Now, calculate the day of the week from the date supplied
      int yearMod100 = fullYear % 100;
      
      /* the following line is along the lines of what I did in
       * java.util.SimpleTimeZone.figureOutRuleDay.  Basically, there
       * used to be this array that had to be static that kept 
       * what amounted to the offsets in the day of the week that all 
       * the months start on (relative to each other).  All the values
       * were small and could be stored in 3 bits, so I made a mask that
       * stored all these values, so all I have to do is suck it out of the mask
       * by shifting over some multiple of three and masking the bits I want!
       */
      int temp       = (yearMod100 / 4) + date + (int)((0xd0e62a121L >> ((month-1) * 3)) & 0x7);
      //int temp       = (yearMod100 / 4) + date + monthKeyValue [month - 1];
      
      //Figure out if this is a leap year
      boolean leap = false;

      if ((fullYear % 4) == 0)
      {
         if ((fullYear % 100) == 0)
         {
            if ((fullYear % 400) == 0)
            {
               leap = true;
            }
         }
         else
            leap = true;
      }

      if (leap)
      {
         if (month < 3)
            temp--;
      }

      //This will not work for pre 1900's or the year 3000.  So there.
      if (fullYear >= 2000)
         temp += 6;

      temp += yearMod100;

      return ((temp % 7) + 1);
   }

   /**
    * Converts a BCD integer byte to its binary integer value.
    *
    * @param bcdVal the BCD integer to convert to a decimal integer
    *
    * @return the converted decimal integer
    *
    * @see #intToBCD(int)
    */
   public static int bcdToInt (int bcdVal)
   {
      bcdVal &= 0x0ff;

      return (bcdVal & 0x0f) + ((bcdVal & 0x0f0) >>> 4) * 10;
   }

   /**
    * Converts a binary integer byte to its BCD value.
    *
    * @param intVal decimal value to convert to BCD
    *
    * @return the converted BCD integer
    *
    * @see #bcdToInt(int)
    */
   public static byte intToBCD (int intVal)
   {
      if (intVal > 99)
         return 0;

      return ( byte ) (((intVal & 0x0ff) % 10)
                       + (((intVal & 0x0ff) / 10) << 4));
   }
}
