
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;



import java.util.*;


/**
 * This class allows byte wide reads and writes to memory mapped IO devices.
 * <p>
 * Two modes are available: Sequential and FIFO (First In First Out).  Sequential
 * mode is the default and automatically increments the address after each byte
 * when block reads and writes occur.  FIFO access only writes to the specified
 * address, and does not increment the address during block reads and writes.
 * <code>setFIFOMode</code> should be called to set the appropriate mode.
 * After each <code>read</code> or <code>write</code> call, the memory address
 * is returned to the start address designated by <code>setAddress</code>.
 * <p>
 * For slower memory devices, the use of stretch cycles access may be needed.
 * Set this parameter via <code>setStretchCycles</code>.
 *
 * <p>
 * Eight chip enables are available, CE0-CE3, PCE0-PCE3.<br>
 * The memory map is divided into two segments.<br>
 * CE decoded address space spans from 0 - 4MB: <code>0x00000000 - 0x003FFFFF</code><br>
 * PCE decoded address space spans from 8MB - 12MB: <code>0x00800000 - 0x00BFFFFF</code><br>
 * <p>
 *
 * <table BORDER=4>
 * <tr>
 * <td colspan=2 align=center>Expected Transfer Speeds</td>
 * </tr>
 * <tr>
 * <td>Function</td>
 * <td>Speed (bytes/s)</td>
 * </tr>
 * <tr>
 * <td>to CE Sequential</td>   <td align=right>655000</td>
 * </tr>
 * <tr>
 * <td>to PCE Sequential</td>  <td align=right>172000</td>
 * </tr>
 * <tr>
 * <td>to CE FIFO</td>         <td align=right>655000</td>
 * </tr>
 * <tr>
 * <td>to PCE FIFO</td>        <td align=right>172000</td>
 * </tr>
 * <tr>
 * <td>single byte only</td>   <td align=right>770</td>
 * </tr>
 * </table>
 *
 */
public class DataPort
{

   /**
    * Use to set the number of memory bus cycles to 2.
    */
   public static final byte STRETCH0 = 0;

   /**
    * Use to set the number of memory bus cycles to 3.
    */
   public static final byte STRETCH1 = 1;

   /**
    * Use to set the number of memory bus cycles to 4.
    */
   public static final byte STRETCH2 = 2;

   /**
    * Use to set the number of memory bus cycles to 5.
    */
   public static final byte STRETCH3 = 3;

   /**
    * Use to set the number of memory bus cycles to 9.
    */
   public static final byte STRETCH7 = 4;

   /**
    * Use to set the number of memory bus cycles to 10.
    */
   public static final byte STRETCH8 = 5;

   /**
    * Use to set the number of memory bus cycles to 11.
    */
   public static final byte STRETCH9 = 6;

   /**
    * Use to set the number of memory bus cycles to 12.
    */
   public static final byte  STRETCH10  = 7;

   // For the native calls
   private static final byte FUNC_WRITE_FIFO       = 0;
   private static final byte FUNC_READ_FIFO        = 1;
   private static final byte FUNC_WRITE_SEQUENTIAL = 2;
   private static final byte FUNC_READ_SEQUENTIAL  = 3;

   /**
    * Starting memory address for mapped IO access.
    */
   public int address = 0;

   /**
    * Set to <code>true</code> to access FIFO type devices.
    */
   public boolean useFIFOAccess = false;

   /**
    * The last written value.
    */
   public int latchValue = 0;

   /**
    * Number of stretch memory cycles to use when accessing memory mapped IO.
    * Valid values are:<br>
    * <code>STRETCH0</code><br>
    * <code>STRETCH1</code><br>
    * <code>STRETCH2</code><br>
    * <code>STRETCH3</code><br>
    * <code>STRETCH7</code><br>
    * <code>STRETCH8</code><br>
    * <code>STRETCH9</code><br>
    * <code>STRETCH10</code>
    */
   public byte    stretchCycles      = STRETCH0;

   // Arrays to use when reading and writing single bytes.
   private byte[] dataPortReadArray  = new byte [1];
   private byte[] dataPortWriteArray = new byte [1];

   private static native int dataPortNative (int function, int address,
                                             int stretch, int offset,
                                             int length, byte[] data)
      throws IllegalAddressException;

   /**
    * Creates a DataPort object with address of 0x000000.
    */
   public DataPort ()
   {
   }

   /**
    * Creates a DataPort object using the address specified.
    *
    * @param address address location for read/write.
    *
    */
   public DataPort (int address)
   {
      this.address = address;
   }

   /**
    * Reads the current latch value.  This is always the last
    * value written.  To read the current value of the pin,
    * call the <code>read</code> method.
    *
    * @return latch value
    */
   public int readLatch ()
   {
      return latchValue;
   }

   /**
    * Reads a single byte from the <code>address</code>.
    *
    * @return byte read
    * @throws IllegalAddressException if <code>address</code> is out of range
    */
   public int read ()
      throws IllegalAddressException
   {
      synchronized (dataPortReadArray)
      {
         dataPortNative(FUNC_READ_SEQUENTIAL, address, stretchCycles, 0,
                        1, dataPortReadArray);

         return (dataPortReadArray [0] & 0xFF);
      }
   }

   /**
    * Reads bytes into an array starting at the <code>address</code>.
    *
    * @param arr array of bytes
    * @param off starting offset to read into
    * @param len length to read
    *
    * @return number of bytes read
    * @throws IllegalAddressException if <code>address</code> is out of range
    */
   public int read (byte[] arr, int off, int len)
      throws IllegalAddressException
   {
      if (useFIFOAccess)
         dataPortNative(FUNC_READ_FIFO, address, stretchCycles, off, len,
                        arr);
      else
         dataPortNative(FUNC_READ_SEQUENTIAL, address, stretchCycles, off,
                        len, arr);

      return len;
   }

   /**
    * Writes a single byte to the <code>address</code>.
    *
    * @param value single byte to write
    * @throws IllegalAddressException if <code>address</code> is out of range
    */
   public void write (int value)
      throws IllegalAddressException
   {
      synchronized (dataPortWriteArray)
      {
         latchValue = value;

         dataPortWriteArray [0] = ( byte ) value;

         dataPortNative(FUNC_WRITE_SEQUENTIAL, address, stretchCycles, 0,
                        1, dataPortWriteArray);
      }
   }

   /**
    * Writes an array of bytes starting at the <code>address</code>.
    *
    * @param arr array of bytes
    * @param off starting offset to send
    * @param len length to send
    * @throws IllegalAddressException if <code>address</code> is out of range
    */
   public void write (byte[] arr, int off, int len)
      throws IllegalAddressException
   {
      if (len > 0)
        latchValue = (arr[off + len - 1]) & 0xFF;

      if (useFIFOAccess)
         dataPortNative(FUNC_WRITE_FIFO, address, stretchCycles, off, len,
                        arr);
      else
         dataPortNative(FUNC_WRITE_SEQUENTIAL, address, stretchCycles, off,
                        len, arr);
   }

   /**
    * Sets the starting address for read/write.
    *
    * @param address address for read/write
    */
   public void setAddress (int address)
   {
      this.address = address;
   }

   /**
    * Gets the starting address for read/write.
    *
    * @return address for read/write
    */
   public int getAddress ()
   {
      return address;
   }

   /**
    * Sets/Clears the FIFO mode.
    *
    * @param useFIFOAddress <code>false</code> for automatic incrementing of address during
    * multiple reads, or <code>true</code> for FIFO style non-incrementing addressing
    */
   public void setFIFOMode (boolean useFIFOAccess)
   {
      this.useFIFOAccess = useFIFOAccess;
   }

   /**
    * Gets the FIFO mode.
    *
    * @return <code>true</code> if using FIFO mode, <code>false</code>
    *         if using sequential mode
    */
   public boolean getFIFOMode ()
   {
      return useFIFOAccess;
   }

   /**
    * Sets the number of stretch memory cycles to use when accessing memory mapped IO.
    * Valid values are:<br>
    * <code>STRETCH0</code><br>
    * <code>STRETCH1</code><br>
    * <code>STRETCH2</code><br>
    * <code>STRETCH3</code><br>
    * <code>STRETCH7</code><br>
    * <code>STRETCH8</code><br>
    * <code>STRETCH9</code><br>
    * <code>STRETCH10</code>
    *
    * @param stretch stretch cycle value
    * @throws IllegalArgumentException if stretch is out of range
    */
   public void setStretchCycles (byte stretch)
      throws IllegalArgumentException
   {
      if ((stretch >= STRETCH0) && (stretch <= STRETCH10))
         stretchCycles = stretch;
      else
         throw new IllegalArgumentException("Invalid stretch value.");
   }

   /**
    * Gets the number of stretch memory cycles to use when accessing memory mapped IO.
    *
    * @return stretch cycle value
    */
   public int getStretchCycles ()
   {
      return stretchCycles;
   }
}
