
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;



import java.io.IOException;
import java.io.PrintStream;
import com.dalsemi.comm.*;


/**
 * This class encapsulates a small collection of hopefully useful debug
 * utilities. All methods are static to keep performance up and memory usage
 * down. We don't want these things to get in the way of the application we're
 * trying to debug.
  */
public class Debug
{

   /** <code>true</code> if default stream has been set;
       <code>false</code> otherwise. */
   public static boolean         defaultStreams;
   private static TINISerialPort serialPort = null;

   // used by the hexDump methods to hold 2 ascii encoded hex character
   private static byte[] ascii = new byte [2];
   private static byte[] crlf  = { 0x0d, 0x0a };
   private static byte[] space = { ( byte ) ' ' };

   /**
    * Don't allow anyone to instantiate.
    */
   private Debug ()
   {
   }

   /**
    * Dumps hex representation of a four byte integer to
    * <code>System.out</code>.
    *
    * @param i integer to dump
    *
    * @see   #hexDump(byte[])
    * @see   #hexDump(byte[], int)
    * @see   #hexDump(PrintStream, int)
    * @see   #hexDump(PrintStream, byte[])
    */
   public static void hexDump (int i)
   {
      byte[] intBytes = new byte [4];

      intBytes [0] = ( byte ) (((i >>> 0) & 0xFF));
      intBytes [1] = ( byte ) (((i >>> 8) & 0xFF));
      intBytes [2] = ( byte ) (((i >>> 16) & 0xFF));
      intBytes [3] = ( byte ) (((i >>> 24) & 0xFF));

      hexDump(System.out, intBytes);
   }

   /**
    * Dumps hex representation of a four byte integer to the given
    * <code>PrintStream</code>.
    *
    * @param out user supplied output <code>PrintStream</code>
    * @param i   integer to dump
    *
    * @see   #hexDump(int)
    * @see   #hexDump(byte[])
    * @see   #hexDump(byte[], int)
    * @see   #hexDump(PrintStream, byte[])
    */
   public static void hexDump (PrintStream out, int i)
   {
      byte[] intBytes = new byte [4];

      intBytes [0] = ( byte ) (((i >>> 0) & 0xFF));
      intBytes [1] = ( byte ) (((i >>> 8) & 0xFF));
      intBytes [2] = ( byte ) (((i >>> 16) & 0xFF));
      intBytes [3] = ( byte ) (((i >>> 24) & 0xFF));

      hexDump(out, intBytes);
   }

   /**
    * Dumps hex representation of a byte array to <code>System.out</code>.
    *
    * @param  b byte array to dump
    *
    * @throws IOException on failed write
    *
    * @see   #hexDump(int)
    * @see   #hexDump(byte[], int)
    * @see   #hexDump(PrintStream, int)
    * @see   #hexDump(PrintStream, byte[])
    */
   public static void hexDump (byte[] b)
   {
      hexDump(System.out, b);
   }

   /**
    * Dumps hex representation of a byte array to the given
    * <code>PrintStream</code>.
    *
    * @param out user supplied output <code>PrintStream</code>
    * @param b   byte array to dump
    *
    * @throws IOException on failed write
    *
    * @see   #hexDump(int)
    * @see   #hexDump(byte[])
    * @see   #hexDump(byte[], int)
    * @see   #hexDump(PrintStream, int)
    */
   public static void hexDump (PrintStream out, byte[] b)
   {

      // hexDump(b, b.length);
      try
      {
         for (int i = 0; i < b.length; i++)
         {
            ascii [0] = ( byte ) ((b [i] & 0xF0) >>> 4);
            ascii [0] = ( byte ) (ascii [0] > 9 ? ascii [0] + 55
                                                : ascii [0] + 48);
            ascii [1] = ( byte ) (b [i] & 0x0F);
            ascii [1] = ( byte ) (ascii [1] > 9 ? ascii [1] + 55
                                                : ascii [1] + 48);

            out.write(ascii);
            out.write(space);
         }

         out.write(crlf);
      }
      catch (IOException io)
      {
         out.println("IOException:");
         out.println(io.getMessage());
      }
   }

   /**
    * Dumps hex representation of a byte array to
    * <code>System.out</code>.
    *
    * @param b      byte array to dump
    * @param length number of bytes to dump
    *
    * @see   #hexDump(int)
    * @see   #hexDump(byte[])
    * @see   #hexDump(PrintStream, int)
    * @see   #hexDump(PrintStream, byte[])
    */
   public static void hexDump (byte[] b, int length)
   {
      try
      {
         for (int i = 0; i < length; i++)
         {
            ascii [0] = ( byte ) ((b [i] & 0xF0) >>> 4);
            ascii [0] = ( byte ) (ascii [0] > 9 ? ascii [0] + 55
                                                : ascii [0] + 48);
            ascii [1] = ( byte ) (b [i] & 0x0F);
            ascii [1] = ( byte ) (ascii [1] > 9 ? ascii [1] + 55
                                                : ascii [1] + 48);

            System.out.write(ascii);
            System.out.write(space);
         }

         System.out.write(crlf);
      }
      catch (IOException io)
      {
         System.out.println("IOException:");
         System.out.println(io.getMessage());
      }
   }

   /**
    * Dumps a <code>String</code> to the debug port.
    *
    * @param out String to dump.
    *
    * @see #debugDump(int)
    * @see #debugDump(byte[], int)
    * @see #dump(String)
    */
   public static void debugDump (String out)
   {
      dump(out);
   }

   /**
    * Dumps a <code>String</code> to the debug port.
    *
    * @param out String to dump
    *
    * @see #dump(int)
    * @see #dump(byte[], int)
    * @see #debugDump(String)
    */
   public static void dump (String out)
   {
      byte[] b = out.getBytes();

      debugDumpBytesNative(b, b.length);
   }

   /**
    * Dumps a byte array to the debug port.  This will print the ASCII
    * representation of the byte array.
    *
    * @param arr    byte array to dump
    * @param length number of bytes to dump
    *
    * @see #debugDump(int)
    * @see #debugDump(String)
    * @see #dump(byte[],int)
    */
   public static void debugDump (byte[] arr, int length)
   {
      dump(arr, length);
   }

   /**
    * Dumps a byte array to the debug port.  This will print the ASCII
    * representation of the byte array.
    *
    * @param arr    byte array to dump
    * @param length number of bytes to dump
    *
    * @see #dump(int)
    * @see #dump(String)
    * @see #debugDump(byte[], int)
    */
   public static void dump (byte[] arr, int length)
   {
      debugDumpBytesNative(arr, length);
   }

   /**
    * Dumps the least significant byte of an integer to the debug port.
    * This will print the hex representation of the byte.
    *
    * @param b integer to dump
    *
    * @see #debugDump(String)
    * @see #debugDump(byte[], int)
    * @see #dump(int)
    */
   public static void debugDump (int b)
   {
      dump(b);
   }

   /**
    * Dumps the least significant byte of an integer to the debug port.
    * This will print the hex representation of the byte.
    *
    * @param b integer to dump
    *
    * @see #dump(String)
    * @see #dump(byte[], int)
    * @see #debugDump(int)
    */
   public static void dump (int b)
   {
      debugDumpNative(b);
   }

   private static final byte[] intBytes = new byte [4];
   private static final byte[] decBytes = new byte [8];
    
   /**
    * Dumps hex representation of a four byte integer to
    * <code>System.out</code>.
    *
    * @param iVal integer to dump
    *
    * @see   #hexDump(int)
    */
   public static void intDump (int iVal)
   {
      synchronized(intBytes)
      {
//      byte[] intBytes = new byte [4];
//      byte[] decBytes = new byte [8];

        // Chop the integer into bytes
        intBytes [0] = ( byte ) (iVal & 0xFF);
        intBytes [1] = ( byte ) ((iVal >>> 8) & 0xFF);
        intBytes [2] = ( byte ) ((iVal >>> 16) & 0xFF);
        intBytes [3] = ( byte ) ((iVal >>> 24) & 0xFF);

        for (int i = 0; i < 4; i++)
        {
            int adj = ( byte ) (intBytes [i] & 0x0F) > 9 ? 0x37
                                                        : 0x30;

            decBytes [7 - ((i << 1))] = ( byte ) ((intBytes [i] & 0x0F) + adj);
            adj                       = ( byte ) (((intBytes [i] & 0xF0) >>> 4))
                                        > 9 ? 0x37
                                            : 0x30;
            decBytes [6 - (i << 1)]   = ( byte ) (((intBytes [i] & 0xF0) >>> 4)
                                                + adj);
        }

        try
        {
            System.out.write(decBytes);
            System.out.write(crlf);
        }
        catch (IOException io)
        {
            System.out.println("IOException:");
            System.out.println(io.getMessage());
        }
      }
   }

   private static native boolean debugDumpNative (int b);

   private static native void debugDumpBytesNative (byte[] b, int length);

   private static native void setNativeVerboseDebugSpewNative (
      boolean verbose);

   /**
    * Sets the old TINI default streams to Serial port 0 at
    * <code>115200</code>.
    * Binds <code>System.out</code>, <code>System.err</code>, and
    * <code>System.in</code> to <code>serial0</code> at
    * <code>115200</code>.
    */
   public static void setDefaultStreams ()
   {
      setDefaultStreams("serial0", 115200);
   }

   /**
    * Sets the old TINI default streams to the given port.  Binds
    * <code>System.out</code>, <code>System.err</code>, and
    * <code>System.in</code> to the given <code>port</code>.
    *
    * @param port  name of the desired port (i.e. "serial0").
    * @param speed speed of the port. (i.e. 115200)
    */
   public static void setDefaultStreams (String port, int speed)
   {

      //Protect the slush serial session.
      if (com.dalsemi.system.TINIOS.isCurrentTaskInit())
      {
         try
         {
            if (serialPort != null)
               serialPort.close();

            serialPort = new TINISerialPort(port);

            serialPort.setBaudRate(speed);
            System.setIn(serialPort.getInputStream());
            System.setOut(new PrintStream(serialPort.getOutputStream()));
            System.setOut(new PrintStream(serialPort.getOutputStream()));

            defaultStreams = true;
         }
         catch (IOException e)
         {
//            debugDump("IOException opening serial ports in Debug.");
            resetDefaultStreams();
         }
      }
   }

   /**
    * Sets <code>System.out</code>, <code>System.err</code>, and
    * <code>System.in</code> to Null Streams.
    */
   public static void resetDefaultStreams ()
   {
      System.setIn(new NullInputStream());

      PrintStream nullOut = new PrintStream(new NullOutputStream());

      System.setOut(nullOut);
      System.setErr(nullOut);

      defaultStreams = false;

      if (serialPort != null)
      {
         serialPort.close();

         serialPort = null;
      }
   }

   /**
    * For internal use only.
    */
   public static void setNativeVerboseDebugSpew (boolean verbose)
   {
      setNativeVerboseDebugSpewNative(verbose);
   }
}
