
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;


/**
 * Class allowing access to I2C bus.<br>
 * This class uses the I2C protocol on one of two resources:
 * <ul>
 * <li>
 * P5.0(SCL) and P5.1(SDA)<br>
 * Max rate 250Kbit/s, Min rate 2.5Kbit/s dependant on <code>clockDelay</code> value.
 * </li>
 * <li>
 * Any two memory mapped IO areas.<br>
 * Masks are required to select bit or bits within byte for SCL and SDA.<br>
 * Max rate 33Kbit/s, Min rate 2.5Kbit/s dependant on <code>clockDelay</code> value.
 * </li>
 * </ul>
 *
 */
public class I2CPort
{
   private static final byte FUNC_WRITE  = 0;
   private static final byte FUNC_READ   = 1;
   private static final byte I2CDEFAULT  = 0;
   private static final byte I2CMAPPEDIO = 1;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 2.
    */
   public static final byte STRETCH0 = DataPort.STRETCH0;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 3.
    */
   public static final byte STRETCH1 = DataPort.STRETCH1;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 4.
    */
   public static final byte STRETCH2 = DataPort.STRETCH2;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 5.
    */
   public static final byte STRETCH3 = DataPort.STRETCH3;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 9.
    */
   public static final byte STRETCH7 = DataPort.STRETCH7;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 10.
    */
   public static final byte STRETCH8 = DataPort.STRETCH8;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 11.
    */
   public static final byte STRETCH9 = DataPort.STRETCH9;

   /**
    * When in memory map mode, use to set the number of memory bus cycles to 12.
    */
   public static final byte STRETCH10 = DataPort.STRETCH10;

   /**
    * Number of stretch memory cycles to use when accessing memory mapped IO.
    * Valid values are:<br>
    * <code>STRETCH0</code><br>
    * <code>STRETCH1</code><br>
    * <code>STRETCH2</code><br>
    * <code>STRETCH3</code><br>
    * <code>STRETCH7</code><br>
    * <code>STRETCH8</code><br>
    * <code>STRETCH9</code><br>
    * <code>STRETCH10</code>
    */
   public byte stretchCycles = STRETCH0;

   /**
    * I2C address of slave.  Right justified, least significant bit of
    * this byte = least significant bit of I2C slave address.
    */
   public byte slaveAddress = 0;

   /**
    * How long to wait between I2C edge transitions.
    * <br>
    * 1 increment of clockDelay adds approximately 109 nanoseconds between edges.
    */
   public byte  clockDelay = 0;

   private byte I2CType;

   /**
    * Memory mapped address of the I2C clock line.
    */
   public int SCLAddress;

   /**
    * Bit mask for clock line within byte at <code>SCLAddress</code>.
    */
   public byte SCLMask;

   /**
    * Memory mapped address of the I2C data line.
    */
   public int SDAAddress;

   /**
    * Bit mask for data line within byte at <code>SDAAddress</code>.
    */
   public byte SDAMask;

   private static native int i2cportNative (int i2ctype, int function,
                                            int slaveaddress, int clockdelay,
                                            int SCLAddress, int SCLMask,
                                            int SDAAddress, int SDAMask,
                                            int stretchcycles, int offset,
                                            int length, byte[] barr)
      throws IllegalAddressException;

   /**
    * Use processor port pin P5.0 for SCL and P5.1 for SDA.
    */
   public I2CPort ()
   {
      I2CType = I2CDEFAULT;
   }

   /**
    * Use memory mapped IO address for SCL and SDA pins.
    *
    * @param SCLAddress address of latch.
    * @param SCLMask mask of bit or bits to use.
    * @param SDAAddress address of latch.
    * @param SDAMask mask of bit or bits to use.
    *
    */
   public I2CPort (int SCLAddress, byte SCLMask, int SDAAddress, byte SDAMask)
   {
      I2CType         = I2CMAPPEDIO;
      this.SCLAddress = SCLAddress;
      this.SCLMask    = SCLMask;
      this.SDAAddress = SDAAddress;
      this.SDAMask    = SDAMask;
   }

   /**
    * Sets the I2C slave address.
    * This address will take effect on the next read or write.
    * Right justified, least significant bit of
    * this byte = least significant bit of I2C slave address.
    *
    * @param address address of I2C slave device.
    */
   public void setAddress (byte address)
   {
      slaveAddress = address;
   }

   /**
    * Sets how long to wait between I2C edge transitions.
    * <br>
    * 1 increment of clockDelay adds approximately 109 nanoseconds between edges.
    *
    * @param delay I2C Clock delay
    */
   public void setClockDelay (byte delay)
   {
      clockDelay = delay;
   }

   /**
    * Sets the number of stretch memory cycles to use when accessing memory mapped IO.
    * Valid values are:<br>
    * <code>STRETCH0</code><br>
    * <code>STRETCH1</code><br>
    * <code>STRETCH2</code><br>
    * <code>STRETCH3</code><br>
    * <code>STRETCH7</code><br>
    * <code>STRETCH8</code><br>
    * <code>STRETCH9</code><br>
    * <code>STRETCH10</code>
    *
    * @param stretch stretch cycle value
    * @throws IllegalArgumentException if stretch is out of range
    */
   public void setStretchCycles (byte stretch)
      throws IllegalArgumentException
   {
      if ((stretch >= STRETCH0) && (stretch <= STRETCH10))
         stretchCycles = stretch;
      else
         throw new IllegalArgumentException("Invalid stretch value.");
   }

   /**
    * Gets the number of stretch memory cycles to use when accessing memory mapped IO.
    *
    * @return stretch cycle value
    */
   public int getStretchCycles ()
   {
      return stretchCycles;
   }

   /**
    * Writes the byte array to the current slave address
    *
    * @param barr array of bytes to send
    * @param off starting offset to send
    * @param len length of bytes to send
    *
    * @return bytes read, or -1 on error (No ACK from I2C device)
    * @throws IllegalAddressException if <code>SCLAddress</code> or <code>SDAAddress</code> is out of range
    */
   public int write (byte[] barr, int off, int len)
      throws IllegalAddressException
   {
      int retVal;

      if ((len < 0) || (off < 0) || ((len + off) > barr.length))
         throw new ArrayIndexOutOfBoundsException();

      switch (I2CType)
      {

         case I2CDEFAULT :
            retVal = i2cportNative(I2CDEFAULT, FUNC_WRITE, slaveAddress,
                                   clockDelay, 0, 0, 0, 0, 0, off, len, barr);

            return retVal;

         case I2CMAPPEDIO :
            retVal = i2cportNative(I2CMAPPEDIO, FUNC_WRITE, slaveAddress,
                                   clockDelay, SCLAddress, SCLMask,
                                   SDAAddress, SDAMask, stretchCycles, off,
                                   len, barr);

            return retVal;

         default :
            return -1;
      }
   }

   /**
    * Reads into the byte array from the current slave address
    *
    * @param barr array to fill from read operation
    * @param off starting offset to read into
    * @param len length of bytes to read
    *
    * @return bytes read, or -1 on error (No ACK from I2C device)
    * @throws IllegalAddressException if <code>SCLAddress</code> or <code>SDAAddress</code> is out of range
    */
   public int read (byte[] barr, int off, int len)
      throws IllegalAddressException
   {
      int retVal;

      if ((len < 0) || (off < 0) || ((len + off) > barr.length))
         throw new ArrayIndexOutOfBoundsException();

      switch (I2CType)
      {

         case I2CDEFAULT :
            retVal = i2cportNative(I2CDEFAULT, FUNC_READ, slaveAddress,
                                   clockDelay, 0, 0, 0, 0, 0, off, len, barr);

            return retVal;

         case I2CMAPPEDIO :
            retVal = i2cportNative(I2CMAPPEDIO, FUNC_READ, slaveAddress,
                                   clockDelay, SCLAddress, SCLMask,
                                   SDAAddress, SDAMask, stretchCycles, off,
                                   len, barr);

            return retVal;

         default :
            return -1;
      }
   }
}
