
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.system;



import com.dalsemi.shell.*;
import com.dalsemi.shell.server.*;
import java.io.*;
import java.util.*;


/**
 * The <code>TINIOS</code> class contains static methods to read
 * the state of the running tasks and set system configuration variables.
 * The default encoding String is also kept here.
 */
public class TINIOS
{

   // The following states form a bitmask

   /**
    * Network configuration was restored on last reboot.
    */
   public static final int NETWORK_CONFIGURATION_RESTORED = 0x01;

   /**
    * Master erase occurred on last reboot.
    */
   public static final int MASTER_ERASE_OCCURRED = 0x02;

   /**
    * Heap clearing occurred on last reboot.
    */
   public static final int HEAP_CLEAR_OCCURRED = 0x04;

   /**
    * Filesystem modification occurred on last reboot.
    */
   public static final int FS_MODIFICATION_OCCURRED = 0x08;

   /**
    * Flag used with blastHeapOnReboot() to reinitializes heap on next boot.
    */
   public static final int BLAST_HEAP = 0;

   /**
    * Flag used with blastHeapOnReboot() to reinitializes heap and stored system variables on next boot.
    */
   public static final int BLAST_ALL = 1;

   //
   private static final int DISPATCH_EXTERNAL_GET_SERIAL_PORT_ENABLE        =
      0;
   private static final int DISPATCH_EXTERNAL_SET_SERIAL_PORT_ENABLE        =
      1;
   private static final int DISPATCH_EXTERNAL_GET_SERIAL_PORT_SEARCH_ENABLE =
      2;
   private static final int DISPATCH_EXTERNAL_SET_SERIAL_PORT_SEARCH_ENABLE =
      3;
   private static final int DISPATCH_EXTERNAL_GET_SERIAL_PORT_ADDRESS       =
      4;
   private static final int DISPATCH_EXTERNAL_SET_SERIAL_PORT_ADDRESS       =
      5;
   private static final int DISPATCH_IRDA_SET_CLOCK_PIN_STATE               =
      6;
   private static final int DISPATCH_SYSTEM_GET_SERIAL_NUMBER               =
      7;
   private static final int DISPATCH_SYSTEM_GET_HW_VERSION                  =
      8;

   private static final int DISPATCH_GET_LCD_PORT_ADDRESS                   =
      0;
   private static final int DISPATCH_SET_LCD_PORT_ADDRESS                   =
      1;
   private static TINIShell shell                                           =
      new DefaultTINIShell();

   /**
    * Don't let anyone instantiate this class
    */
   private TINIOS ()
   {
   }

   /** Returns the recovery hash. This hash can be used on system recovery
     * as the hash of the root password for login purposes.
     * @param byte array to be filled in with the recovery hash.
     * @return true is a hash has been stored. false otherwise.
     */
   public static native boolean getRecoveryHash (byte[] recoveryHash);

   /** Sets the recovery hash. This hash can be used on system recovery
     * as the hash of the root password for login purposes.
     * @param recovery hash to be stored
     */
   public static native void setRecoveryHash (byte[] recoveryHash);

   /** Returns the restore state from last reboot.
     * @return  an integer representing the extent of any restoration
     *          that was performed at boot time. This number will be
     *          the Inclusive-OR of bit fields containing information
     *          on heap restore state, file system restore state and
     *          network restore state.
     */
   public static native int getBootState ();

   /**
    * Sets the system shell.  You must
    * be admin to call this function, and you can only set a
    * shell if one has not been previously set.
    *
    * @param  newShell  new shell for the system to use
    */
   public static void setShell (TINIShell newShell)
      throws SecurityException
   {
      if (!isCurrentUserAdmin())
      {
         throw new SecurityException("Permission denied.");
      }

      if (newShell != null)
      {
         shell = newShell;
      }
   }

   /**
    * Logs a user into the system and sets
    * his privilege level correctly.  Most shells will have a list
    * of valid users and their corresponding passwords and user id's
    * to consult when this method is called.  (Such as a passwd file.)
    *
    * @param  userName   user name of the user logging in
    * @param  password   password for this user
    *
    * @return user ID of the newly logged in user, or -1 if
    * the userName password combination was not correct
    */
   public static int login (String userName, String password)
   {
      return shell.login(userName, password);
   }

   /**
    * Logs the current user out of the system and returns the
    * privilege level of the current process to the default state.
    *
    * @param  info   any extra information needed in the logout process
    */
   public static void logout (Object info)
   {
      shell.logout(info);
   }

   /**
    * Indicates whether or not the current user has admin privileges.
    *
    * @return  <code>true</code> if the user has admin rights,
    * <code>false</code> otherwise
    */
   public static boolean isCurrentUserAdmin ()
   {
      return shell.isCurrentUserAdmin();
   }

   /**
    * Returns the user ID of the current user.
    *
    * @return  current user's ID
    */
   public static byte getCurrentUID ()
   {
      return shell.getCurrentUID();
   }

   /**
    * Determines if the given user ID has administrative rights.
    *
    * @param uid  the user ID.
    *
    * @return  <code>true</code> if the user is an administrator,
    * <code>false</code> otherwise.
    */
   public static boolean isAdmin (byte uid)
   {
      return shell.isAdmin(uid);
   }

   /**
    * Returns the user name that corresponds to the given user ID.
    *
    * @param  uid  user id to check
    *
    * @return  the user name that corresponds to the given user ID,
    * or null if that ID is not valid.
    */
   public static String getUserNameByUID (byte uid)
   {
      return shell.getUserNameByUID(uid);
   }

   /**
    * Returns the ID of the user with the given user name.
    *
    * @param  username user username to check.
    *
    * @return  the ID of the user with the given user name,
    * or -1 if that user does not exist.
    */
   public static int getUIDByUserName (String username)
   {
      return shell.getUIDByUserName(username);
   }

   /**
    * Returns the name of the current user logged in.
    *
    * @return  name of the current user
    */
   public static String getCurrentUserName ()
   {
      return shell.getCurrentUserName();
   }

   /**
    * Returns the system environment.  (i.e. the environment that
    * all processes see when they start up.
    *
    * @return  a copy of the system environment
    */
   public static Hashtable getSystemEnvironment ()
   {
      return shell.getSystemEnvironment();
   }

   /**
    * Returns a copy of the current environment.
    *
    * @return current environment
    */
   public static Hashtable getCurrentEnvironment ()
   {
      return shell.getCurrentEnvironment();
   }

   /**
    * Gets the value of the given key from the current environment.
    *
    * @param  key   The key to lookup in the current environment.
    *
    * @return  The value of the key, or <code>null</code> if the key
    * is not set in the current environment.
    */
   public static String getFromCurrentEnvironment (String key)
   {
      return shell.getFromCurrentEnvironment(key);
   }

   /**
    * Executes a command in the shell.
    *
    * @param commandLine  An Object array containing the command in the first element,
    * followed by any parameters need for that command in a String[] in the second element.
    * @param in  The stream the command will use to get input.
    * @param out  The stream used to report non-critical messages.
    * @param err  The stream used to report critical messages.
    * @param env  A table of environment variables.
    *
    * @throws Exception Any exception raised by the command.
    */
   public static void execute (Object[] commandLine, SystemInputStream in,
                               SystemPrintStream out, SystemPrintStream err,
                               Hashtable env)
      throws Exception
   {
      shell.execute(commandLine, in, out, err, env);
   }

   /**
    * Returns the name of the shell.
    *
    * @return  The name of the shell.
    */
   public static String getShellName ()
   {
      return shell.getName();
   }

   /**
    * Returns the version of the shell.
    *
    * @return   The shell version.
    */
   public static String getShellVersion ()
   {
      return shell.getVersion();
   }

   /**
    * Returns the current shell.
    *
    * @return current shell
    */
   public static TINIShell getShell ()
   {
      return shell;
   }

   private static native boolean getSlushConsoleOutputNative ();

   private static native void setSlushConsoleOutputNative (boolean set);

   /**
    * Returns the state of bootup messages.
    *
    * @return   true if console output is enabled, false if disabled
    */
   public static boolean isConsoleOutputEnabled ()
   {
      return getSlushConsoleOutputNative();
   }

   /**
    * Sets the state of the bootup messages.
    *
    * @param set true if output should be enabled, false if it should be disabled
    */
   public static void setConsoleOutputEnabled (boolean set)
   {
      if (shell.isCurrentUserAdmin())
         setSlushConsoleOutputNative(set);
   }

   /**
    * Gets the state of the external serial port described by portNum. Currently only ports 2
    * and 3 are supported. Throws <code>IllegalArgumentException</code> for all other
    * specified ports.
    *
    * @param portNum specifies number of external serial port
    *
    * @return true if the external serial port enabled
    */
   public static boolean getExternalSerialPortEnable (int portNum)
   {
      if ((portNum < 2) || (portNum > 3))
         throw new IllegalArgumentException("Invalid port number");

      byte[] dispatchBuffer = new byte [4];

      dispatchBuffer [0] = ( byte ) (portNum & 0x0ff);

      return (tinios_DispatchNative(
         DISPATCH_EXTERNAL_GET_SERIAL_PORT_ENABLE, dispatchBuffer) != 0);
   }

   /**
    * Sets the state of the external serial port. Currently only ports 2
    * and 3 are supported. Throws <code>IllegalArgumentException</code> for all other
    * specified ports.
    *
    * @param portNum specifies number of external serial port
    * @param enable true to enable, false to disable
    */
   public static void setExternalSerialPortEnable (int portNum,
           boolean enable)
   {
      if ((portNum < 2) || (portNum > 3))
         throw new IllegalArgumentException("Invalid port number");

      byte[] dispatchBuffer = new byte [4];

      dispatchBuffer [0] = ( byte ) (portNum & 0x0ff);
      dispatchBuffer [1] = ( byte ) ((enable == true) ? 1
                                                      : 0);

      tinios_DispatchNative(DISPATCH_EXTERNAL_SET_SERIAL_PORT_ENABLE,
                            dispatchBuffer);
   }

   /**
    * Returns true if a search for serial ports 2 and 3 will be performed on boot.
    *
    * @return true if enabled, false if disabled
    */
   public static boolean getExternalSerialPortSearchEnable ()
   {
      return (tinios_DispatchNative(
         DISPATCH_EXTERNAL_GET_SERIAL_PORT_SEARCH_ENABLE, new byte [4]) != 0xff);
   }

   /**
    * Searches for serial ports 2 and 3 on boot if enabled.
    *
    * @param enable true if enabled, false if disabled
    */
   public static void setExternalSerialPortSearchEnable (boolean enable)
   {
      byte[] dispatchBuffer = new byte [4];

      dispatchBuffer [0] = ( byte ) ((enable == true) ? 1
                                                      : 0);

      tinios_DispatchNative(DISPATCH_EXTERNAL_SET_SERIAL_PORT_SEARCH_ENABLE,
                            dispatchBuffer);
   }

   /**
    * Gets the address of the external serial port described by portNum. Currently only ports 2
    * and 3 are supported. Throws <code>IllegalArgumentException</code> for all other
    * specified ports.
    *
    * @param portNum specifies number of external serial port
    * @return address of external serial port
    */
   public static int getExternalSerialPortAddress (int portNum)
   {
      if ((portNum < 2) || (portNum > 3))
         throw new IllegalArgumentException("Invalid port number");

      byte[] dispatchBuffer = new byte [4];

      dispatchBuffer [0] = ( byte ) (portNum & 0x0ff);

      return tinios_DispatchNative(DISPATCH_EXTERNAL_GET_SERIAL_PORT_ADDRESS,
                                   dispatchBuffer);
   }

   /**
    * Sets the address of the external serial port described by portNum. Currently only ports 2
    * and 3 are supported. Throws <code>IllegalArgumentException</code> for all other
    * specified ports.
    *
    * @param portNum specifies number of external serial port
    * @param address specifies address of external serial port
    */
   public static void setExternalSerialPortAddress (int portNum, int address)
   {
      if ((portNum < 2) || (portNum > 3))
         throw new IllegalArgumentException("Invalid port number");

      byte[] dispatchBuffer = new byte [4];

      dispatchBuffer [0] = ( byte ) (portNum & 0x0ff);
      dispatchBuffer [1] = ( byte ) (address & 0x0ff);
      dispatchBuffer [2] = ( byte ) ((address >>> 8) & 0x0ff);
      dispatchBuffer [3] = ( byte ) ((address >>> 16) & 0x0ff);

      tinios_DispatchNative(DISPATCH_EXTERNAL_SET_SERIAL_PORT_ADDRESS,
                            dispatchBuffer);
   }

   /**
    * Turn on or off the IrDA clock pin (P3.5).  This pin oscillates
    * at 16 times the baud rate generator for serial port 1.
    * Note, P3.5 is used for internal one wire and will not work
    * when the IrDA clock pin is running.
    *
    * @param on  true to enable IrDA clock, false to disable
    *
    */
   public static void setIrDAClockPinState (boolean on)
   {
      byte[] dispatchBuffer = new byte [4];

      dispatchBuffer [0] = ( byte ) (on ? 1
                                        : 0);
      dispatchBuffer [1] = 0x00;
      dispatchBuffer [2] = 0x00;
      dispatchBuffer [3] = 0x00;

      tinios_DispatchNative(DISPATCH_IRDA_SET_CLOCK_PIN_STATE,
                            dispatchBuffer);
   }

   /**
    *
    */
   private static native int tinios_DispatchNative (int dispatchId,
           byte[] dataBuffer);

   /**
    *
    */
   private static native void sleepNative (int ms);

   /**
    *
    */
   private static native int wos_killTaskNative (int taskid);

   /**
    *
    */
   private static native int wos_getTaskIDsNative (int[] arr);

   /**
    *
    */
   private static native int wos_getTaskNameNative (int taskid, byte[] arr);

   /**
    *
    */
   private static native int wos_getTaskOwnerNative (int taskid);

   /**
    *
    */
   private static native int wos_getTaskStateNative (int taskid);

   /**
    *
    */
   private static native boolean wos_isTaskRunningNative (int taskid);

   /**
    *
    */
   private static native int wos_getCurrentTaskIDNative ();

   /**
    *
    */
   private static final native int getFreeRAMNative ();

   /**
    *
    */
   private static final native int rebootNative ();

   /**
    *
    */
   private static final native byte sys_getDebugMessageStateNative ();

   /**
    *
    */
   private static final native void sys_setDebugMessageStateNative (
      byte state);

   /**
    *
    */
   private static final native boolean wos_isCurrentTaskInitNative ();

   /**
    *
    */
   private static final native byte[] wos_getTINIOSFirmwareVersionNative ();

   /**
    *
    */
   private static final native void watchdog_setTimeoutNative (int mstimeout);

   /**
    *
    */
   private static native void wos_setTaskOwnerNative (int processID,
           int owner);

   /**
    *
    */
   private static final native void watchdog_feedDogNative ();

   /**
    *
    */
   private static final native void wos_blastHeapOnRebootNative (
      int blastType);

   /**
    *
    */
   private static final native void sys_disablePORNative ();

   /**
    *
    */
   private static final native void getSetTimeZoneNative (boolean set,
           byte[] zone);

   /**
    * Sets the TINI Watchdog timeout. If timer is allowed to run without being reset (fed)
    * system will reboot.
    *
    * @param mstimeout - timeout, in milliseconds.  Don't set this too low,
    *                    because Java ain't that quick!  Set to zero to turn off
    *                    the watchdog.
    */
   public static void setWatchdogTimeout (int mstimeout)
   {
      watchdog_setTimeoutNative(mstimeout);
   }

   /**
    * Resets (feeds) the Watchdog timer.  This must be called frequently at some interval less
    * than the watchdog timeout value to prevent the system from rebooting. Watchdog timers are
    * useful for recovering the system when a critical process goes awry.
    */
   public static void feedWatchdog ()
   {
      watchdog_feedDogNative();
   }

   /**
    * Gets a string representing the TINI OS Firmware Version.
    *
    * @return version string
    */
   public static String getTINIOSFirmwareVersion ()
   {
      return new String(wos_getTINIOSFirmwareVersionNative());
   }

   /**
    * Gets a string representing the TINI HW Version.
    *
    * @return version string, or null if version is not known
    */
   public static String getTINIHWVersion ()
   {
      byte[] barr = new byte[255];
      int retVal = tinios_DispatchNative(DISPATCH_SYSTEM_GET_HW_VERSION,barr);

      if (retVal > 0)
      {
        return new String(barr,0,retVal);
      }
      else
        return null;
   }

   /**
    * Gets the hardware serial number.  This is the address of the DS2502
    * configuration memory.
    *
    * @return HW serial number
    */
   public static final String getTINISerialNumber()
   {
      byte[] barr = new byte[8];
      tinios_DispatchNative(DISPATCH_SYSTEM_GET_SERIAL_NUMBER,barr);
      return com.dalsemi.onewire.utils.Address.toString(barr);
   }

   /**
    * Sets the current timezone.
    * If you want to set the time to an offset from GMT, pass in a
    * string of the form "TOffset" - Example "-10" sets the timezone
    * to an offset of GMT-10.
    * @param zone timezone string
    */
   public static void setTimeZone (String zone)
      throws IllegalArgumentException
   {
      TimeZone newTZ;
      if (zone.startsWith("+") || zone.startsWith("-"))
      {

        if (zone.startsWith("+"))
        {
           zone = zone.substring(1);
        }
        newTZ = (TimeZone)(new SimpleTimeZone(((new Integer(zone)).intValue()*3600*1000), zone));
      }
      else
      {
        newTZ = TimeZone.getTimeZone(zone);
      }

      if (newTZ == null)
      {
         throw new IllegalArgumentException("Invalid Timezone.");
      }

      //Set the given Timezone as the default timezone.
      TimeZone.setDefault(newTZ);

      Properties props = System.getProperties();

      props.put("user.timezone", zone);
      System.setProperties(props);

      getSetTimeZoneNative(true, zone.getBytes());
   }

   /**
    * Gets a string representing the current Timezone.
    *
    * @return timezone string
    */
   public static String getTimeZone ()
   {
      byte[] temp = new byte [3];


      getSetTimeZoneNative(false, temp);

      if ((temp [0] == 0) && (temp [1] == 0) && (temp [2] == 0))
      {

         return TimeZone.getDefault().getID();
      }
      else
      {

         return new String(temp);
      }
   }

   /**
    * Returns true if the current process is the init process (HEX file in bank 7).
    *
    * @return true if the current process is the init process (HEX file in bank 7)
    *
    * @see Thread.sleep()
    */
   public static boolean isCurrentTaskInit ()
   {
      return wos_isCurrentTaskInitNative();
   }

   /**
    * Delays a process for <code>ms</code> milliseconds.
    *
    * @param ms the amout of time to sleep
    *
    * @see Thread.sleep()
    */
   public static final void sleepProcess (int ms)
   {
      sleepNative(ms);
   }

   /**
    * Gets this tasks ID
    *
    * @return my Task ID
    */
   public static int getTaskID ()
   {
      return wos_getCurrentTaskIDNative();
   }

   /**
    * Allows root to protect the 1st two processes
    * from being killed by non admin users.  Process 1 is
    * the GC, process 2 is the user program in bank 7.
    * You must be admin to call this function.
    */
   public static void lockInitProcesses ()
   {
      if (shell.isCurrentUserAdmin())
      {
         wos_setTaskOwnerNative(1, shell.getCurrentUID());
         wos_setTaskOwnerNative(2, shell.getCurrentUID());
      }
   }

   /**
    * Gets the owner of specified task ID.
    *
    *
    * @param taskid
    * @return my task ID
    */
   public static int getOwnerIDByTaskID (int taskid)
   {
      return wos_getTaskOwnerNative(taskid);
   }

   /**
    * Checks to see if task with specified ID is running
    *
    * @param taskid ID of task
    *
    * @return true if task is running
    */
   public static boolean isTaskRunning (int taskid)
   {
      return wos_isTaskRunningNative(taskid);
   }

//   private static int[]  ids  = new int [16];
//   private static byte[] name = new byte [256];
   private static int[]  ids;
   private static byte[] name;

   private static void verifyStatics()
   {
      if (ids == null)
        ids = new int[16];
      if (name == null)
        name = new byte[256];
   }

   /**
    * Returns the task table.
    *
    * @return string array with size equal to number of tasks
    */
   public static String[] getTaskTable ()
   {
      String[] retval;

      verifyStatics();

      synchronized (name)
      {
         synchronized (ids)
         {
            int numids = wos_getTaskIDsNative(ids);
            int namelength;

            retval = new String [numids];

            String tempstr;

            for (int i = 0; i < numids; i++)
            {
               namelength = wos_getTaskNameNative(ids [i], name);
               retval [i] = Integer.toString(ids [i]) + ": "
                            + new String(name, 0, namelength);
            }
         }
      }

      return retval;
   }

   /**
    * Returns the task table ID list.
    *
    * @return int array with size equal to number of tasks
    */
   public static int[] getTaskTableIDs ()
   {
      int[] retval;

      verifyStatics();

      synchronized (ids)
      {
         int numids = wos_getTaskIDsNative(ids);

         retval = new int [numids];

         System.arraycopy(ids, 0, retval, 0, numids);
      }

      return retval;
   }

   /**
    * Stops and removes task from system.
    *
    * @param taskID Task ID of task to kill
    *
    * @return task ID killed
    */
   public static int killTask (int taskID)
   {
      return (wos_killTaskNative(taskID));
   }

   /**
    * Gets the free RAM.
    *
    * @return available free RAM
    */
   public static final int getFreeRAM ()
   {
      return getFreeRAMNative();
   }

   /**
    * Performs an orderly system reboot.
    */
   public static final void reboot ()
   {
      rebootNative();
   }

   /**
    * Gets the state of serial boot up messages.
    *
    * @return true or false state of boot messages
    * @deprecated
    */
   public static final boolean getSerialBootMessagesState ()
   {
      return false;
   }

   static final byte SERIAL_DEBUG_MESSAGE_MASK  = 0x01;
   static final byte BITBANG_DEBUG_MESSAGE_MASK = 0x02;

   /**
    * Sets the state of serial boot up messages.
    *
    * @param on sets to true to turn on boot messages, false to turn off
    */
   public static final void setSerialBootMessagesState (boolean on)
   {
      byte state = sys_getDebugMessageStateNative();

      state = (on) ? (state &= ~SERIAL_DEBUG_MESSAGE_MASK)
                   : (state |= SERIAL_DEBUG_MESSAGE_MASK);

      sys_setDebugMessageStateNative(state);
   }

   /**
    * Sets the state of debug messages.
    *
    * @param on sets to true to turn on debug messages, false to turn off
    */
   public static final void setDebugMessagesState (boolean on)
   {
      byte state = sys_getDebugMessageStateNative();

      state = (on) ? (state &= ~BITBANG_DEBUG_MESSAGE_MASK)
                   : (state |= BITBANG_DEBUG_MESSAGE_MASK);

      sys_setDebugMessageStateNative(state);
   }

   /**
    * Gets the address of the LCD port.
    *
    * @return - the address of the LCD Port - defaults to 0x00380000
    */
   public static int getLCDAddress ()
   {
      return tinios_LCDAddress(DISPATCH_GET_LCD_PORT_ADDRESS, 0);
   }

   /**
    * Sets the address of the LCD Port
    * @param - the address of the LCD port
    */
   public static void setLCDAddress (int address)
   {
      tinios_LCDAddress(DISPATCH_SET_LCD_PORT_ADDRESS, address);
   }

   /**
    */
   private static native int tinios_LCDAddress (int funcNum, int address);

   /**
    * Gets the approximate number of milliseconds the system has
    * been running.  This number is guaranteed to be less than or
    * equal to the actual uptime.  It will never be larger.
    *
    * @return number of milliseconds since last boot
    */
   public static native final long uptimeMillis ();

   /**
    * Calling this function will cause the heap to be
    * destroyed (reinitialized) on the next boot.
    * All data in RAM will be destroyed.  Initialization
    * of system variables is conditional.
    *
    * WARNING: This is designed to be used when the application
    * the user wishes to run is already in flash, and does not
    * care if the filesystem is destroyed.
    *
    * NOTE: System variables are kept in a separate portion
    * of the RAM, and are only destroyed with the proper parameter
    * to this function.
    *
    * Calling this function only sets the trigger for the next boot.
    * The trigger is disengaged immediately after it fires after boot.
    * Every subsequent boot will be unaffected, unless the user
    * calls this function again.
    *
    * @param blastType - BLAST_HEAP for heap reinit only
    *                    BLAST_ALL for heap reinit and stored system variable init
    */
   public static final void blastHeapOnReboot (int blastType)
   {
      wos_blastHeapOnRebootNative(blastType);
   }

   /**
    * Disables the ability to recover after a power fail
    * in a critical, atomic heap modification.
    *
    * DANGER: Do not turn this off unless you know what
    * you are doing.  All heap data could become corrupted.
    *
    * The benefit to disabling power fail recovery is an
    * improved speed of the new opcode, as well as a faster
    * garbage collector.  Performance should increase
    * about 30% on these operations.
    */
   public static final void disablePowerFailRecovery ()
   {
      sys_disablePORNative();
   }

   /**
    * Enables serial port 1, this will cause serial 1 to take over the interrupt.
    * Disables one-wire to the DS2480.
    * @deprecated
    */
   public static final void enableSerialPort1()
   {
      enableSer1(true);
   }

   /**
    * Enables or disables serial port 1.  Enabling serial port 1 causes the
    * external 1-Wire to be disabled.  Disabling serial port 1 enables external
    * 1-Wire.
    * @param true to enabled serial port 1, false will disable serial port 1
   */
   public static final void enableSerialPort1(boolean enable)
   {
      enableSer1(enable);
   }

   private static native void enableSer1 (boolean enable);

   /**
    *  Enables RTS/CTS hardware flow control for the given
    *  serial port.  RTS/CTS hardware flow control is only
    *  supported on one serial port at a time.  If
    *  RTS/CTS flow control is enabled for a different
    *  port than the port indicated, an UnsupportedCommOperationException will
    *  be thrown.  If harware flow control is enabled, DTR and DCD can not be
    *  used for BitPort.
    *  This method will always return true for serial2 and serial3.
    *  @param portNumber - the number of the port number to enable RTS/CTS flow
    *                      control
    *  @param enable - enables or disables the RTS/CTS hardware flow control for the given
    *                  port
    *  @return true if the operation was successful, false if the hardware
    *           version of TINI does not support RTS/CTS hardware flow control
    *  @throws javax.comm.UnsupportedCommOperationException
    */
   public native static boolean setRTSCTSFlowControlEnable(int portNumber, boolean enable)
   throws javax.comm.UnsupportedCommOperationException;

   /**
    *  Returns the state of the RTS/CTS flow control for the given port.
    *  @param portNumber - the number of the port to get RTS/CTS flow control
    *                      information.
    *  @return true if RTS/CTS flow control is enabled.
    */
   public native static boolean getRTSCTSFlowControlEnable(int portNumber);


   /*private static native boolean isSer1Enabled();

   public static boolean isSerialPort1Enabled()
   {
       return isSer1Enabled();
   }
   */
//   public static native int GetTempVar();
}
