/*
 * Copyright (C) 1997-2000 Dallas Semiconductor Corporation.
 * All rights Reserved. Printed in U.S.A.
 * This software is protected by copyright laws of
 * the United States and of foreign countries.
 * This material may also be protected by patent laws of the United States
 * and of foreign countries.
 * This software is furnished under a license agreement and/or a
 * nondisclosure agreement and may only be used or copied in accordance
 * with the terms of those agreements.
 * The mere transfer of this software does not imply any licenses
 * of trade secrets, proprietary technology, copyrights, patents,
 * trademarks, maskwork rights, or any other form of intellectual
 * property whatsoever. Dallas Semiconductor retains all ownership rights.
 */

package com.dalsemi.tininet;

import java.io.*;
import com.dalsemi.fs.*;
import com.dalsemi.system.*;
import java.net.*;
/**
 * This class is used to set up parameters for TINI networking.
 *
 * @version 1.0
 * @author  Stephen Hess
 */

public class TININet
{
    /** Network configuration is not committed to flash memory.
      */
    public static final int UNCOMMITTED      = 1;
    /** Network configuration is committed to flash memory.
      */
    public static final int COMMITTED        = 2;
    /** Network configuration restore is disabled.
      */
    public static final int RESTORE_DISABLED = 3;

    private static final String TININET_FILE = "/etc/.tininet";
    private static final String HOSTNAME     = "HostName:";
    private static final String DOMAINNAME   = "DomainName:";

    private static final String NOT_SET      = "";

    private static String hostname           = "TINI";
    private static String domainname         = NOT_SET;

    private static boolean initialized       = false;

    //Options for native get/set method
    private static final int ETHERNET        =  0;
    private static final int PRI_DNS         =  1;
    private static final int SEC_DNS         =  2;
    private static final int DHCP_SERVER     =  3;
    private static final int PROXY_SERVER    =  4;
    private static final int MAILHOST        =  5;
    private static final int DNS_TIMEOUT     =  6;
    private static final int PROXY_PORT      =  7;

    private static final int GATEWAY_IP      =  0;
    private static final int MY_IP           =  1;
    private static final int SUBNET_MASK     =  2;

    private static final Object lock = new Object();

    /**
     * Forces a reload of network settings from the configuration file /etc/.tininet
     */
    public static void update()
    {
        initialize(true);
    }

    private static void initialize(boolean force)
    {
        if (initialized && !force)
            return;

        DSFile tininet = new DSFile(TININET_FILE);
        if (!tininet.exists())
        {
            writeDataFile();
        }
        else
        {
            synchronized (lock)
            {
                //Read the values...
                try
                {
                    DataInputStream tininetIn = new DataInputStream(new BufferedInputStream(new FileInputStream(tininet)));

                    String line = null;
                    while ((line = tininetIn.readLine()) != null)
                    {
                        if (line.startsWith(HOSTNAME))
                        {
                            hostname = line.substring(9);
                        }
                        else if (line.startsWith(DOMAINNAME))
                        {
                            domainname = line.substring(11);
                        }
                    }

                    tininetIn.close();
                }
                catch(FileNotFoundException fnfe)
                {
                    //com.dalsemi.system.Debug.debugDump("Could not open TININet data file.");
                    return;
                }
                catch(IOException ioe)
                {
                    //com.dalsemi.system.Debug.debugDump("IOException reading TININet data file.");
                    return;
                }
            }
        }

        initialized = true;
    }

    /**
     * Checks to determine if the user has the rights to set network settings.
     * NOTE: If slush is not running, this method always returns true.
     *
     * @return true is user has access rights
     */
    private static boolean checkAccessRights()
    {
        return TINIOS.isCurrentUserAdmin();
    }

    private static boolean optionNotSet(byte[] checkThis)
    {
        boolean set = false;

        for (int i=0; i<checkThis.length; i++)
        {
            if (checkThis[i] != 0)
                set = true;
        }

        return !set;
    }

    private static String createIPString(byte[] fromThis)
    {
        String ip = NOT_SET + (fromThis[0] & 0x0FF);
        for (int i=1; i< fromThis.length; i++)
            ip += "." + (fromThis[i] & 0x0FF);
        return ip;
    }

    private static byte[] createIPFromString(String fromThis)
    {
        if (fromThis.equals(NOT_SET))
            return new byte[4];

        byte[] toHere = null;
        try
        {
            int pos = 0;
            int oldPos = 0;
            int index = 0;

            while (pos != -1)
            {
                pos = fromThis.indexOf(".", oldPos);
                oldPos = pos + 1;
                if (pos != -1)
                    index++;
            }

            toHere = new byte[index + 1];

            index = 0;
            pos = 0;
            oldPos = 0;

            pos = fromThis.indexOf(".", oldPos);
            while (pos != -1)
            {
                toHere[index++] = (byte)Integer.parseInt(fromThis.substring(oldPos, pos));
                oldPos = pos + 1;
                pos = fromThis.indexOf(".", oldPos);
            }
            toHere[index++] = (byte)Integer.parseInt(fromThis.substring(oldPos));
        }
        catch(Exception e)
        {
            //com.dalsemi.system.Debug.debugDump("Exception in TININet.createIPFromString: " + e.toString());
        }

        return toHere;
    }

    private static boolean writeDataFile()
    {
        try
        {
            DSFile etc = new DSFile("/etc");
            if (!etc.exists())
            {
                etc.mkdir();
                etc.setOtherPermissions(0x05);
            }

            synchronized (lock)
            {
                DSFile tininet = new DSFile(TININET_FILE);
                FileOutputStream tininetStream = new FileOutputStream(tininet);
                PrintStream tininetOut = new PrintStream(new BufferedOutputStream(tininetStream));
                tininetOut.println(HOSTNAME + hostname);
                tininetOut.println(DOMAINNAME + domainname);
                tininetOut.close();
                tininet.setOtherPermissions(0x04);
            }
        }
        catch(IOException ioe)
        {
            //com.dalsemi.system.Debug.debugDump("IOException writing TININet data file.");
            return false;
        }

        return true;
    }

    /**
     * Gets the hostname of TINI.
     *
     * @return TINI hostname
     */
    public static String getHostname()
    {
        initialize(false);

        return hostname;
    }

    /**
     * Gets the domain name of TINI.
     *
     * @return TINI domain name
     */
    public static String getDomainname()
    {
        initialize(false);

        return domainname;
    }

    /**
     * Gets the primary DNS server address.
     *
     * @return primary DNS server address, if the address has not been set an empty string is returned
     */
    public static String getPrimaryDNS()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, PRI_DNS, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the secondary DNS server address.
     *
     * @return secondary DNS server address, if the address has not been set an empty string is returned
     */
    public static String getSecondaryDNS()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, SEC_DNS, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the IP address of the default mailhost.
     *
     * @return IP address of the default mailhost, if the address has not been set an empty string is returned
     */
    public static String getMailhost()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, MAILHOST, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the timeout value for DNS queries.
     *
     * @return timeout in milliseconds
     */
    public static int getDNSTimeout()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, DNS_TIMEOUT, false);

        return intFromByteArray(scratchPad, 0);
    }

    /**
     * Gets the IP address of the DHCP server.
     * @param address byte array of length four to be filled with the IP DHCP server address, most significant
     * byte of address is stored at index zero
     */
    public static void getDHCPServerIP(byte[] address)
    {
        getSetGeneralNetworkOptionNative(address, DHCP_SERVER, false);
    }

    /**
     * Gets the IP address of DHCP server.
     *
     * @return IP address of the dhcp server in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getDHCPServerIP()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, DHCP_SERVER, false);

        if (optionNotSet(scratchPad))
        {
            return NOT_SET;
        }
        else
        {
            return createIPString(scratchPad);
        }
    }

    /**
     * Gets the IP address of the specified interface. Currently loopback (lo), ethernet (eth0) and ppp are
     * currently supported.
     *
     * @param  interfaceName  network interface (lo, eth0 or pppn)
     * @param  address        byte array of length four to be filled in with the IP address, most significant
     * byte of address is stored at index zero
     */
    public static void getIPAddress(String interfaceName, byte[] address)
    {
        getSetNetworkOptionNative(interfaceName.getBytes(), address, MY_IP, false);
    }

    /**
     * Gets the IP address of the specified interface as a String. Currently loopback (lo),
     * ethernet (eth) and ppp are supported
     *
     * @param  interfaceName network interface (lo, eth or ppp)
     *
     * @return the IP address of the named interface,
     * if the address has not been set an empty string is returned
     */
    public static String getIPAddress(String interfaceName)
    {
        byte[] scratchPad = new byte[4];

        getSetNetworkOptionNative(interfaceName.getBytes(), scratchPad, MY_IP, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the IP address of the default interface as a String.
     *
     * @return the IP address in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getIPAddress()
    {
        byte[] scratchPad = new byte[4];

        getSetNetworkOptionNative(null, scratchPad, MY_IP, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the IP address of the default interface.
     *
     * @param address  byte array of length four to be filled with the IP address, most significant
     * byte of address is stored at index zero
     */
    public static void getIPAddress(byte[] address)
    {
        getSetNetworkOptionNative(null, address, MY_IP, false);
    }

    /**
     * Gets the gateway IP address of the default interface.
     *
     * @param address byte array of length four to be filled with the IP address, most significant
     * byte of address is stored at index zero
     */
    public static void getGatewayIP(byte[] address)
    {
        getSetNetworkOptionNative(null, address, GATEWAY_IP, false);
    }

    /**
     * Gets the gateway IP address of the default interface as a String.
     * @return  ip address of the gateway in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getGatewayIP()
    {
        byte[] scratchPad = new byte[4];

        getSetNetworkOptionNative(null, scratchPad, GATEWAY_IP, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the IP address of the gateway for the specified interface.
     *
     * @param interfaceName network interface (lo, eth or ppp)
     * @param address byte array of length four to be filled with the IP address, most significant
     * byte of address is stored at index zero
     */
    public static void getGatewayIP(String interfaceName, byte[] address)
    {
        getSetNetworkOptionNative(interfaceName.getBytes(), address, GATEWAY_IP, false);
    }

    /**
     * Gets the IP address of the gateway for the specified interface as a
     * string in dotted decimal notation.
     *
     * @return IP address of the gateway in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getGatewayIP(String interfaceName)
    {
        byte[] scratchPad = new byte[4];

        getSetNetworkOptionNative(interfaceName.getBytes(), scratchPad, GATEWAY_IP, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the subnet mask of the default interface.
     *
     * @param address byte array of length four to be filled with the subnet mask, most significant
     * byte of mask is stored at index zero
     */
    public static void getSubnetMask(byte[] address)
    {
        getSetNetworkOptionNative(null, address, SUBNET_MASK, false);
    }

    /**
     * Gets the subnet mask of the default interface as a String.
     *
     * @return the subnet mask in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getSubnetMask()
    {
        byte[] scratchPad = new byte[4];

        getSetNetworkOptionNative(null, scratchPad, SUBNET_MASK, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the subnet mask of the specified interface.
     *
     * @param interfaceName network interface (lo, eth or ppp)
     * @param address byte array of length four to be filled with the subnet mask, most significant
     * byte of mask is stored at index zero
     */
    public static void getSubnetMask(String interfaceName, byte[] address)
    {
        getSetNetworkOptionNative("eth0".getBytes(), address, SUBNET_MASK, false);
    }

    /**
     * Gets the subnet mask of the specified interface as a String.
     *
     * @return the subnet mask in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getSubnetMask(String interfaceName)
    {
        byte[] scratchPad = new byte[4];

        getSetNetworkOptionNative(interfaceName.getBytes(), scratchPad, SUBNET_MASK, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the ethernet MAC address into the given byte array, with the most significant byte first.
     *
     * @param address byte array of length six to be filled with the ethernet address
     */
    public static void getEthernetAddress(byte[] address)
    {
        getSetGeneralNetworkOptionNative(address, ETHERNET, false);
    }

    /**
     * Gets the ethernet MAC address in dotted decimal notation.
     *
     * @return the ethernet address in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getEthernetAddress()
    {
        byte[] ip = new byte[6];

        getSetGeneralNetworkOptionNative(ip, ETHERNET, false);

        if (optionNotSet(ip))
            return NOT_SET;
        else
        {
            StringBuffer sb = new StringBuffer();
            for (int i=0;i<6;i++)
            {
                sb.append(Integer.toHexString(ip[i] & 0x0ff));
                if (i!=5) sb.append(':');
            }
            return sb.toString();
        }
    }

    /**
     * Gets the IP address of the proxy server.
     *
     * @return the IP address of the proxy server in dotted decimal notation,
     * if the address has not been set an empty string is returned
     */
    public static String getProxyServer()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, PROXY_SERVER, false);

        if (optionNotSet(scratchPad))
            return NOT_SET;
        else
            return createIPString(scratchPad);
    }

    /**
     * Gets the proxy server port number.
     *
     * @return the proxy port number
     */

    public static int getProxyPort()
    {
        byte[] scratchPad = new byte[4];

        getSetGeneralNetworkOptionNative(scratchPad, PROXY_PORT, false);

        return intFromByteArray(scratchPad, 0);
    }

    /**
     * Sets the default network interface for use with establishing a connection with an
     * address that doesn't match current known networks.
     *
     * @param defInterface network interface lo(loopback), eth(ethernet) or ppp
     *
     * @return true if successful
     */
    private static boolean setDefaultInterface(String defInterface)
    {
        try
        {
            setDefaultInterface(defInterface.getBytes());
        }
        catch(Exception e)
        {
            return false;
        }
        return true;
    }

    /**
     * Sets the host name to the given string.
     *
     * @param host hostname
     *
     * @return true if successful
     */
    public static boolean setHostname(String host)
    {
        initialize(false);

        if (!checkAccessRights())
            return false;

        hostname = host;

        return writeDataFile();
    }

    /**
     * Sets the domain name to the given string.
     *
     * @param   domain domain name to be set
     *
     * @return true if successful
     */
    public static boolean setDomainname(String domain)
    {
        initialize(false);

        if (!checkAccessRights())
            return false;

        domainname = domain;

        return writeDataFile();
    }

    /**
     * Sets the primary DNS IP address to the given string.
     *
     * @param   primaryDNS  primary DNS IP address in dotted decimal notation
     *
     * @return true if successful
     */
    public static boolean setPrimaryDNS(String primaryDNS)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetGeneralNetworkOptionNative(createIPFromString(primaryDNS), PRI_DNS, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the secondary DNS IP address to the given string.
     *
     * @param secondDNS secondary DNS IP address in dotted decimal notation
     *
     * @return true if successful
     */
    public static boolean setSecondaryDNS(String secondDNS)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetGeneralNetworkOptionNative(createIPFromString(secondDNS), SEC_DNS, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the mail host to the given IP address.
     *
     * @param mailhost IP address of the mail host in dotted decimal notation
     *
     * @return true on success
     */
    public static boolean setMailhost(String mailhost)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            //Allow DNS on setMailhost if possible:
            getSetGeneralNetworkOptionNative(InetAddress.getByName(mailhost).getAddress(), MAILHOST, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the DNS timeout in milliseconds.
     * If the timeout value is set to zero, a fallback and retry procedure
     * will take effect.  The DNS client will retry starting at 2 seconds
     * and doubling until reaching 16 seconds.  The DNS resolver will try
     * four times in this period to resolve the name.
     * * 2s * 4s * 8s * 16s
     *
     * @param dnsTimeout timeout value in milliseconds, zero for fallback and retry
     *
     * @return true on success
     */

    public static boolean setDNSTimeout(int dnsTimeout)
    {
        if (!checkAccessRights())
            return false;

        byte[] scratchPad = new byte[4];

        try
        {
            intToByteArray(scratchPad, 0, dnsTimeout);
            getSetGeneralNetworkOptionNative(scratchPad, DNS_TIMEOUT, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the IP address of the DHCP server to the given string.
     *
     * @param dhcpServer DHCP server IP address in dotted decimal notation
     *
     * @return true on success
     */
    public static boolean setDHCPServerIP(String dhcpServer)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetGeneralNetworkOptionNative(createIPFromString(dhcpServer), DHCP_SERVER, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the IP address of the default interface.
     *
     * @param localIP IP address, MSB of address is index zero
     *
     * @return true on success
     */
    public static boolean setIPAddress(byte[] localIP)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(null, localIP, MY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the IP address of the default interface.
     *
     * @param localIP IP address in dotted decimal notation
     *
     * @return true on success
     */
    public static boolean setIPAddress(String localIP)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(null, createIPFromString(localIP), MY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the IP address of the specified interface.
     *
     * @param interfaceName network interface name
     *
     * @param localIP IP address, MSB of address is index zero
     *
     * @return true on success
     */
    public static boolean setIPAddress(String interfaceName, byte[] localIP)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(interfaceName.getBytes(), localIP, MY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the IP address of the specified interface.
     *
     * @param interfaceName network interface name
     *
     * @param localIP IP address in dotted decimal notation
     *
     * @return true on success
     */
    public static boolean setIPAddress(String interfaceName, String localIP)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(interfaceName.getBytes(), createIPFromString(localIP), MY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the gateway address for the default interface.
     *
     * @param gateway IP address of the gateway, MSB of address is index zero
     *
     * @return true on success
     */
    public static boolean setGatewayIP(byte[] gateway)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(null, gateway, GATEWAY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the gateway IP for the default interface.
    *
    * @param gateway IP address of the gateway in dotted decimal notation
    *
    * @return true on success
    */
    public static boolean setGatewayIP(String gateway)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(null, createIPFromString(gateway), GATEWAY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Sets the gateway IP address for the specified interface.
     *
     * @param interfaceName network interface name
     * @param gateway ip address of the gateway, MSB of address is index zero
     *
     * @return true on success
     */
    public static boolean setGatewayIP(String interfaceName, byte[] gateway)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(interfaceName.getBytes(), gateway, GATEWAY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the gateway IP for the specified interface.
    *
    * @param interfaceName network interface name
    *
    * @param gateway IP address of the gateway in dotted decimal notation
    *
    * @return true on success
    */
    public static boolean setGatewayIP(String interfaceName, String gateway)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(interfaceName.getBytes(), createIPFromString(gateway), GATEWAY_IP, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the subnet mask of the default interface.
    *
    * @param subnetMask subnet mask, MSB of mask is at index zero
    *
    * @return true on success
    */
    public static boolean setSubnetMask(byte[] subnetMask)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(null, subnetMask, SUBNET_MASK, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the subnet mask of the default interface.
    *
    * @param subnetMask subnet mask in dotted decimal notation
    *
    * @return true on success
    */
    public static boolean setSubnetMask(String subnetMask)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(null, createIPFromString(subnetMask), SUBNET_MASK, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the subnet mask of the specified interface.
    *
    * @param interfaceName network interface name
    *
    * @param subnetMask subnet mask, MSB of mask is at index zero
    *
    * @return true on success
    */
    public static boolean setSubnetMask(String interfaceName, byte[] subnetMask)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(interfaceName.getBytes(), subnetMask, SUBNET_MASK, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the subnet mask of the specified interface.
    *
    * @param interfaceName network interface name
    *
    * @param subnetMask subnet mask in dotted decimal notation
    *
    * @return true on success
    */
    public static boolean setSubnetMask(String interfaceName, String subnetMask)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetNetworkOptionNative(interfaceName.getBytes(), createIPFromString(subnetMask), SUBNET_MASK, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the proxy server IP address
    *
    * @param proxyServer IP address of the proxy server, MSB of address is at index zero
    *
    * @return true on success
    */
    public static boolean setProxyServer(String proxyServer)
    {
        if (!checkAccessRights())
            return false;

        try
        {
            getSetGeneralNetworkOptionNative(InetAddress.getByName(proxyServer).getAddress(), PROXY_SERVER, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets the proxy port.
    *
    * @param proxyPort the port number to be used for proxy
    *
    * @return true on success
    */
    public static boolean setProxyPort(int proxyPort)
    {
        if (!checkAccessRights())
            return false;

        byte[] scratchPad = new byte[4];

        try
        {
            intToByteArray(scratchPad, 0, proxyPort);
            getSetGeneralNetworkOptionNative(scratchPad, PROXY_PORT, true);
        }
        catch(Exception e)
        {
            return false;
        }

        return true;
    }

   /**
    * Sets multiple network configuration settings.
    *
    * @param dhcp IP address of the DHCP server, MSB of address is at index zero
    * @param ip IP address of the local machine, MSB of address is at index zero
    * @param gateway IP address of gateway address, MSB of address is at index zero
    * @param subnet subnet mask, MSB of address is at index zero
    * @param dns1 IP address of the primary DNS server, MSB of address is at index zero
    * @param dns2 IP address of the secondary DNS server, MSB of address is at index zero
    * @param domain domain name stored as a byte array
    * @param mailhost IP address of the outgoing mail server, MSB of address is at index zero
    *
    * @return true on success
    */

     public static boolean setOptions(byte[] dhcp, byte[] ip, byte[] gateway,
                                     byte[] subnet, byte[] dns1, byte[] dns2,
                                     byte[] domain, byte[] mailhost)
    {
        initialize(false);

        if (!checkAccessRights())
            return false;

        try
        {
            getSetGeneralNetworkOptionNative(dhcp, DHCP_SERVER, true);

            getSetNetworkOptionNative("eth0".getBytes(), ip, MY_IP, true);

            getSetNetworkOptionNative("eth0".getBytes(), gateway, GATEWAY_IP, true);

            getSetNetworkOptionNative("eth0".getBytes(), subnet, SUBNET_MASK, true);

            getSetGeneralNetworkOptionNative(dns1, PRI_DNS, true);

            getSetGeneralNetworkOptionNative(dns2, SEC_DNS, true);

            setMailhost(new String(mailhost));

        }
        catch(Exception e)
        {
            return false;
        }
        if (domain != null)
            domainname = new String(domain);

        return writeDataFile();
    }

    static void intToByteArray(byte[] outArray, int start, int value)
    {
        //BigEndian
        outArray[start]     = (byte)((value & 0xFF000000) >>> 24);
        outArray[start + 1] = (byte)((value & 0x00FF0000) >>> 16);
        outArray[start + 2] = (byte)((value & 0x0000FF00) >>>  8);
        outArray[start + 3] = (byte) (value & 0x000000FF);
    }

    static int intFromByteArray(byte[] inArray, int start)
    {
        return (((int)inArray[start    ] << 24) & 0xFF000000 |
                ((int)inArray[start + 1] << 16) & 0x00FF0000 |
                ((int)inArray[start + 2] <<  8) & 0x0000FF00 |
                ((int)inArray[start + 3]      ) & 0x000000FF);
    }

    /**
     * Gets the current connection table.
     *
     * @return byte array containing the connection table
     */
    public static byte[] getConnectionTable()
    {
        return getConnectionTableNative();
    }

    /**
     * Gets the current ARP cache table.
     *
     * @return byte array containing the ARP cache table
     */
    public static byte[] getARPCacheTable()
    {
        //This should look more like getConnection table but for now
        //since these native methods exist...
        byte[] arpCache = new byte[getARPCacheSizeNative()];
        getARPCacheTableNative(arpCache);

        return arpCache;
    }

    /**
     * Commits the current network configuration to flash.
     *
     * @exception CommitException if state has already been committed.
     */

    public static void commitNetworkState()
      throws CommitException
    {
      int retVal = commitNetState(0x55);

      switch(retVal)
      {
        case 0:
        return;
        case 1:
          throw new CommitException("Restore disabled");
        case 2:
          throw new CommitException("Network configuration has already been committed");
        case 3:
          throw new CommitException("Bad commit flag");
        case 4:
          throw new CommitException("Write failed");
        default:
          throw new CommitException();
      }
    }
    /**
     *  Returns the commit state as UNCOMMITED, COMMITTED or RESTORE_DISABLED
     *
     * @return the current commit state
     */

    public static int getNetworkCommitState()
    {
      switch(getNetCommitState())
      {
        case 0x0ff:
        return UNCOMMITTED;

        case 0x055:
        return COMMITTED;

        default:
        return RESTORE_DISABLED;
      }
    }
    /**
     *  Disables the ability to commit or restore network configuration
     * until the bank 7 application is re-flashed.
     *
     */
    public static void disableNetworkRestore()
    {
      commitNetState(0);
    }

    private static native int getARPCacheSizeNative();
    private static native void getARPCacheTableNative(byte[] arpCache);
    private static native byte[] getConnectionTableNative();
    private static native void getSetGeneralNetworkOptionNative(byte[] newData, int option, boolean set);
    private static native void getSetNetworkOptionNative(byte[] interfaceName, byte[] newData, int option, boolean set);

    /**
     * Sets the default network interface.
     *
     * @param name network interface, lo (loopback), eth (ethernet) or ppp
     *
     * @return true if inteface was successfully set as the default
     */

    public static native boolean setDefaultInterface(byte[] name);

    /** Gets interface information for given interface and stores it in byte array. Returns true
      * if information was found using supplied interface number.
      *
      * @param interfaceNum network interface number
      *
      * @param data buffer for interface info
      *
      * @return true if interface info was found
      */

    public static native boolean getInterfaceInfo(int interfaceNum, byte[] data);

    /** Adds an interface to the network interface list with the given parameters.
      *
      * @param name network interface lo(loopback), eth(ethernet) or ppp
      * @param ipAddr IP address of the new interface
      * @param subnet subnet mask of the new interface
      * @param gateway gateway IP address of the new interface
      * @param flags network configuration flags mapped as follows:
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.0 = used - should always be set
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.1 = active - should be set if the interface is to be active
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.2 = default - set if this should be the default network interface
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.3 = UNUSED - must be cleared
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.4 = routable - set if interface is to carry routable packets
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.5 = multicasting enabled - set if interface supports multicast
      * <br>&nbsp&nbsp&nbsp&nbsp&nbsp flags.6 = UNUSED - must be cleared
      * @param type type of network interface ethernet = 1, ppp = 2, loopback = 3
      * @param transmitter 24 bit transmitter routine address represented as a 32-bit int
      *
      */

    public static native void addInterfaceEntry(byte[] name, byte[] ipAddr, byte[] subnet,
                                                byte[] gateway, byte flags, byte type, int transmitter);

    /** Removes specified interface from the network interface list.
      *
      * @param name network interface name to remove
      */
    public static native void removeInterfaceEntry(byte[] name);

    private static native int commitNetState(int flag);
    private static native int getNetCommitState();
}
