
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.dhcp;

import java.io.*;
import java.net.*;
import com.dalsemi.system.*;
import com.dalsemi.tininet.*;
import java.util.*;


/**
 * The DHCPClient supports dynamically obtaining a new IP Address or reclaiming an 
 * IP Address lost on reboot from a DHCP server.  
 * Both constructors are passed a DHCPListener so that DHCP status messages can be 
 * relayed. After the DHCPClient object is constructed, the process in invoked through 
 * the start() method. When the user no longer needs the IP Address it is recommended 
 * that the stopDHCP method is called to release the IP to the server and 
 * close the socket.
 *<p>
 * Refer to the following RFCs:
 *<pre>            951 - BOOTSTRAP PROTOCOL (BOOTP)
 *           1048 - BOOTP Vendor Information Extensions
 *           2131 - Dynamic Host Configuration Protocol
 *           2132 - DHCP Options and BOOTP Vendor Extentions</pre>
 * <A HREF="http://www.ietf.org/rfc.html"> RFC repository http://www.ietf.org/rfc.html </A>
 * @author CM
 * @version %I%, %G%
 */
public class DHCPClient extends Thread
{
   int            state;
   int            retries    = 0;
   int            count      = 0;
   long           T1, T2;
   byte[]         ip         = new byte [4];
   byte[]         ServerIP   = new byte [4];
   byte[]         data       = new byte [548];
   short          index      = 0;
   boolean        dhcprun    = true;
   boolean        resend     = false;
   static Object  lock       = new Object();   // lock for network synchonization
   DatagramSocket sock;
   Vector         dhcpListeners = new Vector(3);   //Pick a small initial vector size

   private native static boolean isDHCPRunningNative ();

   private native static void setDHCPRunningNative (boolean set);

   /**
    * Constructs a dhcp object and intializes for obtaining a new IP.
    * 
    *
    * @param DL  DHCPListener to report DHCP status messages
    *
    * @see DHCPListener.java
    */
   public DHCPClient (DHCPListener DL)
      throws IllegalStateException
   {

      if (isDHCPRunningNative())
      {
         throw new IllegalStateException("DHCP is already running.");
      }

      try
      {

         // open a socket to port 68
         if (sock == null)
            sock = new DatagramSocket(68);

         // start in init state (no previous ip)
         state = 1;

         dhcpListeners.addElement(DL);
         setDHCPRunningNative(true);
      }
      catch (SocketException e)
      {
      }
   }

   /**
    * Constructs a dhcp object and intializes to attempt to rebind to previous 
    * owned IP that was lost in a reboot.
    *
    *
    * @param DL DHCPListener
    * @param serverIP IP of DHCP server where previous IP was obtained
    * @param localIP  IP to obtain
    *
    */
   public DHCPClient (DHCPListener DL, byte[] serverIP, byte[] localIP)
      throws IllegalStateException
   {
      if (isDHCPRunningNative())
      {
         throw new IllegalStateException("DHCP is already running.");
      }

      // init reboot constructor (previous ip, want to try to regain)
      try
      {
         if ((serverIP == null) || (localIP == null))
         { // should inform ip lost
              state = 1;
              notifyListenersIPLost();
         }
         else
         {
            System.arraycopy(serverIP, 0, ServerIP, 0, 4);
            System.arraycopy(localIP, 0, ip, 0, 4);

            state = 7;
         }
         // open a socket to port 68
         if (sock == null)
            sock = new DatagramSocket(68);

         dhcpListeners.addElement(DL);
         setDHCPRunningNative(true);
      }
      catch (SocketException e)
      {
        //drain
      }
   }

   /**
    * Starts dhcp thread.
    *
    *
    */
   public void run ()
   {
      while (dhcprun)
      {
         switch (state)
         {

            case 1 :
               state = init();
               break;

            case 2 :
               state = selecting();
               break;

            case 3 :
               state = requesting();
               break;

            case 4 :
               state = bound();
               break;

            case 5 :
               state = renewing();
               break;

            case 6 :
               state = rebinding();
               break;

            case 7 :
               state = initReboot();
               break;

            case 8 :
               state = rebooting();
               break;

            case -1 :
               // call a shutdown to release ip if we have one
               count = retries = 0;
               state = 1;
               break;

            default :
               notifyListenersIPError(toString(11));
               dhcprun = false;
         }
      }

      // close socket
      if (sock != null)
         sock.close();
   }

   /**
    * Stops dhcp thread, releases ip, and closes any open sockets.
    *
    */
   public void stopDHCPThread ()
   {
      try
      {
         // protect so we don't step on the bound code toes
         synchronized (lock)
         {

            // can't tell if the socket is open or closed
            // so must close it and then open a new one to
            // guarantee we have a open socket
            dhcprun = false;

            if (sock != null)
            {
               sock.close();
            }

            sock = new DatagramSocket(68);

            // Send the message out to release current ip if we have one
            releaseIP();
            sock.close();
            setDHCPRunningNative(false);
            this.interrupt();
         }
      }
      catch (SocketException e){}
   }

   private int init ()
   {
      T1 = T2 = 0;
      backOff(1);
      DatagramPacket dp;

      // If it gets kicked back to init from other states >5x send error and continue
      if (retries >= 5)
      {

         notifyListenersIPError(toString(1));

         return -1;
      }

      try
      {
         // build DHCPOFFER Msg
         buildBasicMsg(0x01);

         dp = new DatagramPacket(data, data.length, InetAddress.getByName("255.255.255.255"), 67);

         sock.send(dp);

         return 2;   // goto Selecting state*/
      }
      catch (UnknownHostException e)
      {
         notifyListenersIPError(toString(9));
         return -1;
      }
      catch (InterruptedIOException e)
      {
         return 1;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }

   private int selecting ()
   {
      int i, j;
      byte[] recvdata = new byte[548];
      DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);   // recv
      DatagramPacket dp;      // Field offer, if timeout resend packet
      byte[]         domainName = null;
      byte[]         gateway    = new byte [4];
      byte[]         netmask    = new byte [4];
      byte[]         dns1       = new byte [4];
      byte[]         dns2       = new byte [4];
      byte[]         mailhost   = new byte [4];


      try
      {
         if (count >= 5)
         {
            notifyListenersIPError(toString(2));
            return -1;
         }

         sock.setSoTimeout(3000);
         sock.receive(rdp);

         // if trans id does not match or not an offer
         if (!((recvdata [4] == data [4]) && (recvdata [5] == data [5])
              && (recvdata [6] == data [6]) && (recvdata [7] == data [7])) ||
              (messageType(recvdata)!=0x02))
         {
            count++;
            return 2;                 // selecting
         }

         // copy yiaddr and siaddr into request packet
         for (i = 16; i < 24; i++)
            data [i] = recvdata [i];

         // DHCPREQUEST
         data [242] = 0x03;

         // get gateway, lease times, ip, server ip
         getOpts(recvdata, gateway, netmask, dns1, dns2, domainName, mailhost);

         System.arraycopy(recvdata, 16, ip, 0, 4);

         // accept first offer unless fields are bad, continue to field more offers
         if ((ServerIP[0]==0) || (T1 == 0) || (T2 == 0) || (ip[0] == 0))
         {

            count++;
            return 2;                 // INIT
         }

         index--;                     // write over ending 0xff

         // copy requested ip from yiaddr into option field
         // request ip offered
         data [index++] = 0x32;
         data [index++] = 0x04;

         for (i = 16; i < 20; i++)
            data [index++] = recvdata [i];

         // copy server ip into options
         data [index++] = 0x36;
         data [index++] = 0x04;

         for (i = 0; i < 4; i++)
            data [index++] = ServerIP [i];

         // finish packet
         data [index++] = ( byte ) 0xff;

         System.arraycopy(recvdata, 16, ip, 0, 4);

         // construct packet
         dp = new DatagramPacket(data, data.length,
                              InetAddress.getByName("255.255.255.255"), 67);

         sock.send(dp);

         count = 0;

         return 3;
      }
      catch (UnknownHostException e)
      {

         notifyListenersIPError(toString(9));

         return -1;
      }
      catch (InterruptedIOException e)
      {

         // go back to init stage to resend UDP
         retries++;
         backOff(retries);
         return 1;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }

   private int requesting ()
   {
      byte[] recvdata = new byte[548];
      DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);
      DatagramPacket dp;

      byte[]         domainName = null;
      byte[]         dns1       = new byte [4];
      byte[]         dns2       = new byte [4];
      byte[]         gateway    = new byte [4];
      byte[]         netmask    = new byte [4];
      byte[]         mailhost   = new byte [4];

      try
      {
         if (count >= 5)
         {
            notifyListenersIPError(toString(3));
            return -1;
         }
         if (resend == true)
         {
            String serverIPS = "";
            for (int i = 0; i < 4; i++)
            {
                serverIPS += Integer.toString(ServerIP [i] & 0xFF);

                if (i < 3)
                    serverIPS += '.';
            }
            dp = new DatagramPacket(data, data.length, InetAddress.getByName(serverIPS), 67);
            sock.send(dp);
            resend = false;
         }
         // time out in 3 secs, assume package lost
         sock.setSoTimeout(3000);
         sock.receive(rdp);

         int len = rdp.getLength();


         if (((recvdata [4] == data [4]) && (recvdata [5] == data [5])
              && (recvdata [6] == data [6]) && (recvdata [7] == data [7])) ||
              (messageType(recvdata)==0x05))
         {
            getOpts(recvdata, gateway, netmask, dns1, dns2, domainName, mailhost);
            // set network params we extracted in selecting

            
            TININet.setOptions(ServerIP, ip, gateway, netmask, dns1, dns2,
                            domainName, mailhost);
            notifyListenersIPLeased();
            //return 4;
         }
         else
         {
            count++;
            return 3;
         }

         // Arp ip to make sure no one owns it.
         // T1 and T2 are already set go and spin
         return 4;                    // go to bound state

         /*
          * }
          * else
          * {
          * //send decline
          * data[242] = 0x04;
          * dp.setAddress(InetAddress.getByName(serverIPS));
          * sock.send(dp);
          * retries++;
          * return 1; //go to init state
          * }
          */

         // return 4;
      }
      catch (UnknownHostException e)
      {

         notifyListenersIPError(toString(9));
         return -1;
      }
      catch (InterruptedIOException e)
      {
         backOff(count);
         count++;
         return 3;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }

   private int bound ()
   {
      DatagramPacket dp;
      try
      {
         sock.close();
         count = retries = 0;

         this.sleep(T1 * 1000);   // need to be *1000 for release
         T1 = 0;

         synchronized (lock)
         {

            // open socket again
            sock = new DatagramSocket(68);

            // set message type to request
            data [242] = 0x03;
            // Send out renew request

            String serverIPS = "";
            for (int i = 0; i < 4; i++)
            {
                serverIPS += Integer.toString(ServerIP [i] & 0xFF);
                if (i < 3)
                    serverIPS += '.';
            }
            dp = new DatagramPacket(data, data.length,InetAddress.getByName(serverIPS), 67);

            sock.send(dp);

            return 5;                // go to renew state
         }
      }
      catch (InterruptedException e)
      {
         T1 = 0;
         return 4;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }

   private int renewing ()
   {
      byte[] recvdata = new byte[548];
      int j;

      DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);
      DatagramPacket dp;

      try
      {
         if ((T2 - System.currentTimeMillis())/1000 < 4)
         {
            // Send out renew request
            dp = new DatagramPacket(data, data.length, InetAddress.getByName("255.255.255.255"), 67);
            sock.send(dp);
            return 6;
         }

         if (count >= 2)
         {
            notifyListenersIPLost();
            return 1;
         }

         if (resend == true)
         {
            String serverIPS = "";
            for (int i = 0; i < 4; i++)
            {
                serverIPS += Integer.toString(ServerIP [i] & 0xFF);

                if (i < 3)
                serverIPS += '.';
            }
            dp = new DatagramPacket(data, data.length, InetAddress.getByName(serverIPS), 67);
            sock.send(dp);
            resend = false;
         }
         // time out in 3 secs, assume package lost
         sock.setSoTimeout(3000);
         sock.receive(rdp);

         if ((recvdata [4] == data [4]) && (recvdata [5] == data [5])
              && (recvdata [6] == data [6]) && (recvdata [7] == data [7]))
         {
            int type = messageType(recvdata);

            if (type == 0x05)
            {
                // Reset T1 and T2
                resetTimes(recvdata);

                if ((T1 > 0) && (T2 > 0))
                {
                    notifyListenersIPRenewed();
                    return 4;
                }
                else
                {
                    // times are hosed - see if there is another offer
                    count++;
                    return 5;
                }
            }

            // NAK
            else if (type == 0x06)
            {
                notifyListenersIPLost();
                return 1;
            }
         }
         count++;
         return 5;


      }
      catch (UnknownHostException e)
      {
         notifyListenersIPError(toString(9));
         return -1;
      }
      catch (InterruptedIOException e)
      {

         count++;
         backOff(count);
         resend = true;
         return 5;

      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }
   // Timer 2 has expired, one last valient effort to get renew ip, by broadcast sent from renew
   private int rebinding ()
   {
       byte[] recvdata = new byte[548];
       DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);
       try
       {
           sock.setSoTimeout(3000);
           sock.receive(rdp);

           // ACK
           if (!((recvdata [4] == data [4]) && (recvdata [5] == data [5])
              && (recvdata [6] == data [6]) && (recvdata [7] == data [7])) ||
              (messageType(recvdata)!=0x05))
           {
               resetTimes(recvdata);
               if ((T1 > 0) && (T2 > 0))
               {
                   notifyListenersIPRenewed();
                   return 4;
               }
           }
           // notify and start back in init
           notifyListenersIPLost();
           return 1;
      }
      catch (InterruptedIOException e)
      {
         notifyListenersIPLost();
         return 1;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         notifyListenersIPLost();
         return -1;
      }
   }

   private int initReboot ()
   {

      // read stored ip and dhcp server ip
      //
      DatagramPacket dp;

      backOff(1);

      try
      {
         buildBasicMsg(0x03);

         index--;    // write over ending 0xff

         // copy ip from ipconfig
         data [index++] = 0x32;
         data [index++] = 0x04;

         for (int i = 0; i < 4; i++)
            data [index++] = ( byte ) ip [i];

         data [index++] = ( byte ) 0xff;

         // create datagram packet with stored data and send to dhcpserver
         String serverIPS = "";
         for (int i = 0; i < 4; i++)
         {
            serverIPS += Integer.toString(ServerIP [i] & 0xFF);

            if (i < 3)
               serverIPS += '.';
         }

         dp = new DatagramPacket(data, data.length, InetAddress.getByName(serverIPS), 67);

         sock.send(dp);

         return 8;   //rebooting
      }
      catch (UnknownHostException e)
      {

         notifyListenersIPError(toString(9));
         return -1;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }

   private int rebooting ()
   {
      byte[] recvdata = new byte[548];
      byte[]         domainName = null;
      byte[]         dns1       = new byte [4];
      byte[]         dns2       = new byte [4];
      byte[]         gateway    = new byte [4];
      byte[]         netmask    = new byte [4];
      byte[]         mailhost   = new byte [4];

      try
      {
         DatagramPacket rdp = new DatagramPacket(recvdata, recvdata.length);

         if (count > 1)
         {
            notifyListenersIPError(toString(8));
            return -1;
         }

         sock.setSoTimeout(3000);
         sock.receive(rdp);

         if ((recvdata [4] == data [4]) && (recvdata [5] == data [5])
              && (recvdata [6] == data [6]) && (recvdata [7] == data [7]))
         {
            int type = messageType(recvdata);

            if (type == 0x05)
            {

                getOpts(recvdata,  gateway, netmask, dns1, dns2, domainName, mailhost);

                
                TININet.setOptions(ServerIP, ip, gateway, netmask, dns1, dns2,
                            domainName, mailhost);


                notifyListenersIPLeased();

                return 4;
            }
            else if (type == 0x06)
            {
                notifyListenersIPLost();
                count = 0;
                return 1;
            }
        }
        count++;
        return 7;
      }
      catch (InterruptedIOException e)
      {
         count++;
         backOff(count);
         return 7;
      }
      catch (IOException e)
      {
         notifyListenersIPError(toString(10));
         return -1;
      }
   }

   private void buildBasicMsg (int type)
   {
      index = 0;
      byte[] temp = new byte [6];
      String hostname = TININet.getHostname();
      int len = hostname.length();

      // build dhcp request
      data [index++] = 1;      // op
      data [index++] = 1;      // htype
      data [index++] = 0x06;   // hlen
      data [index++] = 0x00;   // hops

      // 4 bytes of xid to least 4 bytes of mac id
      TININet.getEthernetAddress(temp);

      for (int i = 2; i < 6; i++)
         data [index++] = temp [i];

      data [index++] = 0x00;             //
      data [index++] = 0x00;

      // set first byte of flags (broadcast bit)
      data [index++] = ( byte ) 0x80;

      for (int i = 0; i < 17; i++)
        data[index++] = 0x00;   // zero fill

      // set mac addr
      for (int i = 0; i < 6; i++)
      {
         data [index++] = ( byte ) temp [i];
      }

      for (int i = 0; i < 202; i++)
        data[index++] = 0x00;   // zero fill

      // options field - DHCP format
      data [index++] = 0x63;
      data [index++] = ( byte ) 0x82;
      data [index++] = 0x53;
      data [index++] = 0x63;

      // DHCP message type DHCPDiscover
      data [index++] = 0x35;
      data [index++] = 0x01;
      data [index++] = ( byte ) type;

      if (type != 0x07)                    // not a release message
      {
         // add hostname
         data [index++] = (byte) 0x0C;
         data [index++] = (byte) len;
         for (int i = 0; i < len; i++)
         {
            data [index++] = (byte) hostname.charAt(i);
         }

         // subnet mask and gateway request
         data [index++] = ( byte ) 0x37;
         data [index++] = 0x05;
         data [index++] = 0x01;            //subnet
         data [index++] = 0x03;            //gateway
         data [index++] = 0x45;            //smtp servers
         data [index++] = 0x06;            //dns servers
         data [index++] = ( byte ) 0x0F;   //domain name
      }

      data [index++] = ( byte ) 0xFF;   // end message
   }

   // method to get dhcp message type and to make sure message is ours
   private int messageType (byte[] recvdata)
   {
    
        int i;
        // start the search after standard dhcp options
        i = 240;


        while (recvdata [i] != ( byte ) 0xff)
        {
            if (recvdata [i] == 0x35)
                return (recvdata [i + 2]);

            i += recvdata [i + 1] + 2;
       }

       return (-1);
   }

   // loop through and find lease times, gateway, subnet mask, serverip
   private void getOpts (byte[] recvdata, byte[] gateway, byte[] netmask,
                         byte[] dns1, byte[] dns2, byte[] domainName, byte[] mailhost)
   {

      int j, i;
      long lease = 0;
      // Find Lease time
      i = 243;   // start the search after standard dhcp options
      while (recvdata [i] != ( byte ) 0xff)
      {
         switch (recvdata [i])
         {

            // ip lease time
            case 0x33 :
            {
               lease = 0;

               for (j = 0; j < 4; j++)
               {
                  lease = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j)))
                          | lease;
               }

               break;
            }

            // T1
            case 0x3A :
            {
               T1 = 0;

               for (j = 0; j < 4; j++)
               {
                  T1 = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j))) | T1;
               }

               T1 &= 0x7FFFFFFF;

               break;
            }

            // T2
            case 0x3B :
            {
               T2 = 0;

               for (j = 0; j < 4; j++)
               {
                  T2 = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j))) | T2;
               }

               T2      &= 0x7FFFFFFF;
               //sleepT2 = ( int ) (T2 / 4);

               break;
            }

            // server ip
            case 0x36 :
            {
               System.arraycopy(recvdata, i + 2, ServerIP, 0, 4);
               break;
            }

            // gateway
            case 0x03 :
            {
               System.arraycopy(recvdata, i + 2, gateway, 0, 4);
               break;
            }

            // subnet mask
            case 0x01 :
            {
               System.arraycopy(recvdata, i + 2, netmask, 0, 4);
               break;
            }

            // domain name
            case 0x0F :
            {
               byte[] domain = new byte [recvdata [i + 1]];
               System.arraycopy(recvdata, i + 2, domain, 0, recvdata [i + 1]);
               //domainName = new String(domain);
               break;
            }

            // smtp servers
            case 0x45 :
            {
               System.arraycopy(recvdata, i + 2, mailhost, 0, recvdata [i + 1]);
               break;
            }

            // dns servers
            case 0x06 :
            {
               // primary dns
               System.arraycopy(recvdata, i + 2, dns1, 0, 4);

               // secondary dns
               if (recvdata [i + 1] > 4)
                    System.arraycopy(recvdata, i + 2, dns2, 0, 4);

               break;
            }
         }

         i += recvdata [i + 1] + 2;

      }
      if (((T1 == 0) || (T2 == 0)) && (lease > 0))
      {
         T1 = ( int ) ((lease / 2) & 0x7FFFFFFF);
         T2 = ( int ) (((lease * 85) / 10) & 0x7FFFFFFF);
      }
      if (T2 > 0)
        T2 = System.currentTimeMillis() + T2*1000;

   }

   /* Function used to reset timers after a lease renewal*/
   private void resetTimes (byte[] recvdata)
   {
      int i, j;
      T1 = T2 = 0;
      long lease = 0;
      // Reset T1 and T2
      i = 243;   // start the search after standard dhcp options

      while (recvdata [i] != ( byte ) 0xff)
      {
         switch (recvdata [i])
         {

            // ip lease time
            case 0x33 :
            {
               lease = 0;

               for (j = 0; j < 4; j++)
               {
                  lease = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j)))
                          | lease;
               }

               break;
            }

            // T1
            case 0x3A :
            {
               for (j = 0; j < 4; j++)
               {
                  T1 = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j))) | T1;
               }

               T1 &= 0x7FFFFFFF;

               break;
            }

            // T2
            case 0x3B :
            {
               for (j = 0; j < 4; j++)
               {
                  T2 = ((recvdata [i + j + 2] & 0xFF) << (8 * (3 - j))) | T2;
               }

               T2      &= 0x7FFFFFFF;

               break;
            }
            default :
               break;
         }

         i += recvdata [i + 1] + 2;
      }

      if (((T1 == 0) || (T2 == 0)) && (lease > 0))
      {
         T1 = ( int ) ((lease / 2) & 0x7FFFFFFF);
         T2 = ( int ) (((lease * 85) / 10) & 0x7FFFFFFF);
      }
      if (T2 > 0)
        T2 = System.currentTimeMillis() + T2*1000;
   }

   private void backOff (int bcount)
   {
      try
      {
         int    timeout = 0;
         Random r       = new Random();

         timeout = r.nextInt();
         timeout &= 0x3F;

         bcount &= 0x07;

         if (bcount < 4)
            timeout %= (bcount * 7);
         else                     // keep timeouts around 1 minute
         {
            timeout %= (60);

            if (timeout < 40)
               timeout += 40;
         }
         this.sleep(timeout * 1000);
      }
      catch (InterruptedException e){}
   }

   private void releaseIP ()
   {
      DatagramPacket dp;

      try
      {
         buildBasicMsg(0x07);

         String serverIPS = "";

         for (int i = 0; i < 4; i++)
         {
            serverIPS += Integer.toString(ServerIP [i] & 0xFF);

            if (i < 3)
               serverIPS += '.';
         }

         dp = new DatagramPacket(data, data.length, InetAddress.getByName(serverIPS), 67);
         sock.send(dp);
         sock.close();
      }
      catch (InterruptedIOException e)
      {
      }
      catch (IOException e)
      {
      }
   }
   /**
    * Converts dhcp integer error to error String.
    *
    *
    * @param err  error returned from dhcp
    *
    * @return String associated with error, ie socket error, no reply in init, etc.
    *
    */
   private String toString (int err)
   {

      // Error Codes to be returned in error method of DHCPListener
      switch (err)
      {

         case 0 :
            return "Socket Error";

         case 1 :
            return "No reply in Init";

         case 2 :
            return "No reply in Selecting";

         case 3 :
            return "No reply in Requesting";

         case 4 :
            return "No reply in Bound";

         case 5 :
            return "No reply in Renewing";

         case 6 :
            return "No reply in Rebinding";

         case 7 :
            return "No reply in Init Reboot";

         case 8 :
            return "No reply in Rebooting";

         case 9 :
            return "Unknown Host Exception";

         case 10 :
            return "IOException";

         case 11 :
            return "Fatal Error";

         default :
            return "Unknown Error";
      }
   }

   /**
    * Adds listener for accepting dhcp status messages.
    *
    *
    * @param newListener DHCPListener for accepting status messages
    *
    */
   public void addDHCPListener (DHCPListener newListener)
   {
      dhcpListeners.addElement(newListener);
   }

   /**
    * Removes listener, to discontinue acceptance of status messages.
    *
    *
    * @param thisListener DHCPListener to be removed from list
    *
    */
   public void removeDHCPListener (DHCPListener thisListener)
   {
      dhcpListeners.removeElement(thisListener);
   }

   private void notifyListenersIPError (String message)
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipError(message);
      }
   }

   private void notifyListenersIPLost ()
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipLost();
      }
   }

   private void notifyListenersIPRenewed ()
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipRenewed();
      }
   }

   private void notifyListenersIPLeased ()
   {
      for (int i = 0; i < dhcpListeners.size(); i++)
      {
         (( DHCPListener ) dhcpListeners.elementAt(i)).ipLeased();
      }
   }
}
    
