
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.icmp;



import java.net.InetAddress;


/**
 * <P>
 * Ping provides a TINI Java application the ability to send ICMP echo request
 * and read the raw ICMP response. When a machine receives an ICMP echo request
 * from a remote host it responds with an ICMP echo reply message that contains
 * an exact copy of the request packet.
 * </P>
 * <P>
 * Ping allows an application to programmatically determine whether
 * a particular service running on a remote host has died or 
 * whether the host machine itself has become unreachable on the network.
 * </P>
 *
 * <H2>Usage</H2>
 * Send a single ICMP echo request to a remote host
 * <PRE><CODE>
 *     InetAddress addr = InetAddress.getByName("www.dalsemi.com");
 *     boolean response = Ping.pingNode(addr);
 *     System.out.println(response ? "Got reply from node" : "Node was unreachable");
 * </CODE></PRE>
 *
 */
public class Ping
{

   /** Field DEFAULT_TTL           */
   public static final byte DEFAULT_TTL = ( byte ) 0xFF;

   /**
    * Don't allow instantiation.
    */
   private Ping ()
   {
   }

   /**
    * Transmit multiple ICMP echo request packets to a node.
    *
    * @param addr ip address to be pinged
    * @param count number of times to ping
    *
    * @return the number of packets received
    *
    * @see java.net.InetAddress
    */
   public static int pingNode (InetAddress addr, int count)
   {
      byte[] ipAddress       = addr.getAddress();
      int    packetsReceived = 0;
      byte[] response        = new byte [256];

      for (int i = 0; i < count; i++)
         if (pingNative(ipAddress, DEFAULT_TTL, response) != -1)
            ++packetsReceived;

      return packetsReceived;
   }

   /**
    * Transmit a single ICMP echo request packet to a node.
    *
    * @param addr ip address to be pinged
    *
    * @return true if a response is received
    *
    * @see java.net.InetAddress
    */
   public static boolean pingNode (InetAddress addr)
   {
      byte[] response = new byte [256];

      return (pingNative(addr.getAddress(), DEFAULT_TTL, response) != -1)
             ? true
             : false;
   }

   /**
    * <P>
    * Transmit a single ICMP echo request packet to a node.
    * </P>
    * <P>
    * This version of <CODE>pingNode</CODE> requires two additional parameters:
    * the ttl byte specifying the time to live field in the IP datagram header
    * of the oubound ICMP echo request and a byte array which is filled in with
    * entire IP datagram received in response to the echo request. The method
    * returns the time, measured in milliseconds, between transmitting the
    * ICMP echo request message and receiving a response. If no response is
    * received prior to the time-out period <CODE>pingNode</CODE> returns -1.
    * The round trip time (RTT) estimate is measured in the native network 
    * stack and is therefore reasonably (within a few milliseconds) accurate.
    * Any inaccuracy is on the high side and therefore the return value of
    * <CODE>pingNode</CODE> provides an upper bound on the true round trip time.
    * </P>
    * <P>
    * As mentioned the response buffer receives the entire IP datagram containing
    * the echo request response. Within the buffer the ICMP header and ICMP data are
    * located after the IP header and can be located by calculating the IP header
    * length as <CODE>int ipHdrLength = (response[0] & 0x0f) << 2);</CODE>. Using the
    * IP header length as an offset the first field of the ICMP header, ICMP packet type,
    * can be accessed as <CODE>int type = response[ipHdrLength];</CODE>
    * </P>
    * <P>
    * ICMP is specified in RFC 792 
    * </P>
    * @param addr ip address to be pinged
    * @param ttl Time To Live (hop count) used in the IP header
    * @param response buffer filled in with the repsonse, the
    *                 length of the array should be >= 128
    *
    * @return the round trip time (between request and reply) in
    *         milliseconds, -1 if no response received
    *
    * @see java.net.InetAddress
    */
   public static long pingNode (InetAddress addr, byte ttl, byte[] response)
   {
      return pingNative(addr.getAddress(), ttl, response);
   }

   //Native methods
   private static native long pingNative (byte[] ipAddress, byte ttl,
                                          byte[] response);
}
