
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.tininet.ppp;

import java.util.*;

/**
 * Propagates <CODE>PPPEvents</CODE>.
 *
 * <P>
 * It is easiest to understand the operation of PPP as a Finite State Machine (FSM). Note that there are actually
 * two Finite State Machines at work here. The true PPP FSM that is implemented as part of the network stack
 * beneath the IP module and the high level state machine implemented by a <CODE>PPPEventListener</CODE> whose
 * state transitions are driven by events generated by the PPP deamon thread and method invocations on a PPP object.
 * The PPP FSM is more complex than the event listener FSM and includes several additional states. For the most part
 * the arcane details of native implementation are hidden from the application developer by the PPP class. The purpose
 * of the <CODE>PPPEventListener</CODE> interface is to provide a mechanism to drive a much simpler, higher level
 * state machine that gives the application an opportunity to control the physical data link, user authentication
 * and handling of error information.
 * </P>
 *
 * <P>
 * After creating a new PPP object the FSM is in the <CODE>INIT</CODE> state. At this point there is no PPP traffic
 * traveling across the physical link. To transition to the <CODE>STARTING</CODE> state the owner of the PPP object
 * invokes the <CODE>PPP.open</CODE> method.
 * </P>
 *
 * <P>
 * The <CODE>STARTING</CODE> state provides the application with a chance to intialize the physical data link, a
 * serial port object. Intialization of the physical link can be as simple as openeing the port and setting the speed
 * and bit counts or as complicated as initializing a modem and initiating or accepting a call. Once the physical layer
 * has been initialized and the application is prepared to start PPP link negotiation the application invokes the
 * <CODE>PPP.up</CODE> method passing a serial port object as the argument. Once the serial port reference is passed
 * to PPP it assumes exclusive access to the serial port. At this point PPP attempts to bring up the link. If any
 * error occurs during the link negotiation phase PPP will issue a <CODE>STOPPED</CODE> event at which point the
 * user application can examine the error code associated with the <CODE>STOPPED</CODE> event.
 * </P>
 *
 * <P>
 * Most errors are fatal with the exception of timeouts. In the event of a timeout the user application can re-envoke the
 * <CODE>ppp.UP</CODE> method. Error codes and thier descriptions can be found in the <CODE>PPP.event</CODE> class
 * documentation.
 * </P>
 *
 * <P>
 * Once the initial negotiation has completed and if this peer has been configured to request authentication from the remote
 * peer an <CODE>AUTHENTICATION_REQUESTED</CODE> event will be generated when the remote peer presents it's authentication
 * password and username. After the application has examined the authentication infromation the application then invokes
 * <CODE>PPP.authenticate</CODE> with an argument of <CODE>true</CODE> to signal the lower level FSM that link negotiation
 * can continue or with and argument with <CODE>false</CODE> argument indicating that link negotiation should be terminated.
 * </P>
 *
 * <P>
 * When link negotiation is successfuly completed an <CODE>UP</CODE> event is generated at which point the application should
 * add the PPP interface to the network interface table by invoking the <CODE>PPP.addInterface()</CODE> method.
 * </P>
 *
 * <P>
 * <H2>Figure1. PPP "client/server" FSM Diagram</H2>
 * <PRE>
 *                  +-----------+
 *                  |           |
 *                  |   INIT    |
 *                  |           |
 *                  +-----------+
 *                        |
 *                        | open()
 *                        V
 *                  +-----------+
 *          open()  |           |       up()
 *       +--------->|   START   |------------------OR---------------------+
 *       |          |           |                  |                      |
 *       |          +-----------+                  | "server" operation   |
 *       |                |                        V with authentication  |
 * +-----------+          |                  +-----------+                |
 * |           |          | Link             |           |                |
 * |  CLOSED   |          | Negotiation      |   AUTH    |                |
 * |           |          | Error            |           |                |
 * +-----------+          |                  +-----------+                |
 *       ^                |                    |      |                   |
 *       |                V                    |      |     "server" or   |
 *       |          +-----------+              |      |      "client"     |
 *       |  close() |           |              |      |      operation    |
 *       +----------|  STOPPED  |<-------------+      |      with out     |
 *       |          |           | authenticate(false) |    authentication |
 *       |          +-----------+                     |                   |
 *       |                ^                           |                   |
 *       |                | remote termination        |                   |
 *       |                |                           |                   |
 *       |          +-----------+ authenticate(true)  |                   |
 *       |  close() |           |<--------------------+                   |
 *       +----------|    UP     |                                         |
 *                  |           |<----------------------------------------+
 *                  +-----------+
 * </PRE>
 * </P>
 *
 * @see PPP
 * @see PPPEvent
 */


public interface PPPEventListener
   extends EventListener
{

   /**
    *  Propagates a PPPEvent event.
    */
   public abstract void pppEvent (PPPEvent ev);
}
